/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: EAIndexer.cpp,v 1.12 2007/12/30 21:30:13 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <Ferris/EAIndexer.hh>
#include <Ferris/EAIndexer_private.hh>
#include <Ferris/FullTextIndexer_private.hh>
#include <Ferris/Configuration_private.hh>
#include <Ferris/FerrisBoost.hh>
#include <Ferris/Trimming.hh>
#include <Ferris/Iterator.hh>
#include <Ferris/EAQuery.hh>

using namespace std;

namespace Ferris
{
    namespace EAIndex 
    {
        const string IDXMGR_EXPLICIT_EANAMES_TO_INDEX_K = "idxmgr-explicit-eanames-to-index-k";

        const string IDXMGR_EANAMES_TO_INDEX_REGEX_K = "idxmgr-eanames-to-index-regex-k";


        const string IDXMGR_EANAMES_IGNORE_K = "idxmgr-eanames-ignore-k";
//         const string IDXMGR_EANAMES_IGNORE_DEFAULT = "as-xml,as-text,content,"
//         /**/  "block-count,dontfollow-block-count,inode,dontfollow-inode,dontfollow-size,"
//         /**/  "fs-available-block-count,fs-file-nodes-free,fs-free-block-count,"
//         /**/  "size-human-readable,atime-ctime,ctime-ctime,mtime-ctime,"
//         /**/  "atime-display,ctime-display,mtime-display,"
//         /**/  "md2,mdc2,attribute-count,block-size,device-type,"
//         /**/  "fs-block-count,fs-block-size,fs-file-name-length-maximum,fs-file-nodes-total,"
//         /**/  "fs-type,protection-ls,force-passive-view,"
//         /**/  "recommended-ea,recommended-ea-union,recommended-ea-union-view,"
//         /**/  "download-if-mtime-since-ctime,download-if-mtime-since-display"
//         /**/  "";
        
        const string IDXMGR_EANAMES_REGEX_IGNORE_K = "idxmgr-eanames-regex-ignore-k";
//         const string IDXMGR_EANAMES_REGEX_IGNORE_DEFAULT = (string)"schema:.*" + '\0' 
//         /**/                                                     + "recursive-.*" + '\0'
//         /**/                                                     + "dontfollow-.*" + '\0'
//         /**/                                                     + "subcontext.*";
        const string IDXMGR_MAX_VALUE_SIZE_K = "idxmgr-max-value-size-k";
//        const string IDXMGR_MAX_VALUE_SIZE_DEFAULT = "200";

        const string IDXMGR_FULLTEXT_INDEX_PATH_K = "idxmgr-fulltext-index-path-k";

        
        
        void SET_EAINDEX_EANAMES_NOT_TO_INDEX_DEFAULT( const std::string& v )
        {
            setEDBString( GET_FDB_GENERAL(), "EAINDEX_EANAMES_NOT_TO_INDEX", v );
        }
        void SET_EAINDEX_EANAMES_REGEX_IGNORE_DEFAULT( const std::string& v )
        {
            setEDBString( GET_FDB_GENERAL(), "EAINDEX_EANAMES_REGEX_IGNORE", v );
        }
        void SET_EAINDEX_MAX_VALUE_SIZE_TO_INDEX_DEFAULT( const std::string& v )
        {
            setEDBString( GET_FDB_GENERAL(), "EAINDEX_MAX_VALUE_SIZE_TO_INDEX", v );
        }
        void SET_EAINDEX_EANAMES_TO_INDEX_REGEX_DEFAULT( const std::string& v )
        {
            setEDBString( GET_FDB_GENERAL(), "EAINDEX_EANAMES_TO_INDEX_REGEX", v );
        }

        std::string GET_EAINDEX_EANAMES_NOT_TO_INDEX_DEFAULT()
        {
            return getEDBString(
                GET_FDB_GENERAL(), "EAINDEX_EANAMES_NOT_TO_INDEX",
                "name,language-human,content,as-xml,as-text,as-rdf,"
                "block-count,dontfollow-block-count,inode,dontfollow-inode,dontfollow-size,"
                "dontfollow-block-size,dontfollow-device,dontfollow-device-type,"
                "dontfollow-filesystem-filetype,dontfollow-hard-link-count,"
                "recommended-ea-short,"
                "associated-branches,depth-per-color,xfs-ea-names,exif:ea-names,"
                "fs-available-block-count,fs-file-nodes-free,fs-free-block-count,"
                "size-human-readable,"
                "attribute-count,block-size,device-type,"
                "fs-block-count,fs-block-size,fs-file-name-length-maximum,fs-file-nodes-total,"
                "fs-type,protection-ls,force-passive-view,"
                "recommended-ea,recommended-ea-union,recommended-ea-union-view,"
                "is-dir-try-automounting,"
                "rgba-32bpp,exif:thumbnail-update,exif:thumbnail-rgba-32bpp,md2,sha1,crc32,crc32-is-valid,mdc2,download-if-mtime-since,force-passive-view,path,realpath,emblem:upset,emblem:list,emblem:list-ui" );
        }
        std::string GET_EAINDEX_EANAMES_REGEX_IGNORE_DEFAULT( bool commaSeperated )
        {
            string def = (std::string)"^schema:.*" + '\0'
                /**/ + "^recursive-.*" + '\0'
                /**/ + "^subcontext.*" + '\0'
                /**/ + "^branchfs-.*" + '\0'
                /**/ + "^associated-branches-.*" + '\0'
                /**/ + "^as-.*" + '\0'
                /**/ + "^medallion.*" + '\0'
                /**/ + ".*-ctime$" + '\0'
                /**/ + ".*-display$" + '\0'
                /**/ + ".*-cdrom-count$" + '\0'
                /**/ + ".*-dvd-count$" + '\0'
                /**/ + ".*-human-readable$" + '\0'
                /**/ + ".*-granularity$" + '\0'
                /**/ + "^emblem:id-fuzzy.*$" + '\0'
                /**/ + "^emblem:has-fuzzy.*$" + '\0'
                /**/ + "^emblem:.*-mtime.*$" + '\0'
                /**/ + "^geospatial-.*$" + '\0'
                ;
            
            string ret = getEDBString( GET_FDB_GENERAL(), "EAINDEX_EANAMES_REGEX_IGNORE", def );

//             cerr << "GET_EAINDEX_EANAMES_REGEX_IGNORE_DEFAULT(cs:" << commaSeperated << ") "
//                  << " FDB_GENERAL:" << GET_FDB_GENERAL()
//                  << " ret:" << ret
//                  << endl;
            if( commaSeperated )
                replace( ret.begin(), ret.end(), '\0', ',' );
            
            return ret;
        }
        std::string GET_EAINDEX_MAX_VALUE_SIZE_TO_INDEX_DEFAULT()
        {
            return getEDBString(
                GET_FDB_GENERAL(), "EAINDEX_MAX_VALUE_SIZE_TO_INDEX",
                "1024" );
        }
        std::string GET_EAINDEX_EANAMES_TO_INDEX_REGEX_DEFAULT()
        {
            return getEDBString( GET_FDB_GENERAL(), "EAINDEX_EANAMES_TO_INDEX_REGEX", ".*" );
        }
        



        

        const string IDXMGR_NONRESOLVABLE_NOT_TO_REMOVE_REGEX_K = "idxmgr-nonresolvable-not-to-remove-regex-k";
        void SET_EAINDEX_NONRESOLVABLE_NOT_TO_REMOVE_REGEX_DEFAULT( const std::string& v )
        {
            setEDBString( GET_FDB_GENERAL(), "EAINDEX_NONRESOLVABLE_NOT_TO_REMOVE_REGEX", v );
        }
        std::string GET_EAINDEX_NONRESOLVABLE_NOT_TO_REMOVE_REGEX_DEFAULT( bool commaSeperated )
        {
            string def = (std::string)"" + '\0';
            string ret = getEDBString( GET_FDB_GENERAL(), "EAINDEX_NONRESOLVABLE_NOT_TO_REMOVE_REGEX", def );

            if( commaSeperated )
                replace( ret.begin(), ret.end(), '\0', ',' );
            
            return ret;
        }
        
        
        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/

        DocumentIndexer::DocumentIndexer( fh_idx idx )
            :
            m_idx( idx ),
            m_dontCheckIfAlreadyThere( false ),
            m_filesIndexedCount( 0 ),
            m_haveCalledPrepareForWrites( false ),
            m_AddEvenIfAlreadyCurrent( false ),
            m_haveReportedIndexDoesNotSupportAddEvenIfAlreadyCurrent( false ),
            m_retireNonExistentDocumentsToMuliversion( false ),
            m_NonExistentDocumentsState( 0 )
        {
//             cerr << "m_idx->getEANamesIgnore() s:" << m_idx->getEANamesIgnore() << endl;
            setIgnoreEANames( m_idx->getEANamesIgnore() );
            appendIgnoreEARegexs( m_idx->getEANamesIgnoreRegexes() );
            setMaximunValueSize( m_idx->getMaxValueSize() );
            appendNonResolvableURLsNotToRemoveRegexes( m_idx->getNonResolvableURLsNotToRemoveRegexes() );
            setEANamesToIndexRegex( m_idx->getEANamesToIndexRegex() );
        }

        DocumentIndexer::~DocumentIndexer()
        {
            if( m_idx )
                m_idx->allWritesComplete();
                
            if( m_idx )
                m_idx->sync();
        }
        
        AddToEAIndexProgress_Sig_t&
        DocumentIndexer::getProgressSig()
        {
            return m_progressSig;
        }


        void
        DocumentIndexer::setIgnoreEANames( const std::string& s )
        {
            clearIgnoreEANames();
            appendIgnoreEANames( s );
        }

        void
        DocumentIndexer::appendIgnoreEANames( const std::string& s )
        {
            stringlist_t sl = Util::parseCommaSeperatedList( s );
            copy( sl.begin(), sl.end(), inserter( m_ignoreEANames, m_ignoreEANames.end() ));
        }

        void
        DocumentIndexer::clearIgnoreEANames()
        {
            m_ignoreEANames.clear();
        }
        
        stringset_t&
        DocumentIndexer::getIgnoreEANames()
        {
            return m_ignoreEANames;
        }
        

        void
        DocumentIndexer::appendIgnoreEARegexs( const std::string& s )
        {
            m_ignoreEARegexs.append( s );
        }

        void
        DocumentIndexer::appendIgnoreEARegexs( const stringlist_t& sl )
        {
            m_ignoreEARegexs.append( sl );
        }
        
        void
        DocumentIndexer::clearIgnoreEARegexs()
        {
            m_ignoreEARegexs.clear();
        }

        DocumentIndexer::m_ignoreEARegexs_t&
        DocumentIndexer::getIgnoreEARegexs()
        {
            return m_ignoreEARegexs;
        }



        
        void
        DocumentIndexer::appendNonResolvableURLsNotToRemoveRegexes( const std::string& s )
        {
            m_nonResolvableURLsNotToRemoveRegexes.append( s );
        }

        
        void DocumentIndexer::appendNonResolvableURLsNotToRemoveRegexes( const stringlist_t& sl )
        {
            m_nonResolvableURLsNotToRemoveRegexes.append( sl );
        }
        
        void DocumentIndexer::clearNonResolvableURLsNotToRemoveRegexes()
        {
            m_nonResolvableURLsNotToRemoveRegexes.clear();
        }
        RegexCollection&
        DocumentIndexer::getNonResolvableURLsNotToRemoveRegexes()
        {
            return m_nonResolvableURLsNotToRemoveRegexes;
        }
        
        void
        DocumentIndexer::setEANamesToIndex( const stringlist_t& sl )
        {
            m_EANamesToIndex.clear();
            copy( sl.begin(), sl.end(), inserter( m_EANamesToIndex, m_EANamesToIndex.end() ) );
        }
        
        stringlist_t
        DocumentIndexer::getEANamesToIndex()
        {
            stringlist_t ret;
            copy( m_EANamesToIndex.begin(),
                  m_EANamesToIndex.end(),
                  back_inserter( ret ));
            return ret;
        }
        
        bool
        DocumentIndexer::EANamesToIndexContains( const std::string& s )
        {
            if( !m_EANamesToIndex.empty() )
            {
                if( m_EANamesToIndex.find( s ) == m_EANamesToIndex.end() )
                    return false;
            }
            return true;
        }

        

        void
        DocumentIndexer::setEANamesToIndexRegex( const std::string& s )
        {
            m_EANamesToIndexRegex = 0;
            m_EANamesToIndexRegexString = "";
            if( !s.empty() )
                m_EANamesToIndexRegex = toregexh( s );
            m_EANamesToIndexRegexString = s;
        }
        
        std::string&
        DocumentIndexer::getEANamesToIndexRegex()
        {
            return m_EANamesToIndexRegexString;
        }
        
        bool
        DocumentIndexer::EANamesToIndexRegexContains( const std::string& s )
        {
            if( m_EANamesToIndexRegex )
            {
                if( !regex_match( s, m_EANamesToIndexRegex ))
                {
                    return false;
                }
            }
            return true;
        }
        
        
        
        void
        DocumentIndexer::setMaximunValueSize( std::streamsize sz )
        {
            m_maxValueSize = sz;
        }

        std::streamsize
        DocumentIndexer::getMaximunValueSize()
        {
            return m_maxValueSize;
        }
        
        void
        DocumentIndexer::setDontCheckIfAlreadyThere( bool v )
        {
            m_dontCheckIfAlreadyThere = v;
        }
        
        bool
        DocumentIndexer::getDontCheckIfAlreadyThere()
        {
            return m_dontCheckIfAlreadyThere;
        }

        int
        DocumentIndexer::getFilesIndexedCount()
        {
            return m_filesIndexedCount;
        }

        void
        DocumentIndexer::setAddEvenIfAlreadyCurrent( bool v )
        {
            m_AddEvenIfAlreadyCurrent = v;
        }

        void
        DocumentIndexer::setRetireNonExistentDocumentsToMuliversion( bool v )
        {
            m_retireNonExistentDocumentsToMuliversion = v;
        }

        class NonExistentDocumentsState
            :
            public Handlable
        {
            fh_context m_dirContext;

            stringset_t m_indexedDocs;
        public:
            NonExistentDocumentsState( fh_context dc );
            void add( fh_context c );
            void retire( fh_context dc, fh_docindexer di );
        };
        
        NonExistentDocumentsState::NonExistentDocumentsState( fh_context dc )
            :
            m_dirContext( dc )
        {
            LG_EAIDX_D << "NonExistentDocumentsState(ctor) dc:" << dc->getURL() << endl;
        }
        
        void
        NonExistentDocumentsState::add( fh_context c )
        {
            LG_EAIDX_D << "NonExistentDocumentsState::add() c:" << c->getURL() << endl;
            m_indexedDocs.insert( c->getURL() );
        }
        
        void
        NonExistentDocumentsState::retire( fh_context dc, fh_docindexer di )
        {
            LG_EAIDX_D << "NonExistentDocumentsState::retire() dc:" << dc->getURL() << endl;
            // Steps:
            // find all the filenames for directory dc
            // remove the filenames we have seen
            // retire the filenames that were there that we didn't find this time


            // Find existing docids from database
            stringset_t dbDocs;
            {
                stringstream qss;
                qss << "(parent-url==" << dc->getURL() << ")";
                ExecuteQueryToToURLs( qss.str(), di->m_idx, dbDocs );
            }

            // Find which are in the database but not seen now.
            stringset_t delme;
            set_difference( dbDocs.begin(), dbDocs.end(),
                            m_indexedDocs.begin(), m_indexedDocs.end(),
                            inserter( delme, delme.end() ) );

            // Remove old docs.
            if( !delme.empty() )
            {
                if( LG_EAIDX_D_ACTIVE )
                {
                    LG_EAIDX_D << "NonExistentDocumentsState::retire() removing docids.sz:"
                               << delme.size() << endl;
                    for( stringset_t::iterator si = delme.begin(); si!=delme.end(); ++si )
                    {
                        LG_EAIDX_D << "retire doc:" << *si << endl;
                    }
                }

                di->m_idx->retireDocumentsFromIndex( delme );
            }
        }
        
        
        void
        DocumentIndexer::EnteringContext(fh_context ctx)
        {
            if( m_retireNonExistentDocumentsToMuliversion )
            {
                m_NonExistentDocumentsState = new NonExistentDocumentsState( ctx );
            }
        }
        
        void
        DocumentIndexer::LeavingContext(fh_context ctx)
        {
            if( m_NonExistentDocumentsState )
            {
                m_NonExistentDocumentsState->retire( ctx, this );
                m_NonExistentDocumentsState = 0;
            }
        }
        
        
        void
        DocumentIndexer::addContextToIndex( fh_context c )
        {
            if( !m_haveCalledPrepareForWrites )
            {
                m_haveCalledPrepareForWrites = true;
                int f = MetaEAIndexerInterface::PREPARE_FOR_WRITES_NONE;
                if( !m_AddEvenIfAlreadyCurrent )
                    f |= MetaEAIndexerInterface::PREPARE_FOR_WRITES_ISNEWER_TESTS;

                m_idx->prepareForWrites( f );
            }

            LG_EAIDX_D << "DocumentIndexer::addContextToIndex() c:" << c->getURL() << endl;
            
            try
            {
                LG_EAIDX_D << "DocumentIndexer::addContextToIndex(AA) c:" << c->getURL() << endl;
                
                if( !m_AddEvenIfAlreadyCurrent )
                {
                    bool sup = m_idx->getIndexMethodSupportsIsFileNewerThanIndexedVersion();
                    if( !sup && !m_haveReportedIndexDoesNotSupportAddEvenIfAlreadyCurrent )
                    {
                        m_haveReportedIndexDoesNotSupportAddEvenIfAlreadyCurrent = true;
                        cerr << "WARNING: index doesn't support isNewer() checks" << endl;
                    }
                    
                    if( sup )
                    {
                        bool isNewer = m_idx->isFileNewerThanIndexedVersion( c );
                        if( !isNewer )
                        {
                            LG_EAIDX_D << "Skipping because index is current for c:"
                                       << c->getURL() << endl;

                            // retain previous metadata as is.
                            if( m_retireNonExistentDocumentsToMuliversion )
                            {
                                m_NonExistentDocumentsState->add( c );
                            }
                            
                            return;
                        }
                    }
                }

                LG_EAIDX_D << "DocumentIndexer::addContextToIndex(BB) c:" << c->getURL() << endl;
                
                m_idx->visitingContext( c );

                
                LG_EAIDX_D << "DocumentIndexer::addContextToIndex(calling addToIndex) c:"
                           << c->getURL() << endl;

                if( m_retireNonExistentDocumentsToMuliversion )
                {
                    m_NonExistentDocumentsState->add( c );
                }

                m_idx->addToIndex( c, this );
                ++m_filesIndexedCount;
            }
            catch( exception& e )
            {
                // The file existed, but there was an error, don't wipe the old metadata.
                if( m_retireNonExistentDocumentsToMuliversion )
                {
                    m_NonExistentDocumentsState->add( c );
                }
                
                LG_EAIDX_W << "DocumentIndexer::addContextToIndex(1) e:" << e.what() << endl;
                m_idx->sync();
                LG_EAIDX_W << "DocumentIndexer::addContextToIndex(2) e:" << e.what() << endl;
                throw e;
            }
        }


        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/
        
    };
};

