/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: FerrisKDE3.cpp,v 1.1 2005/07/04 08:56:42 ben Exp $

    *******************************************************************************
    *******************************************************************************
    ******************************************************************************/

#include "config.h"
#include "FerrisKDE3.hh"

#include <Ferris.hh>
#include <Native.hh>

#ifdef HAVE_KDE3
#include <qapplication.h>
#include <kmimetype.h>
#include <kdebug.h>
#include <kapplication.h>
#include <kaboutapplication.h>
#include <kcmdlineargs.h>
#include <kmimemagic.h>
#endif



using namespace std;


namespace Ferris
{
    namespace KDE
    {
#ifdef HAVE_KDE3

        /**
         * This exists because folks might want to use libferris compiled
         * with KDE mime sniffing from an app which is not a KDE client.
         *
         * ie. We dont want to force command line clients to know anything
         * about KDE if they just want to getStrAttr( ctx, "mimetype" );
         */
        class FERRISEXP_DLLLOCAL KAppHolder
        {
            KApplication& getKApp()
                {
                    static KApplication a( false, false );
                    return a;
                }

            bool m_weCreatedKApp;
    
        public:
            KAppHolder()
                :
                m_weCreatedKApp( false )
                {
                    if( !KApplication::kApplication() )
                    {
                        m_weCreatedKApp = true;
                
                        int argc = 1;
                        char* argv[] = { "libferris", 0 };
                        KCmdLineArgs::init( argc, argv, argv[0], argv[0], "1.0", true );
                        getKApp();
                    }
                }
    
            ~KAppHolder()
                {
                    if( m_weCreatedKApp )
                    {
                        getKApp().unlock();
                    }
                }
        };

        typedef Loki::SingletonHolder< KAppHolder > KAppSingleton;
#endif
    
    
        string getMimeType( fh_context c, bool fromContent  )
        {
#ifdef HAVE_KDE3
            KAppSingleton::Instance();

            if( fromContent )
            {
                const int sz = 4096;
                QByteArray array( sz + 1 );

                fh_istream iss = c->getIStream();
                iss.read( array.data(), sz );
            
                KMimeMagicResult *result = KMimeMagic::self()->findBufferType( array );
                return result->mimeType().ascii();
            }
            else
            {
                KMimeType::Ptr type = KMimeType::findByURL( c->getDirPath().c_str() );
                return type->name().ascii();
            }
#endif
            return "document/unknown";
        }
    
//         string getMimeType( NativeContext* c, bool fromContent  )
//         {
// #ifdef HAVE_KDE3
//             KAppSingleton::Instance();
//             KMimeMagicResult* result = KMimeMagic::self()->findFileType( c->getDirPath().c_str() );
//             return result->mimeType().ascii();
// #endif
//             return "document/unknown";
//         }
    };
    
};
