/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: Ferrisls_AggregateData.cpp,v 1.2 2006/07/13 16:25:43 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <Ferrisls_AggregateData.hh>
#include <Ferrisls.hh>
#include <Native.hh>

namespace Ferris
{

    Ferrisls_aggregate_data::Ferrisls_aggregate_data()
    {
        reset();
        RecordSeperator = "";
    }


    void
    Ferrisls_aggregate_data::UpdateAggregateData( Ferrisls_aggregate_t& d, fh_context ctx )
    {
        d.count++;
            
        Context* c = GetImpl(ctx);
        if( NativeContext* nc = dynamic_cast<NativeContext*>(c) )
        {
            const struct stat& sb = nc->getStat_Follow();

            d.hardlinkcount  = sb.st_nlink;
            d.size          += sb.st_size;
            d.sizeinblocks  += sb.st_blocks; // * sb.st_blksize;

            if(S_ISREG(sb.st_mode))        d.filecount++;
            else if(S_ISDIR(sb.st_mode))   d.dircount++;

            if( sb.st_atime < d.oldestatime )
            {
                d.oldestatime     = sb.st_atime;
                d.oldestatime_url = ctx->getURL();
            }
            if( sb.st_mtime < d.oldestmtime )
            {
                d.oldestmtime     = sb.st_mtime;
                d.oldestmtime_url = ctx->getURL();
            }
            if( sb.st_ctime < d.oldestctime )
            {
                d.oldestctime     = sb.st_ctime;
                d.oldestctime_url = ctx->getURL();
            }
        }
    }
    
    
    void
    Ferrisls_aggregate_data::ShowAttributes( fh_context ctx )
    {
        if( isDirectSubContextOfRootContext( ctx ) )
        {
            UpdateAggregateData( subd, ctx );
        }
        
        UpdateAggregateData( recd, ctx );
    }
        
    void
    Ferrisls_aggregate_data::ShowHeader(fh_context ctx)
    {
        recd.maxdepth = std::max( recd.maxdepth, getDepthOfContext( ctx ));
    }

    bool
    Ferrisls_aggregate_data::isDirectSubContextOfRootContext( fh_context c )
    {
        if( !isBound( RootContext ) || !c->isParentBound() )
        {
            return false;
        }

        return c->getParent() == GetImpl(RootContext);
    }

    guint32
    Ferrisls_aggregate_data::getDepthOfContext( fh_context c )
    {
        guint32 d = 0;
        
        while( GetImpl(c) != GetImpl(RootContext) )
        {
            c = c->getParent();
            d++;
        }
        return d;
    }
    
    
    
    void
    Ferrisls_aggregate_data::reset()
    {
        recd.reset();
        subd.reset();
    }

    Ferrisls_aggregate_t&
    Ferrisls_aggregate_data::getRecursiveData()
    {
        return recd;
    }

    Ferrisls_aggregate_t&
    Ferrisls_aggregate_data::getSubData()
    {
        return subd;
    }
    
    Ferrisls_aggregate_t&
    Ferrisls_aggregate_data::getData( int m )
    {
        if( m & AGGDATA_RECURSIVE )
        {
            return getRecursiveData();
        }
        return getSubData();
    }
    
    aggdata_mode_t
    Ferrisls_aggregate_data::getMode()
    {
        return Mode;
    }
    
    void
    Ferrisls_aggregate_data::setRootContext( fh_context c )
    {
        RootContext = c;
    }
    
    
        
    /*******************************************************************************/
    /*******************************************************************************/
    /*******************************************************************************/
        
    void
    Ferrisls_aggregate_t::reset()
    {
        count = 0;
        maxdepth = 0;
        hardlinkcount = 0;
        size = 0;
        filecount = 0;
        dircount = 0;
        sizeinblocks = 0;
        oldestmtime = 0;
        oldestctime = 0;
        oldestatime = 0;
        oldestmtime_url = "";
        oldestctime_url = "";
        oldestatime_url = "";
    }
        

    fh_display_aggdata getAggregateData( fh_context c,
                                         aggdata_mode_t m )
    {
        fh_display_aggdata  d = new Ferrisls_aggregate_data();
        fh_ls     ls          = new Ferrisls();

        d->reset();
        d->setRootContext( c );

        ls->setURL( c->getURL() );
        ls->setDisplay( d );
        ls->setRecursiveList( m & AGGDATA_RECURSIVE );
        ls->setDontDecendIntoFiles( true );
        
        ls->operator()();
        d->UpdateAggregateData( d->getData( m ), c );
            
        return d;
    }
        
    
};

