/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: LinkContextSchemeClients.cpp,v 1.6 2007/04/08 21:30:13 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <LinkContextSchemeClients.hh>
#include "config.h"

namespace Ferris
{
    namespace Private
    {
        const std::string link_ctx_applications_name = "applications";
//        const std::string link_ctx_applications_url  = "~/.ferris/apps.db";
        const std::string link_ctx_applications_url  = "apps://";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_applications_name, link_ctx_applications_url > >
        link_ctx_applications_singleton;
        static RootContextDropper& applicationsd = link_ctx_applications_singleton::Instance();

        const std::string link_ctx_apps_name = "apps";
        const std::string link_ctx_apps_url  = "~/.ferris/apps.db";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_apps_name, link_ctx_apps_url > > link_ctx_apps_singleton;
        static RootContextDropper& appsd = link_ctx_apps_singleton::Instance();

        
        const std::string link_ctx_events_name = "events";
        const std::string link_ctx_events_url  = "~/.ferris/eventbind.db";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_events_name, link_ctx_events_url > > link_ctx_events_singleton;
         static RootContextDropper& eventsd = link_ctx_events_singleton::Instance();

        
        const std::string link_ctx_icons_name = "icons";
        const std::string link_ctx_icons_url  = AUTOTOOLS_CONFIG_DATADIR + "/ferris/icons";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_icons_name, link_ctx_icons_url > > link_ctx_icons_singleton;
         static RootContextDropper& iconsd = link_ctx_icons_singleton::Instance();

        const std::string link_ctx_gnomeicons_name = "gnomeicons";
        const std::string link_ctx_gnomeicons_url  = "/usr/share/pixmaps";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_gnomeicons_name, link_ctx_gnomeicons_url > > link_ctx_gnomeicons_singleton;
         static RootContextDropper& gnomeiconsd = link_ctx_gnomeicons_singleton::Instance();
        
        
        const std::string link_ctx_mime_name = "mime";
        const std::string link_ctx_mime_url  = "~/.ferris/mimebind.db";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_mime_name, link_ctx_mime_url > > link_ctx_mime_singleton;
         static RootContextDropper& mimed = link_ctx_mime_singleton::Instance();

        const std::string link_ctx_fileclip_name = "file-clipboard";
        const std::string link_ctx_fileclip_url  = "~/.ferris/file-clipboard.db";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_fileclip_name, link_ctx_fileclip_url > > link_ctx_fileclip_singleton;
         static RootContextDropper& fileclipd = link_ctx_fileclip_singleton::Instance();


//         const std::string link_ctx_news_name = "news";
//         const std::string link_ctx_news_url  = "~/.fnews/feeds";
//         typedef Loki::SingletonHolder<
//             LinkContextSchemeVFS_RootContextDropper<
//             link_ctx_news_name, link_ctx_news_url > > link_ctx_news_singleton;
//          static RootContextDropper& newsd = link_ctx_news_singleton::Instance();

        const std::string link_ctx_bookmarks_name = "bookmarks";
        const std::string link_ctx_bookmarks_url  = "~/.ferris/bookmarks.db";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_bookmarks_name, link_ctx_bookmarks_url > > link_ctx_bookmarks_singleton;
         static RootContextDropper& bookmarksd = link_ctx_bookmarks_singleton::Instance();

        const std::string link_ctx_eaq_name = "eaq";
        const std::string link_ctx_eaq_url  = "eaquery://filter/";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_eaq_name, link_ctx_eaq_url > > link_ctx_eaq_singleton;
         static RootContextDropper& eaqd = link_ctx_eaq_singleton::Instance();

        const std::string link_ctx_ftxq_name = "ftxq";
        const std::string link_ctx_ftxq_url  = "fulltextquery://";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_ftxq_name, link_ctx_ftxq_url > > link_ctx_ftxq_singleton;
         static RootContextDropper& ftxqd = link_ctx_ftxq_singleton::Instance();
        

        
        const std::string link_ctx_rdf_name = "myrdf";
        const std::string link_ctx_rdf_url  = "~/.ferris/rdfdb/myrdf-sp2o.db";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_rdf_name, link_ctx_rdf_url > > link_ctx_rdf_singleton;
         static RootContextDropper& rdfd = link_ctx_rdf_singleton::Instance();
        static bool ctx_rdf_b = ImplementationDetail::appendToStaticLinkedRootContextNames("myrdf");


//         const std::string link_ctx_schema_name = "schema";
//         const std::string link_ctx_schema_url  = "~/.ferris/schema.xml";
//         typedef Loki::SingletonHolder<
//             LinkContextSchemeVFS_RootContextDropper<
//             link_ctx_schema_name, link_ctx_schema_url > > link_ctx_schema_singleton;
//          static RootContextDropper& schemad = link_ctx_schema_singleton::Instance();

        const std::string link_ctx_camera_name = "camera";
        const std::string link_ctx_camera_url  = "gphoto:/";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_camera_name, link_ctx_camera_url > > link_ctx_camera_singleton;
         static RootContextDropper& camerad = link_ctx_camera_singleton::Instance();

        const std::string link_ctx_pg_name = "pg";
        const std::string link_ctx_pg_url  = "postgresql:/";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_pg_name, link_ctx_pg_url > > link_ctx_pg_singleton;
         static RootContextDropper& pgd = link_ctx_pg_singleton::Instance();

        const std::string link_ctx_flickr_name = "flickr";
        const std::string link_ctx_flickr_url  = "webphotos://flickr";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_flickr_name, link_ctx_flickr_url > > link_ctx_flickr_singleton;
         static RootContextDropper& flickrd = link_ctx_flickr_singleton::Instance();

        const std::string link_ctx_23hq_name = "23hq";
        const std::string link_ctx_23hq_url  = "webphotos://23hq";
        typedef Loki::SingletonHolder<
            LinkContextSchemeVFS_RootContextDropper<
            link_ctx_23hq_name, link_ctx_23hq_url > > link_ctx_23hq_singleton;
         static RootContextDropper& _23hqd = link_ctx_23hq_singleton::Instance();
        
    };
};


