/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: MetadataServer_private.hh,v 1.1 2007/09/03 00:52:27 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#ifndef _ALREADY_INCLUDED_FERRIS_METADATA_SERVER_PRIV_H_
#define _ALREADY_INCLUDED_FERRIS_METADATA_SERVER_PRIV_H_

#include <string>

#include <dbus-c++/dbus.h>
#include <dbus-c++/pendingcall.h>
#include <dbus-c++/types.h>
#include <dbus-c++/message.h>

#include <Ferris/Ferris.hh>
#include <Ferris/Image_private.hh>
#include <Ferris/FerrisEAGeneratorPlugin_private.hh>

namespace Ferris
{
    DBus::Connection getSessionBus();
    std::string syncMetadataServerGet( const std::string& earl,
                                       const std::string& rdn,
                                       DBus::Connection con = getSessionBus() );
    void syncMetadataServerPut( const std::string& earl,
                                const std::string& rdn,
                                const std::string& value,
                                DBus::Connection con = getSessionBus() );
    

    void resetup_glib_idle_to_close_callback_function( int interval, bool* running );


    stringset_t& getImageMetadataAttributes();
    bool isImageMetadataAttribute( const std::string& rdn );
    s_ImageEAGeneratorsExtensionToShortName_t& getOutOfProcess_ImageEAGeneratorsExtensionToShortName();


    StaticEAGenFactorys_t& getOutOfProcess_EAGeneratorsStaticFactories( StaticEAGenFactorys_t& ret );
    const stringset_t& getOutOfProcess_EAGeneratorsStaticFactoriesShortNamesToUse();

    bool isOutOfProcessMetadataAttribute( const std::string& rdn );
    
};


namespace Ferris
{
    inline std::ostream& operator << ( std::ostream& oss, const std::vector<DBus::Byte>& val )
    {
        std::copy( val.begin(), val.end(), std::ostream_iterator<char>( oss, "" ) );
        return oss;
    }

    /////////////////////////

    inline DBus::MessageIter& operator << ( DBus::MessageIter& iter, const DBus::String& val )
    {
        iter.append_string(val.c_str());
        return iter;
    }

    inline DBus::MessageIter& operator >> ( DBus::MessageIter& iter, DBus::String& val )
    {
        val = iter.get_string();
        return ++iter;
    }

    inline DBus::MessageIter& operator >> ( DBus::MessageIter& iter, DBus::Int32& val )
    {
        val = iter.get_int32();
        return ++iter;
    }

    template<typename E>
    inline DBus::MessageIter& operator >> ( DBus::MessageIter& iter, std::vector<E>& val )
    {
        if(!iter.is_array())
            throw DBus::ErrorInvalidArgs("array expected");

        DBus::MessageIter ait = iter.recurse();

        while(!ait.at_end())
        {
            E elem;

            ait >> elem;

            val.push_back(elem);
        }
        return ++iter;
    }

    template<>
    inline DBus::MessageIter& operator >> ( DBus::MessageIter& iter, std::vector<DBus::Byte>& val )
    {
        if(!iter.is_array())
            throw DBus::ErrorInvalidArgs("array expected");

        if(iter.array_type() != 'y')
            throw DBus::ErrorInvalidArgs("byte-array expected");

        DBus::MessageIter ait = iter.recurse();

        DBus::Byte* array;
        size_t length = ait.get_array(&array);

        val.insert(val.end(), array, array+length);

        return ++iter;
    }

    template<typename E>
    inline DBus::MessageIter& operator << ( DBus::MessageIter& iter, const std::vector<E>& val )
    {
        const std::string sig = DBus::type<E>::sig();
        DBus::MessageIter ait = iter.new_array(sig.c_str());

        typename std::vector<E>::const_iterator vit;
        for(vit = val.begin(); vit != val.end(); ++vit)
        {
            ait << *vit;
        }

        iter.close_container(ait);
        return iter;
    }

    template<>
    inline DBus::MessageIter& operator << ( DBus::MessageIter& iter, const std::vector<DBus::Byte>& val )
    {
        DBus::MessageIter ait = iter.new_array("y");
        ait.append_array('y', &val.front(), val.size());
        iter.close_container(ait);
        return iter;
    }
    
};


#endif
