/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: WebPhotosContext.cpp,v 1.4 2007/04/14 21:30:08 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/


#include "Ferris.hh"
#include "WebPhotos_private.hh"
#include "Resolver_private.hh"
#include <Ferris/Enamel.hh>


using namespace std;

namespace Ferris
{

    


    /**
     * root context for remembrance://
     */
    class FERRISEXP_DLLLOCAL WebPhotosRootContext
        :
        public StateLessEAHolder< WebPhotosRootContext, FakeInternalContext >
    {
        typedef WebPhotosRootContext _Self;
        typedef StateLessEAHolder< WebPhotosRootContext, FakeInternalContext > _Base;
        
        virtual void priv_FillCreateSubContextSchemaParts( CreateSubContextSchemaPart_t& m )
            {}
        
    protected:

        virtual void priv_read();
        
    public:

        WebPhotosRootContext()
            :
            _Base( 0, "/" )
            {
                createStateLessAttributes();
            }
    
        
        virtual ~WebPhotosRootContext()
            {}
        
        virtual fh_context
        createSubContext( const std::string& rdn, fh_context md = 0 )
            throw( FerrisCreateSubContextFailed, FerrisCreateSubContextNotSupported )
            {
                fh_stringstream ss;
                ss << "webphotos:// directory can not have new items created in this way" << endl;
                Throw_FerrisCreateSubContextNotSupported( tostr(ss), this );
            }

        void createStateLessAttributes( bool force = false )
            {
                if( force || isStateLessEAVirgin() )
                {
                    _Base::createStateLessAttributes( true );
                    supplementStateLessAttributes( true );
                }
            }
            
    };
    
    /************************************************************/
    /************************************************************/
    /************************************************************/

    class FERRISEXP_DLLLOCAL WebPhotosObjectTopLevelContext;
    FERRIS_CTX_SMARTPTR( WebPhotosObjectTopLevelContext, fh_WebPhotosObjectTopLevelContext );
    class FERRISEXP_DLLLOCAL WebPhotosObjectTopLevelContext
        :
        public StateLessEAHolder< WebPhotosObjectTopLevelContext, FakeInternalContext >
    {
        typedef StateLessEAHolder< WebPhotosObjectTopLevelContext, FakeInternalContext > _Base;
        
        virtual void priv_FillCreateSubContextSchemaParts( CreateSubContextSchemaPart_t& m )
            {}

        fh_webPhotos m_wf;
        
    protected:

        virtual void priv_read();

    public:

        WebPhotosObjectTopLevelContext( Context* parent,
                                        const std::string& rdn,
                                        fh_webPhotos wf )
            :
            _Base( parent, rdn ),
            m_wf( wf )
            {
                createStateLessAttributes();
            }
        
        virtual ~WebPhotosObjectTopLevelContext()
            {}


        virtual fh_context
        createSubContext( const std::string& rdn, fh_context md = 0 )
            throw( FerrisCreateSubContextFailed, FerrisCreateSubContextNotSupported )
            {
                fh_stringstream ss;
                ss << "webphotos:// directory can not have new items created in this way" << endl;
                Throw_FerrisCreateSubContextNotSupported( tostr(ss), this );
            }

        void createStateLessAttributes( bool force = false )
            {
                if( force || isStateLessEAVirgin() )
                {
                    _Base::createStateLessAttributes( true );
                    supplementStateLessAttributes( true );
                }
            }

        fh_webPhotos getWebPhotos()
            {
                return m_wf;
            }
    };


    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/

    class FERRISEXP_DLLLOCAL WebPhotosUploadFileContext;
    FERRIS_CTX_SMARTPTR( WebPhotosUploadFileContext, fh_WebPhotosUploadFileContext );
    class FERRISEXP_DLLLOCAL WebPhotosUploadFileContext
        :
        public StateLessEAHolder< WebPhotosUploadFileContext, leafContext >
    {
        typedef StateLessEAHolder< WebPhotosUploadFileContext, leafContext > _Base;
        typedef WebPhotosUploadFileContext _Self;

        string m_id;

        virtual void priv_FillCreateSubContextSchemaParts( CreateSubContextSchemaPart_t& m )
            {
                LG_WEBPHOTO_D << "upload file. setting file creation schema" << endl;

                m["ea"] = SubContextCreator(
                    SL_SubCreate_ea,
                    "	<elementType name=\"ea\">\n"
                    "		<elementType name=\"name\" default=\"new ea\">\n"
                    "			<dataTypeRef name=\"string\"/>\n"
                    "		</elementType>\n"
                    "		<elementType name=\"value\" default=\"\">\n"
                    "			<dataTypeRef name=\"string\"/>\n"
                    "		</elementType>\n"
                    "	</elementType>\n");
            }


        fh_iostream
        getNullEAStream( Context*, const std::string&, EA_Atom* attr )
            {
                fh_stringstream ss;
                return ss;
            }
        void
        NullEAStreamClosed( Context* c, const std::string& rdn, EA_Atom* atom, fh_istream ss )
            {
            }
        

        void addNullAttr( const std::string& rdn )
            {
                addAttribute( rdn,
                              this, &_Self::getNullEAStream,
                              this, &_Self::getNullEAStream,
                              this, &_Self::NullEAStreamClosed,
                              XSD_BASIC_STRING );
            }
        

        
        fh_context
        SubCreate_ea( fh_context c, fh_context md )
            {
                LG_WEBPHOTO_D << "SubCreate_ea(1)" << endl;
                string rdn = getStrSubCtx( md, "name", "", true, true );
                addNullAttr( rdn );
                return c;
            }
        
        
    public:

        fh_webPhotos getWebPhotos();
        
        WebPhotosUploadFileContext( Context* parent, const std::string& rdn )
            :
            _Base( parent, rdn ),
            m_id("")
            {
                createStateLessAttributes();

                addNullAttr( "user-owner-number" );
                addNullAttr( "group-owner-number" );
                addNullAttr( "ferris-type" );
                addNullAttr( "dontfollow-selinux-context" );
                addNullAttr( "mode" );
                addNullAttr( "mtime" );
                addNullAttr( "atime" );
                addNullAttr( "ctime" );
            }
        
        virtual ~WebPhotosUploadFileContext()
            {}

        void createStateLessAttributes( bool force = false )
            {
                if( force || isStateLessEAVirgin() )
                {
                    _Base::createStateLessAttributes( true );

                    
                    supplementStateLessAttributes( true );
                }
            }

        fh_stringstream
        real_getIOStream( ferris_ios::openmode m )
            throw (FerrisParentNotSetError,
                   AttributeNotWritable,
                   exception)
            {
                fh_stringstream ss;
                LG_WEBPHOTO_D << "returning an empty stream" << endl;
                return ss;
            }

        virtual void priv_postCopyAction( fh_context c )
            {
                LG_WEBPHOTO_D << "postCopyAction() c:" << c->getURL()
                              << " m_id:" << m_id
                              << endl;
                if( m_id.empty() )
                {
                    LG_WEBPHOTO_W << "postCopyAction() c:" << c->getURL()
                                  << " no web photo ID! not tagging remote file."
                                  << endl;
                    return;
                }

                fh_webPhotos wf = getWebPhotos();
                // Save off the photo-id on the source image for later use
                try
                {
                    string sn = wf->getImplementationShortName();
                    setStrAttr( c, sn + "-photo-id", m_id, true, true );
                    setStrAttr( c, "webphoto-service", sn, true, true );
                    setStrAttr( c, "webphoto-photo-id", m_id, true, true );
                }
                catch( exception& e )
                {
                    LG_WEBPHOTO_D << "Error linking source with photo id:" << e.what() << endl;
                }
                
                
                fh_webPhotoUpload wu = new WebPhotoUpload( wf );
                string tags = wu->getTagString( c );

                string id = m_id;
                
                stringmap_t args;
                args["method"]   = "flickr.photos.addTags";
                args["tags"]     = tags;
                args["photo_id"] = id;

                LG_WEBPHOTO_D << "postCopyAction() photo_id:" << id << endl;
                
                try
                {
                    wu->post( args );
                }
                catch( exception& e )
                {
                    stringstream ss;
                    ss << "Error setting tags for image:" << c->getURL() << endl
                       << " reason:" << e.what() << endl;
                    cerr << ss.str() << endl;
                    LG_WEBPHOTO_W << ss.str() << endl;
                }
            }
        
        void
        OnStreamClosed( fh_istream& ss, std::streamsize tellp, ferris_ios::openmode m )
            {
                if( !(m & std::ios::out) )
                    return;
                
                AdjustForOpenMode_Closing( ss, m, tellp );
                const string s = StreamToString(ss);

                fh_webPhotos wf = getWebPhotos();
                fh_webPhotoUpload wu = new WebPhotoUpload( wf );

                bool Verbose = true;
                bool DebugVerbose = true;
                if( Verbose )
                    wu->setVerboseStream( Factory::fcout() );
                wu->setDebugVerbose( DebugVerbose );

                string extension = "jpg";
                int dotpos = getDirName().rfind('.');
                if( dotpos != string::npos )
                    extension = getDirName().substr( dotpos );
                string tmp_url = (string)"/tmp/libferris-webupload-fs-temp-image." + extension;
                LG_WEBPHOTO_D << "tmp_url:" << tmp_url << endl;
                fh_context c = Shell::acquireContext( tmp_url, 600, false );
                {
                    fh_iostream oss = c->getIOStream( ios::out | ios::trunc );
                    oss << s << flush;
                }

                wu->setPublicViewable( wf->isDefaultImageProtectionPublic() );
                wu->setFriendViewable( wf->isDefaultImageProtectionFriend() );
                wu->setFamilyViewable( wf->isDefaultImageProtectionFamily() );
                
                
                LG_WEBPHOTO_D << "OnStreamClosed() c:" << c->getURL() << endl;
                wu->upload( c, getDirName() );
                LG_WEBPHOTO_D << "OnStreamClosed() s.sz:" << s.size() << endl;
                
                {
                    LG_WEBPHOTO_D << "OnStreamClosed() upload-list.sz:"
                                  << wu->getUploadedPhotoIDList().size() << endl;
                    stringlist_t& sl = wu->getUploadedPhotoIDList();
                    if( !sl.empty() )
                    {
                        stringlist_t::iterator iter = sl.end();
                        --iter;
                        m_id = *iter;
                        LG_WEBPHOTO_D << "OnStreamClosed() setting m_id:" << m_id << endl;
                    }
                }
            }
        
        ferris_ios::openmode
        getSupportedOpenModes()
            {
                return ios::in | ios::out | ios::trunc | ios::binary;
            }


        fh_iostream
        priv_getIOStream( ferris_ios::openmode m )
            throw (FerrisParentNotSetError,
                   AttributeNotWritable,
                   CanNotGetStream,
                   std::exception)
            {
                fh_stringstream ret = real_getIOStream( m );
                ret->getCloseSig().connect( bind( SigC::slot(*this, &_Self::OnStreamClosed ), m )); 
                return ret;
            }
        fh_istream
        priv_getIStream( ferris_ios::openmode m )
            throw (FerrisParentNotSetError,
                   CanNotGetStream,
                   std::exception)
            {
                fh_stringstream ret = real_getIOStream( m );
                return ret;
            }
        
        
        
    };

    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/

    class FERRISEXP_DLLLOCAL WebPhotosUploadDirectoryContext;
    FERRIS_CTX_SMARTPTR( WebPhotosUploadDirectoryContext, fh_WebPhotosUploadDirectoryContext );
    class FERRISEXP_DLLLOCAL WebPhotosUploadDirectoryContext
        :
        public StateLessEAHolder< WebPhotosUploadDirectoryContext, FakeInternalContext >
    {
        typedef StateLessEAHolder< WebPhotosUploadDirectoryContext, FakeInternalContext > _Base;

        virtual void priv_FillCreateSubContextSchemaParts( CreateSubContextSchemaPart_t& m )
            {
                LG_WEBPHOTO_D << "upload dir. setting file creation schema" << endl;
                m["file"] = SubContextCreator(SL_SubCreate_file,
                                              "	<elementType name=\"file\">\n"
                                              "		<elementType name=\"name\" default=\"new file\">\n"
                                              "			<dataTypeRef name=\"string\"/>\n"
                                              "		</elementType>\n"
                                              "	</elementType>\n");
            }
        
    protected:

    void priv_read()
    {
        staticDirContentsRAII _raii1( this );

        if( empty() )
        {
            // We have to setup the children the first time..
            fh_context child = 0;
        }
    }
        
    public:

        fh_webPhotos getWebPhotos()
            {
                WebPhotosObjectTopLevelContext* p = 0;
                p = getFirstParentOfContextClass( p );
                return p->getWebPhotos();
            }
        
        WebPhotosUploadDirectoryContext( Context* parent, const std::string& rdn )
            :
            _Base( parent, rdn )
            {
                createStateLessAttributes();
            }
        
        virtual ~WebPhotosUploadDirectoryContext()
            {}

        bool isDir()
            {
                return true;
            }
        
        

        void createStateLessAttributes( bool force = false )
            {
                if( force || isStateLessEAVirgin() )
                {
                    _Base::createStateLessAttributes( true );
                    supplementStateLessAttributes( true );
                }
            }

        
        virtual fh_context SubCreate_file( fh_context c, fh_context md )
            {
                string rdn = getStrSubCtx( md, "name", "" );
                LG_WEBPHOTO_D << "create_file for rdn:" << rdn << endl;
                
                fh_context child = 0;
                child = new WebPhotosUploadFileContext( this, rdn );
                Insert( GetImpl(child), false, true );

                
                LG_WEBPHOTO_D << "create_file OK for rdn:" << rdn << endl;
                return child;
            }
        
        
    };


    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/


    
    class FERRISEXP_DLLLOCAL WebPhotosFromPhotoMetadataContext;
    FERRIS_CTX_SMARTPTR( WebPhotosFromPhotoMetadataContext, fh_WebPhotosFromPhotoMetadataContext );
    class FERRISEXP_DLLLOCAL WebPhotosFromPhotoMetadataContext
        :
        public StateLessEAHolder< WebPhotosFromPhotoMetadataContext, leafContext >
    {
        typedef WebPhotosFromPhotoMetadataContext _Self;
        typedef StateLessEAHolder< WebPhotosFromPhotoMetadataContext, leafContext > _Base;

    protected:

        fh_PhotoMetadata m_md;


        ferris_ios::openmode
        getSupportedOpenModes()
            {
                return ios::in | ios::binary;
            }


        
        fh_istream
        priv_getIStream( ferris_ios::openmode m )
            throw (FerrisParentNotSetError,
                   CanNotGetStream,
                   std::exception)
            {
                LG_WEBPHOTO_D << "getting remote image as large as possible..." << endl;

                string id = tostr(m_md->getID());
                fh_webPhotos wf = getWebPhotos();
                
                stringmap_t args;
                args["api_key"] = wf->getAPIKey();
                args["photo_id"] = id;
                fh_context res = wf->signedCall( "flickr.photos.getSizes", args );
                fh_context c = res->getSubContext( "sizes" );
                string earl = "";
                bool getOriginal = false;
                Context::iterator ci = c->begin();
                Context::iterator ce = c->end();
                string currentsz = "Thumbnail";
                for( ; ci!=ce; ++ci )
                {
                    string sz = getStrAttr( *ci, "label", "" );
                    string e = getStrAttr( *ci, "source", "" );

                    if( sz == "Original" )
                    {
                        earl = e;
                        break;
                    }
                    if( sz == "Large" )
                    {
                        earl = e;
                        currentsz = sz;
                    }
                    if( sz == "Medium" )
                    {
                        if( currentsz != "Large" )
                        {
                            earl = e;
                            currentsz  = sz;
                        }
                    }
                    if( earl.empty() )
                    {
                        earl = e;
                        currentsz  = sz;
                    }
                }
                
                LG_WEBPHOTO_D << "original image URL:" << earl << endl;

                fh_context ret = Resolve( earl );
                return ret->getIStream( m );
            }
        
    public:

        fh_webPhotos getWebPhotos()
            {
                WebPhotosObjectTopLevelContext* p = 0;
                p = getFirstParentOfContextClass( p );
                return p->getWebPhotos();
            }
        
        WebPhotosFromPhotoMetadataContext( Context* parent, const std::string& rdn, fh_PhotoMetadata md )
            :
            _Base( parent, rdn ),
            m_md( md )
            {
                createStateLessAttributes();
                LG_WEBPHOTO_D << "WebPhotosFromPhotoMetadataContext() id:"
                              << md->getID() << endl;
                
            }
        
        virtual ~WebPhotosFromPhotoMetadataContext()
            {}

        bool isDir()
            {
                return false;
            }

        static fh_stringstream SL_getID( WebPhotosFromPhotoMetadataContext* c, const std::string&, EA_Atom*)
            {
                fh_stringstream ss;
                ss << c->m_md->getID();
                LG_WEBPHOTO_D << "SL_getID() ret:" << tostr(ss) << endl;
                return ss;
            }

        static fh_stringstream SL_getSecret( WebPhotosFromPhotoMetadataContext* c, const std::string&, EA_Atom*)
            {
                fh_stringstream ss;
                ss << c->m_md->getSecret();
                return ss;
            }

        static fh_stringstream SL_getTitle( WebPhotosFromPhotoMetadataContext* c, const std::string&, EA_Atom*)
            {
                fh_stringstream ss;
                ss << c->m_md->getTitle();
                LG_WEBPHOTO_D << "SL_getTitle() ret:" << tostr(ss) << endl;
                return ss;
            }

        static fh_stringstream SL_getServer( WebPhotosFromPhotoMetadataContext* c, const std::string&, EA_Atom*)
            {
                fh_stringstream ss;
                ss << c->m_md->getServer();
                return ss;
            }

        static fh_stringstream SL_isPrimary( WebPhotosFromPhotoMetadataContext* c, const std::string&, EA_Atom*)
            {
                fh_stringstream ss;
                ss << c->m_md->isPrimary();
                return ss;
            }
        
        void createStateLessAttributes( bool force = false )
            {
                if( force || isStateLessEAVirgin() )
                {
                    _Base::createStateLessAttributes( true );

#define SLEA  tryAddStateLessAttribute         

                    SLEA( "webphoto-id", &_Self::SL_getID, XSD_BASIC_INT );
                    SLEA( "secret", &_Self::SL_getSecret, XSD_BASIC_STRING );
                    SLEA( "title",  &_Self::SL_getTitle, XSD_BASIC_STRING );
                    SLEA( "server", &_Self::SL_getServer, XSD_BASIC_STRING );
                    SLEA( "is-primary", &_Self::SL_isPrimary, XSD_BASIC_BOOL );

#undef SLEA                    
                    supplementStateLessAttributes( true );
                }
            }

        virtual std::string getRecommendedEA()
            {
                return _Base::getRecommendedEA() + ",webphoto-id,title,";
            }
        
        
        
        
    };


    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/

    class FERRISEXP_DLLLOCAL WebPhotosRecentDirectoryContext;
    FERRIS_CTX_SMARTPTR( WebPhotosRecentDirectoryContext, fh_WebPhotosRecentDirectoryContext );
    class FERRISEXP_DLLLOCAL WebPhotosRecentDirectoryContext
        :
        public StateLessEAHolder< WebPhotosRecentDirectoryContext, FakeInternalContext >
    {
        typedef WebPhotosRecentDirectoryContext _Self;
        typedef StateLessEAHolder< WebPhotosRecentDirectoryContext, FakeInternalContext > _Base;

    protected:

        void priv_read()
            {
                staticDirContentsRAII _raii1( this );

                if( empty() )
                {
                    fh_webPhotos wf = getWebPhotos();
                    LG_WEBPHOTO_D << "recent/priv_read()" << endl;
                    photolist_t pl = wf->getMyRecent();
                    for( rs<photolist_t> pi( pl ); pi; ++pi )
                    {
                        string rdn = tostr((*pi)->getID());
                        fh_context child = new WebPhotosFromPhotoMetadataContext( this, rdn, *pi );
                        addNewChild( child );
                    }
                }
            }
        
    public:

        fh_webPhotos getWebPhotos()
            {
                WebPhotosObjectTopLevelContext* p = 0;
                p = getFirstParentOfContextClass( p );
                return p->getWebPhotos();
            }


        WebPhotosRecentDirectoryContext( Context* parent, const std::string& rdn )
            :
            _Base( parent, rdn )
            {
                createStateLessAttributes();
            }
        
        virtual ~WebPhotosRecentDirectoryContext()
            {}

        bool isDir()
            {
                return true;
            }
        
        
        void createStateLessAttributes( bool force = false )
            {
                if( force || isStateLessEAVirgin() )
                {
                    _Base::createStateLessAttributes( true );

#define SLEA  tryAddStateLessAttribute         

                    SLEA( "webphoto-id", &_Self::SL_getNothingStream, XSD_BASIC_INT );
                    SLEA( "secret", &_Self::SL_getNothingStream, XSD_BASIC_STRING );
                    SLEA( "title",  &_Self::SL_getNothingStream, XSD_BASIC_STRING );
                    SLEA( "server", &_Self::SL_getNothingStream, XSD_BASIC_STRING );
                    SLEA( "is-primary", &_Self::SL_getNothingStream, XSD_BASIC_BOOL );

#undef SLEA                    
                    
                    supplementStateLessAttributes( true );
                }
            }

        virtual std::string getRecommendedEA()
            {
                return _Base::getRecommendedEA() + ",webphoto-id,title,";
            }
        

    };

    

    /************************************************************/
    /************************************************************/
    /************************************************************/
    /************************************************************/
    /************************************************************/
    /************************************************************/
    

    /************************************************************/
    /************************************************************/
    /************************************************************/

    void
    WebPhotosObjectTopLevelContext::priv_read()
    {
        staticDirContentsRAII _raii1( this );

        if( empty() )
        {
            // We have to setup the children the first time..
            fh_context child = 0;
            
            child = new WebPhotosUploadDirectoryContext( this, "upload" );
            addNewChild( child );


            child = new WebPhotosRecentDirectoryContext( this, "recent" );
            addNewChild( child );
        }
    }


    /************************************************************/
    /************************************************************/
    /************************************************************/
    

    void
    WebPhotosRootContext::priv_read()
    {
        staticDirContentsRAII _raii1( this );

        if( empty() )
        {
            // We have to setup the children the first time..
            fh_context child = 0;
            
            {
                fh_fcontext hc = new FakeInternalContext( this, "flickr" );
                Insert( GetImpl(hc), false );

                fh_webPhotos wf = Factory::getDefaultFlickrWebPhotos();
                child = new WebPhotosObjectTopLevelContext( GetImpl(hc),
                                                            wf->getUserName(),
                                                            wf );
                hc->addNewChild( child );
            }

            {
                fh_fcontext hc = new FakeInternalContext( this, "23hq" );
                Insert( GetImpl(hc), false );

                fh_webPhotos wf = Factory::getDefault23hqWebPhotos();
                child = new WebPhotosObjectTopLevelContext( GetImpl(hc),
                                                            wf->getUserName(),
                                                            wf );
                hc->addNewChild( child );
            }
            
        }
    }

    /************************************************************/
    /************************************************************/
    /************************************************************/

    fh_webPhotos
    WebPhotosUploadFileContext::getWebPhotos()
            {
                WebPhotosUploadDirectoryContext* p = 0;
                p = getFirstParentOfContextClass( p );
                return p->getWebPhotos();
            }
    
    /************************************************************/
    /************************************************************/
    /************************************************************/
    
    class FERRISEXP_DLLLOCAL WebPhotosRootContext_RootContextDropper
        :
        public RootContextDropper
    {
    public:
        WebPhotosRootContext_RootContextDropper()
            {
                ImplementationDetail::appendToStaticLinkedRootContextNames("webphotos");
                RootContextFactory::Register( "webphotos", this );
            }

        fh_context Brew( RootContextFactory* rf )
            throw( RootContextCreationFailed )
            {
                static fh_context c = 0;
                if( !isBound(c) )
                {
                    c = new WebPhotosRootContext();
                }
                return c;
            }
    };
    static WebPhotosRootContext_RootContextDropper ___WebPhotosRootContext_static_init;
};
