/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: WebPhotos_private.hh,v 1.5 2007/04/14 21:30:08 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/
#ifndef _ALREADY_INCLUDED_FERRIS_WEBPHOTOS_H_
#define _ALREADY_INCLUDED_FERRIS_WEBPHOTOS_H_

#include "config.h"
#include <Ferris/HiddenSymbolSupport.hh>
#include <Ferris/TypeDecl.hh>
#include <FerrisStreams/Streams.hh>

#ifdef HAVE_WEBPHOTOS_API

#include <Ferris/FerrisBoost.hh>

namespace Ferris
{
    class WebPhotos;
    class WebPhotoUpload;

    class PhotoMetadata
        :
        public Handlable
    {
//         <photo id="2484" secret="123456" server="1"
//      		title="my photo" isprimary="0" /> 
        friend class WebPhotos;

        long m_id;
        std::string m_secret;
        std::string m_server;
        std::string m_title;
        bool m_isPrimary;

    protected:

        PhotoMetadata( fh_context md );

    public:
        
        ~PhotoMetadata();

        long        getID();
        std::string getSecret();
        std::string getServer();
        std::string getTitle();
        bool        isPrimary();
    };
    FERRIS_SMARTPTR( PhotoMetadata, fh_PhotoMetadata );
    typedef std::list< fh_PhotoMetadata > photolist_t;
    
    class WebPhotos
        :
        public Handlable
    {
        friend class WebPhotoUpload;

        std::string m_username;
        fh_rex m_includeEAIsPresentRegex;
        fh_rex m_includeEAandValueRegex;
        
    protected:

        virtual std::string getRESTURL();

    public:

        virtual std::string getUploadURL();
        virtual std::string getAuthURL();
        virtual std::string getPostUploadURLPrefix();
        virtual std::string getAPIKey();
        virtual std::string getAPISecret();
        
    public:

        WebPhotos( const std::string& username );
        virtual ~WebPhotos();

        virtual std::string getImplementationShortName();
        
        long getDefaultLargestDimension();
        void setDefaultLargestDimension( long v );

        fh_rex getDefaultIncludeEAIsPresentRegex();
        std::string getDefaultIncludeEAIsPresentRegexString();
        void setDefaultIncludeEAIsPresentRegex( const std::string& s );
        fh_rex getDefaultIncludeEAandValueRegex();
        std::string getDefaultIncludeEAandValueRegexString();
        void setDefaultIncludeEAandValueRegex( const std::string& s );

        bool isDefaultImageProtectionPublic();
        void setDefaultImageProtectionPublic( bool v );
        bool isDefaultImageProtectionFriend();
        void setDefaultImageProtectionFriend( bool v );
        bool isDefaultImageProtectionFamily();
        void setDefaultImageProtectionFamily( bool v );
        
        
        bool haveAPIKey();
        std::string getToken();
        bool haveToken();
        std::string getDefaultUsername();
        std::string getUserName();
        std::string getFullName();

        fh_context handleCallIO( stringmap_t args );
        std::string makeSignedString( stringmap_t args );
        fh_context signedCall( std::string methodname,
                               stringmap_t args,
                               bool dontAutoIncludeToken = false );

        photolist_t getMyRecent( time_t tt = 0, int pagenum = 1, int limit = 500 );
        
        ///////////
        // These are only needed during / just after an auth
        ///////////
        void setToken( const std::string& token );
        void setUserName( const std::string& s );
        void setFullName( const std::string& s );
        void setDefaultUsername( const std::string& s );
    };
    FERRIS_SMARTPTR( WebPhotos, fh_webPhotos );


    class WebPhotoUpload
        :
        public Handlable
    {
        fh_webPhotos m_wf;
        stringlist_t m_uploadedPhotoIDList;
        fh_ostream m_verboseStream;
        bool m_debugVerbose;
        fh_rex m_includeEAisPresentRegex;
        fh_rex m_includeEAandValueRegex;
        bool m_publicViewable;
        bool m_FriendViewable;
        bool m_FamilyViewable;
        std::string m_DescriptionFromEA;
        long m_largestDimensionExplicit;


        fh_context makeScaledImage( fh_context c );

        void throwIfStatusIsBad( fh_context c, const std::string& res );
        
    public:
    
        WebPhotoUpload( fh_webPhotos wf );
        ~WebPhotoUpload();

        void setVerboseStream( fh_ostream m_verboseStream );
        void setDebugVerbose( bool v );
        void setIncludeEAisPresentRegex( fh_rex r );
        void setIncludeEAandValueRegex( fh_rex r );
        void setPublicViewable( bool v );
        void setFriendViewable( bool v );
        void setFamilyViewable( bool v );
        void setDescriptionFromEA( const std::string& s );
        void setLargestDimensionExplicit( long v );
    
        std::string getTagString( fh_context c );

        void upload( fh_context c,
                     const std::string& title = "",
                     const std::string& desc = "" );
        void post( stringmap_t args );


        
        std::string getPostUploadURL();
        stringlist_t& getUploadedPhotoIDList();
    };
    FERRIS_SMARTPTR( WebPhotoUpload, fh_webPhotoUpload );
    

    namespace Factory
    {
        fh_webPhotos getDefaultFlickrWebPhotos();
        fh_webPhotos getDefault23hqWebPhotos();
        fh_webPhotos getDefaultZooomrWebPhotos();

        fh_webPhotos getDefaultWebPhotosForShortName( const std::string& n );
    };
    
    
    



    
    
};
#endif
#endif
