/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2007 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: XQilla.cpp,v 1.8 2007/08/21 21:30:03 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/


#include <Ferris/XQilla_private.hh>
#include <Ferris/FerrisDOM.hh>
#include <Ferris/Hashing.hh>

#include <xqilla/context/DynamicContext.hpp>
#include <xqilla/utils/XPath2Utils.hpp>

#include <xqilla/schema/DocumentCache.hpp>
#include <xqilla/utils/XPath2Utils.hpp>
#include <xqilla/schema/DocumentCacheImpl.hpp>
#include <xqilla/functions/FunctionConstructor.hpp>
#include <xqilla/axis/NodeTest.hpp>

#include <xqilla/functions/FuncFactory.hpp>

#include <xqilla/exceptions/FunctionException.hpp>
#include <xqilla/exceptions/XMLParseException.hpp>


#include <xqilla/context/impl/XQDynamicContextImpl.hpp>

using namespace XERCES_CPP_NAMESPACE;
using namespace std;

namespace Ferris
{
    namespace qilla
    {
        void copyChildren( const DynamicContext* dc,
                           DOMDocument* document, fh_context c, DOMElement* target );
        
        std::string
        canonicalElementName( std::string rdn )
        {
            LG_XQILLA_D << "canonicalElementName(top) rdn:" << rdn << endl;
            
            string nn = rdn;
            if( starts_with( nn, "." ))
                nn = (string)"_" + nn;
            nn = Util::replace_all( nn, " ", "_space_" );
            nn = Util::replace_all( nn, "+", "_plus_" );
            nn = Util::replace_all( nn, "(", "_oper_" );
            nn = Util::replace_all( nn, ")", "_cper_" );
            nn = Util::replace_all( nn, ",", "_comma_" );
            nn = Util::replace_all( nn, "-", "_dash_" );
            nn = Util::replace_all( nn, ":", "_colon_" );

            static const boost::regex r = toregex( "^[0-9]" );
            if( regex_search( nn, r, boost::match_any ) )
            {
                nn = "number_" + nn;
            }
            

            
            LG_XQILLA_D << "canonicalElementName(bottom) rdn:" << nn << endl;
            return nn;
        }
        std::string
        canonicalElementName( fh_context c )
        {
            return canonicalElementName( c->getDirName() );
        }


        /////////////////////////////////////////
        /////////////////////////////////////////
        /////////////////////////////////////////
        /////////////////////////////////////////
        
            
        ShouldIncludeAttribute::ShouldIncludeAttribute()
            :
            m_showColumnsRegexString("")
        {
        }
            
        bool
        ShouldIncludeAttribute::nameContainsIllegalXMLAttributeChar( const std::string& eaname )
        {
            LG_XQILLA_D << "nameContainsIllegalXMLAttributeChar() eaname:" << eaname << endl;
            static const boost::regex r = toregex( "[-A-Za-z0-9_]+" );
            return( !regex_match( eaname, r, boost::match_any ) );
        }
        

            
        void
        ShouldIncludeAttribute::setShowColumns( const std::string& s )
        {
            Util::parseCommaSeperatedList( s, m_showColumnsList );
        }
            
        void
        ShouldIncludeAttribute::setShowColumnsRegex( const std::string& s )
        {
            m_showColumnsRegexString = s;
            m_showColumnsRegex       = toregex( s );
        }
            
        bool
        ShouldIncludeAttribute::shouldInclude( fh_context c, const std::string& eaname )
        {
            
            if( m_showColumnsList.count( eaname ) )
            {
                if( nameContainsIllegalXMLAttributeChar( eaname ) )
                    return false;
                return true;
            }

            if( !m_showColumnsRegexString.empty()
                && regex_search( eaname, m_showColumnsRegex, boost::match_any ))
            {
                if( nameContainsIllegalXMLAttributeChar( eaname ) )
                    return false;
                return true;
            }

            return false;
        }
            

        typedef map< const DynamicContext*, fh_shouldIncludeAttribute > getShouldIncludeAttribute_cache_t;
        getShouldIncludeAttribute_cache_t& getShouldIncludeAttribute_cache()
        {
            static getShouldIncludeAttribute_cache_t ret;
            return ret;
        }
        
        
        fh_shouldIncludeAttribute getShouldIncludeAttribute( const DynamicContext* context )
        {
            getShouldIncludeAttribute_cache_t::iterator iter = getShouldIncludeAttribute_cache().find( context );
            if( iter != getShouldIncludeAttribute_cache().end() )
            {
                return iter->second;
            }
            fh_shouldIncludeAttribute n = new ShouldIncludeAttribute();
            n->setShowColumnsRegex(".*");
            getShouldIncludeAttribute_cache().insert( make_pair( context, n ) );
            return getShouldIncludeAttribute_cache()[ context ];
        }
        
        void setShouldIncludeAttribute( DynamicContext* context, fh_shouldIncludeAttribute sia )
        {
            getShouldIncludeAttribute_cache()[ context ] = sia;
        }
        
        

        ////////////////////////////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////////////////////////////
        ////////////////////////////////////////////////////////////////////////////////
        
        class FQilla_ResultImpl : public ResultImpl
        {
        protected:
            fh_context m_ctx;
            const NodeTest* m_nodeTest;
            string m_NodeNameWildcard;
            string m_NodeNamespaceURIWildcard;
            SequenceType::ItemType* m_itemType;
            boost::regex m_NodeNameWildcardRegex;
            
        
            bool haveItemType()
                {
                    return m_itemType;
                }
        
            Item::Ptr handleNodeTestForItemType( DynamicContext *dynamicContext, fh_context ctx, const std::string& eaname )
                {
                    SequenceType::ItemType *itemType = m_nodeTest->getItemType();
                    if(itemType != 0)
                    {
//                    Node::Ptr result = factory_.createNode(node, dynamicContext);
                        Node::Ptr result = new FQilla_Node( dynamicContext, ctx, eaname );
                        if(itemType->matches(result, dynamicContext, this)) {
                            return result;
                        }
                    }
                    return 0;
                }

            bool passesElementTests( fh_context ctx )
                {
                    LG_XQILLA_D << "passesElementTests(1) ctx:" << ctx->getURL() << endl;
                    
                    if(!m_nodeTest->getTypeWildcard()
                       && m_nodeTest->getNodeType() != Node::element_string)
                    {
                        return false;
                    }
                    
                    LG_XQILLA_D << "passesElementTests(2) ctx:" << ctx->getURL() << endl;
                    std::string localName = canonicalElementName( ctx );
                    if( !m_NodeNameWildcard.empty() &&
                        !regex_match( localName, m_NodeNameWildcardRegex, boost::match_any ))
                    {
                        LG_XQILLA_D << "passesElementTests(2-bad) ctx:" << ctx->getURL()
                                    << " localName:" << localName
                                    << " m_NodeNameWildcard:" << m_NodeNameWildcard
                                    << endl;
                        return false;
                    }

//                     std::string localName = canonicalElementName( ctx );
//                     if( !m_NodeNameWildcard.empty() && m_NodeNameWildcard != localName )
//                     {
//                         return false;
//                     }
                    
                    LG_XQILLA_D << "passesElementTests(3) ctx:" << ctx->getURL() << endl;
                    if( !m_NodeNamespaceURIWildcard.empty() && m_NodeNamespaceURIWildcard != "FIXME" )
                    {
                        return false;
                    }

                    LG_XQILLA_D << "passesElementTests(ok) ctx:" << ctx->getURL() << endl;
                    return true;
                }
        
        
        public:
            FQilla_ResultImpl( const LocationInfo *o,
                               const NodeTest* nodeTest,
                               fh_context ctx )
                :
                ResultImpl( o ),
                m_nodeTest( nodeTest ),
                m_ctx( ctx ),
                m_itemType( 0 ),
                m_NodeNameWildcard(""),
                m_NodeNamespaceURIWildcard("")
                {
                    if( m_nodeTest )
                    {
                        m_itemType = m_nodeTest->getItemType();

                        if(!m_nodeTest->getNameWildcard())
                        {
                            m_NodeNameWildcard = tostr(m_nodeTest->getNodeName());
                            m_NodeNameWildcardRegex = toregex( m_NodeNameWildcard + "(--[0-9]+)?" );
                        }
                        
                        if(!m_nodeTest->getNamespaceWildcard())
                            m_NodeNamespaceURIWildcard = tostr(m_nodeTest->getNodeUri());
                    }
                }
            virtual ~FQilla_ResultImpl()
                {
                }
        
//        virtual Item::Ptr next(DynamicContext *context);
            virtual std::string asString(DynamicContext *context, int indent) const
                {
                    LG_XQILLA_W << "asString() called but not implemented!" << endl;
                }
        };


        class FQilla_Attr_ResultImpl : public FQilla_ResultImpl
        {
            AttributeCollection::AttributeNames_t an;
            AttributeCollection::AttributeNames_t::iterator e;
            AttributeCollection::AttributeNames_t::iterator iter;
        public:
            FQilla_Attr_ResultImpl( const LocationInfo *o,
                                    const NodeTest* nodeTest,
                                    fh_context ctx )
                :
                FQilla_ResultImpl( o, nodeTest, ctx )
                {
                    ctx->getAttributeNames( an );
                    e = an.end();
                    iter = an.begin();
                }
            virtual ~FQilla_Attr_ResultImpl()
                {
                }
            virtual Item::Ptr next( DynamicContext* dynamicContext )
                {
                    LG_XQILLA_D << "FQilla_Attr_ResultImpl::next()" << endl;
                    while( iter != e )
                    {
                        string eaname = *iter;
                        ++iter;
                        LG_XQILLA_D << "FQilla_Attr_ResultImpl::next() ea:" << eaname << endl;

                        if( haveItemType() )
                        {
                            LG_XQILLA_D << "FQilla_Attr_ResultImpl::next() haveItemType..." << endl;
                            return handleNodeTestForItemType( dynamicContext, m_ctx, eaname );
                        }
                    

                        if(m_nodeTest->getTypeWildcard())
                        {
                            if(m_nodeTest->getHasChildren())
                                continue;
                        }
                        else
                        {
                            if(m_nodeTest->getNodeType() != Node::attribute_string)
                            {
                                continue;
                            }
                        }

                        string localName = eaname;
                        if( !m_NodeNameWildcard.empty() && m_NodeNameWildcard != localName )
                        {
                            LG_XQILLA_D << "FQilla_Attr_ResultImpl::next() ln-bad m_NodeNameWildcard:" << m_NodeNameWildcard << "  ea:" << eaname << endl;
//                        if( m_ctx->isAttributeBound( m_NodeNameWildcard ) )
                            {
                                // FIXME:
                                // skip ahead to the attribute, need to update iter to be the correct attr as well
                            }
                            continue;
                        }
                        LG_XQILLA_D << "FQilla_Attr_ResultImpl::next() ln-ok m_NodeNameWildcard:" << m_NodeNameWildcard << "  ea:" << eaname << endl;
                        if( !m_NodeNamespaceURIWildcard.empty() && m_NodeNamespaceURIWildcard != "FIXME" )
                            continue;

                        LG_XQILLA_D << "FQilla_Attr_ResultImpl::next() MATCH  ea:" << eaname
                                    << " m_ctx:" << m_ctx->getURL()
                                    << endl;
//                    return new FQilla_Node( m_ctx, "" );
                        return new FQilla_Node( dynamicContext, m_ctx, eaname );
                    }
                    return 0;
                }
        };

        template< class IteratorType >
        class FQilla_Iterator_ResultImpl : public FQilla_ResultImpl
        {
            string eaname;
            IteratorType m_iter;
            IteratorType m_begin;
            IteratorType m_end;
            ctxlist_t* m_ctxlist;
            bool selectingSingleNodeWithWildCard;
        
        public:
            FQilla_Iterator_ResultImpl( const LocationInfo *o,
                                        const NodeTest *nodeTest,
                                        fh_context ctx,
                                        IteratorType begin,
                                        IteratorType end,
                                        ctxlist_t* cl = 0 )
                :
                FQilla_ResultImpl( o, nodeTest, ctx ),
                m_iter( begin ),
                m_begin( begin ),
                m_end( end ),
                m_ctxlist( cl ),
                selectingSingleNodeWithWildCard( false )
                {
//                 if( !m_NodeNameWildcard.empty() )
//                 {
//                     Context::iterator ci = ctx->find( m_NodeNameWildcard );
//                     if( ci != m_end )
//                     {
//                         m_iter = ci;
//                         selectingSingleNodeWithWildCard = true;
//                     }
//                 }
                }
            virtual ~FQilla_Iterator_ResultImpl()
                {
                    if( m_ctxlist )
                    {
                        m_ctxlist->clear();
                        delete m_ctxlist;
                    }
                }

        
            virtual Item::Ptr next(DynamicContext *dynamicContext)
                {
                    LG_XQILLA_D << "FQilla_Iterator_ResultImpl::next()" << endl;
                    while( m_iter != m_end )
                    {
                        fh_context ctx = *m_iter;
                        ++m_iter;
                        LG_XQILLA_D << "FQilla_Iterator_ResultImpl::next() ctx:" << ctx->getURL() << endl;

                        if( selectingSingleNodeWithWildCard )
                        {
                            LG_XQILLA_D << "FQilla_Iterator_ResultImpl::next() selecting single via wildcard..." << endl;
                            m_iter = m_end;
                        }
                    

                        if( haveItemType() )
                            return handleNodeTestForItemType( dynamicContext, ctx, "" );

                        LG_XQILLA_D << "FQilla_Iterator_ResultImpl::next() calling passesElementTests" << endl;
                        if( !passesElementTests( ctx ) )
                        {
                            continue;
                        }
                    
                    
//                     if(!m_nodeTest->getTypeWildcard()
//                        && m_nodeTest->getNodeType() != Node::element_string)
//                     {
//                         continue;
//                     }
                    
//                     std::string localName = canonicalElementName( ctx );
//                     if( !m_NodeNameWildcard.empty() && m_NodeNameWildcard != localName )
//                     {
//                         continue;
//                     }

//                     if( !m_NodeNamespaceURIWildcard.empty() && m_NodeNamespaceURIWildcard != "FIXME" )
//                     {
//                         continue;
//                     }

                        LG_XQILLA_D << "FQilla_Iterator_ResultImpl::next() ok ctx:" << ctx->getURL() << endl;
                        return new FQilla_Node( dynamicContext, ctx );
                    }
                    return 0;
                }
        
        };


        class FQilla_SingleNode_ResultImpl : public FQilla_ResultImpl
        {
            const FQilla_Node* m_node;
            const FQilla_Node* m_iter;

        public:
            FQilla_SingleNode_ResultImpl( const LocationInfo *o,
                                          const NodeTest *nodeTest,
                                          fh_context ctx,
                                          const Node* n )
                :
                FQilla_ResultImpl( o, nodeTest, ctx ),
                m_node( (FQilla_Node*)n ),
                m_iter( (FQilla_Node*)n )
                {
                }
        
            virtual ~FQilla_SingleNode_ResultImpl()
                {
                }
        
            virtual Item::Ptr next(DynamicContext* dynamicContext)
                {
                    LG_XQILLA_D << "FQilla_SingleNode_ResultImpl::next()" << endl;
                    if( !m_iter )
                        return 0;
                    
                    const FQilla_Node* ret = m_iter;
                    m_iter = 0;

                    fh_context ctx = ret->getFerrisContext();
                
                    if( haveItemType() )
                        return handleNodeTestForItemType( dynamicContext, ctx, "" );

                    if( !passesElementTests( ctx ) )
                        return 0;
                
                    return ret;
                }
        };
    

    
    
    
        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/
    
        struct fr_hash : public std::unary_function< pair< const DynamicContext*, Context* const >, size_t >
        {
            inline size_t operator()( const pair< const DynamicContext*, Context* const >& s ) const
                {
                    return GPOINTER_TO_INT( s.first ) | GPOINTER_TO_INT( s.second );
                }
        };
        struct fr_equal_to
            : public std::binary_function<
            pair< const DynamicContext*, Context* const >,
            pair< const DynamicContext*, Context* const >, bool> 
        {
            inline bool operator()(const pair< const DynamicContext*, Context* const >& __x,
                                   const pair< const DynamicContext*, Context* const >& __y) const
                {
                    return __x.first == __y.first && __x.second == __y.second;
                }
        };
    
        typedef FERRIS_STD_HASH_MAP<
            pair< const DynamicContext*, Context* const >,
            bool,
            fr_hash, fr_equal_to > m_fqilla_fakeroots_t;
        m_fqilla_fakeroots_t& get_fqilla_fakeroots()
        {
            static m_fqilla_fakeroots_t m_fqilla_fakeroots;
            return m_fqilla_fakeroots;
        }
    
        
    
        const XMLCh FQilla_Node::fQilla[] =   // Points to "fQilla"
        {
            XERCES_CPP_NAMESPACE_QUALIFIER chLatin_F,
            XERCES_CPP_NAMESPACE_QUALIFIER chLatin_Q,
            XERCES_CPP_NAMESPACE_QUALIFIER chLatin_i,
            XERCES_CPP_NAMESPACE_QUALIFIER chLatin_l,
            XERCES_CPP_NAMESPACE_QUALIFIER chLatin_l,
            XERCES_CPP_NAMESPACE_QUALIFIER chLatin_a,
            XERCES_CPP_NAMESPACE_QUALIFIER chNull
        };
    
        FQilla_Node::FQilla_Node( DynamicContext* dynamicContext,
                                  fh_context ctx,
                                  const std::string& eaname,
                                  bool forceNoParent )
            :
            m_ctx( ctx ),
            m_eaname( eaname ),
            m_fSerializer(0),
            m_forceNoParent( forceNoParent )
        {
            if( forceNoParent )
            {
                get_fqilla_fakeroots().insert(
                    make_pair( make_pair( dynamicContext, GetImpl(m_ctx) ), 1 )
                    );
            }
        }
        
    
        FQilla_Node::~FQilla_Node()
        {
            if(m_fSerializer != 0)
            {
                m_fSerializer->release();
            }
        }

        void *
        FQilla_Node::getInterface(const XMLCh *name) const
        {
            if(XPath2Utils::equals(name,FQilla_Node::fQilla))
            {
                return (void*)this;
            }
            else if(XPath2Utils::equals(name,Node::gXerces))
            {
                bool hideXMLAttribute = true;
                bool hideEmblemAttributes = true;
                bool hideSchemaAttributes = true;
            
                fh_domdoc dom = Factory::makeDOM( m_ctx,
                                                  hideXMLAttribute,
                                                  hideEmblemAttributes,
                                                  hideSchemaAttributes );

                cacheDOM( dom );
                return dom->getDocumentElement();
            }
            return 0;
        }

        void
        FQilla_Node::cacheDOM( fh_domdoc d ) const
        {
            dom_cache.push_back( d );
        }
    
    
    
        bool
        FQilla_Node::isNode() const
        {
            LG_XQILLA_D << "FQilla_Node::isNode()" << endl;
//        return m_eaname.empty();
            return true;
        }
    
        bool
        FQilla_Node::isAtomicValue() const
        {
            LG_XQILLA_D << "FQilla_Node::isAtomicValue()" << endl;
            return false;
        }

//#define X(str) XStr(str).unicodeForm()
    
        const XMLCh*
        FQilla_Node::asString(const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "FQilla_Node::asString()" << endl;
            LG_XQILLA_D << "FQilla_Node::asString()  ctx:" << m_ctx->getURL() << endl;
//        dmNodeName( dynamicContext );
            
            {
                
                fh_domdoc dom = Factory::makeDOM( "libferris" );
                DOMElement* parent = dom->getDocumentElement();
                copyChildren( dynamicContext, GetImpl(dom), m_ctx, parent );
                DOMNode* n = dom->getDocumentElement();
                XML::domnode_list_t nl;
                XML::getChildren( nl, (DOMElement*)n );
                if( nl.size() == 1 )
                {
                    n = nl.front();
                
                    fh_stringstream iss = tostream( *n, true );
                    static string s = StreamToString( iss );
//                    cout << s << endl;
                    return X(s.c_str());
                }

                return 0;
            }
            


            

//         stringstream ss;
//         ss << "<context url=\"" << m_ctx->getURL() << "\">" << endl;
//         ss << "</context>";
//         static string s;
//         s = ss.str();
//         return X(s.c_str());


        
            if(!m_fSerializer)
            {
                const static XMLCh ls_string[] =
                    {
                        XERCES_CPP_NAMESPACE_QUALIFIER chLatin_L,
                        XERCES_CPP_NAMESPACE_QUALIFIER chLatin_S,
                        XERCES_CPP_NAMESPACE_QUALIFIER chNull
                    };
            
                DOMImplementation *impl = DOMImplementationRegistry::getDOMImplementation(ls_string);
                m_fSerializer =
                    ((DOMImplementationLS*)impl)->createDOMWriter( dynamicContext->getMemoryManager());
                m_fSerializer->setFeature(XMLUni::fgDOMXMLDeclaration, false);
            }

            fh_domdoc dom = Factory::makeDOM( "root" );
            DOMElement* de = XML::createElement( dom, dom->getDocumentElement(), "context" );
            setAttribute( de, "url", m_ctx->getURL() );
        
        
//         bool hideXMLAttribute = true;
//         bool hideEmblemAttributes = true;
//         bool hideSchemaAttributes = true;
        
//         fh_domdoc dom = Factory::makeDOM( m_ctx,
//                                           hideXMLAttribute,
//                                           hideEmblemAttributes,
//                                           hideSchemaAttributes );

            DOMNode* n = dom->getDocumentElement();
            return m_fSerializer->writeToString( *n );
        }
    
        bool FQilla_Node::hasInstanceOfType(const XMLCh* typeURI, const XMLCh* typeName, const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "hasInstanceOfType() typeURI:" << tostr(typeURI)
                        << " typeName:" << tostr(typeName)
                        << " dc:" << toVoid(dynamicContext)
                        << endl;
            return false;
        }
    
        Sequence FQilla_Node::dmBaseURI(const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "dmBaseURI()"
                        << " dc:" << toVoid(dynamicContext)
                        << endl;
            return Sequence(dynamicContext->getMemoryManager());
        }
    
        const XMLCh* FQilla_Node::dmNodeKind() const
        {
            LG_XQILLA_D << "dmNodeKind()" << endl;
            if( isAttrNode() )
                return attribute_string;
            return element_string;
        }

//     std::string
//     FQilla_Node::canonicalElementName( fh_context c ) const
//     {
//         string nn = m_ctx->getDirName().c_str();
//         if( starts_with( nn, "." ))
//             nn = (string)"_" + nn;
//         nn = Util::replace_all( nn, "+", "_plus_" );
//         return nn;
//     }
    
    
        ATQNameOrDerived::Ptr
        FQilla_Node::dmNodeName(const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "dmNodeName()" << endl;
            LG_XQILLA_D << "dmNodeName() ctx:" << m_ctx->getURL() << endl;

            XMLCh* ns = 0;
            XMLCh* prefix = 0;
            string nn = canonicalElementName( m_ctx );
            XMLCh* localname = XMLString::transcode( nn.c_str() );

            LG_XQILLA_D << "dmNodeName() ret:" << nn << endl;
        
            return dynamicContext->getItemFactory()->createQName(
                ns, prefix, localname, dynamicContext );
        }
    
        const XMLCh* FQilla_Node::dmStringValue(const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "dmStringValue()" << endl;
            if( isAttrNode() )
            {
                string v = getStrAttr( m_ctx, m_eaname, "" );
                return dynamicContext->getMemoryManager()->getPooledString( X(v.c_str()) );
            }

            fh_istream iss = m_ctx->getIStream();
            string v = StreamToString( iss );
            return dynamicContext->getMemoryManager()->getPooledString( X(v.c_str()) );
        }
    
        Sequence FQilla_Node::dmTypedValue(DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "dmTypedValue()" << endl;

            if( isAttrNode() )
            {
                Item::Ptr item = 0;
                item = (const Item::Ptr)dynamicContext->getItemFactory()->createUntypedAtomic(
                    dmStringValue(dynamicContext), dynamicContext);
                return Sequence(item, dynamicContext->getMemoryManager()); 
            }

            Item::Ptr item = 0;
            item = (const Item::Ptr)dynamicContext->getItemFactory()->createUntypedAtomic(
                dmStringValue(dynamicContext), dynamicContext);
            return Sequence(item, dynamicContext->getMemoryManager()); 
        
//        return Sequence(dynamicContext->getMemoryManager());
        }
    
        Sequence FQilla_Node::dmDocumentURI(const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "dmDocumentURI()" << endl;
            return Sequence(dynamicContext->getMemoryManager());
        }
    
        ATQNameOrDerived::Ptr FQilla_Node::dmTypeName(const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "dmTypeName()" << endl;

            const XMLCh* uri = FunctionConstructor::XMLChXPath2DatatypesURI;
            const XMLCh* name = DocumentCacheParser::g_szUntyped;
        
            if( isAttrNode() )
            {
                uri=FunctionConstructor::XMLChXPath2DatatypesURI;
                name=ATUntypedAtomic::fgDT_UNTYPEDATOMIC;
            }

            return dynamicContext->getItemFactory()->
                createQName(uri, XMLUni::fgZeroLenString, name, dynamicContext);
        }
    
        ATBooleanOrDerived::Ptr FQilla_Node::dmNilled(const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "dmNilled()" << endl;

            if( isAttrNode() )
            {
                return 0;
            }
        
            return dynamicContext->getItemFactory()->createBoolean( false, dynamicContext );
        }
    
        bool FQilla_Node::lessThan(const Node::Ptr &other, const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "lessThan(top)" << endl;
            const FQilla_Node* otherImpl = (const FQilla_Node*)other->getInterface(FQilla_Node::fQilla);
        
            fh_context otherc = otherImpl->m_ctx;

            string earl  = m_ctx->getURL();
            string oearl = otherc->getURL();

            LG_XQILLA_D << "lessThan()  earl:" << earl << endl;
            LG_XQILLA_D << "lessThan() oearl:" << oearl << endl;
            LG_XQILLA_D << "lessThan() ret:" << (earl < oearl) << endl;
        
            return earl < oearl;
        }
    
        bool FQilla_Node::equals(const Node::Ptr &other) const
        {
            LG_XQILLA_D << "equals(top)" << endl;
            const FQilla_Node* otherImpl = (const FQilla_Node*)other->getInterface(FQilla_Node::fQilla);
            fh_context otherc = otherImpl->m_ctx;

            string earl  = m_ctx->getURL();
            string oearl = otherc->getURL();

            LG_XQILLA_D << "equals()  earl:" << earl << endl;
            LG_XQILLA_D << "equals() oearl:" << oearl << endl;
            LG_XQILLA_D << "equals() ret:" << (earl == oearl) << endl;
        
            return earl == oearl;
        }
    
        bool FQilla_Node::uniqueLessThan(const Node::Ptr &other, const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "uniqueLessThan()" << endl;
            return lessThan( other, dynamicContext );
        }
        Node::Ptr FQilla_Node::dmParent(const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "dmParent()" << endl;
            LG_XQILLA_D << "m_ctx:" << m_ctx->getURL() << endl;

            if( m_forceNoParent )
                return 0;

            if( m_ctx->isParentBound() )
            {
                return new FQilla_Node( (DynamicContext*)dynamicContext, m_ctx->getParent(), "" );
            }
            return 0;
        }
        Result FQilla_Node::dmAttributes(const DynamicContext* dynamicContext, const LocationInfo *info) const
        {
            LG_XQILLA_D << "dmAttributes()" << endl;
            if( m_eaname.empty() )
            {
                return new FQilla_Attr_ResultImpl( info, 0, m_ctx );
//            return new AttributeAxis(info, fNode, this, NULL);
            }
            return 0;
        }
        Result FQilla_Node::dmNamespaceNodes(const DynamicContext* dynamicContext, const LocationInfo *info) const
        {
            LG_XQILLA_D << "dmNamespaceNodes()" << endl;
            return 0;
        }
        Result FQilla_Node::dmChildren(const DynamicContext *context, const LocationInfo *info) const
        {
            LG_XQILLA_D << "dmChildren()" << endl;
            if( m_eaname.empty() )
            {
                return new FQilla_Iterator_ResultImpl<Context::iterator>( info, 0, m_ctx, m_ctx->begin(), m_ctx->end() );
//            return new ChildAxis(info, fNode, this, NULL);
            }
            return 0;
        }

        void
        FQilla_Node::getAllChildren( fh_context c, ctxlist_t* cl ) const
        {
            Context::iterator    e = c->end();
            Context::iterator iter = c->begin();
            for( ; iter != e; ++iter )
            {
                cl->push_back( *iter );
                getAllChildren( *iter, cl );
            }
        }
    

    
        Result FQilla_Node::getAxisResult(XQStep::Axis axis,
                                          const NodeTest *nodeTest,
                                          const DynamicContext* dynamicContext,
                                          const LocationInfo *info) const
        {
            LG_XQILLA_D << "getAxisResult() axis:" << axis << endl;
            LG_XQILLA_D << "getAxisResult() ctx:" << m_ctx->getURL() << endl;

            switch(axis)
            {
            case XQStep::ANCESTOR_OR_SELF:
            case XQStep::ANCESTOR:
            {
                LG_XQILLA_D << "XQStep::ANCESTOR or XQStep::ANCESTOR_OR_SELF" << endl;
            
                ctxlist_t* cl = new ctxlist_t;
                fh_context c = m_ctx;
                if( axis == XQStep::ANCESTOR_OR_SELF )
                {
                    cl->push_back( c );
                }
                while( c && c->isParentBound() )
                {
                    c = c->getParent();

                    cl->push_back( c );
                    if( get_fqilla_fakeroots().end() !=
                        get_fqilla_fakeroots().find( make_pair( dynamicContext, GetImpl(c) ) ) )
                    {
                        break;
                    }
                }
                return new FQilla_Iterator_ResultImpl<ctxlist_t::iterator>( info, nodeTest, m_ctx, cl->begin(), cl->end(), cl );
            }
            case XQStep::ATTRIBUTE:
            {
                LG_XQILLA_D << "XQStep::ATTRIBUTE m_eaname:" << m_eaname << endl;

                if( m_eaname.empty() )
                {
                    return new FQilla_Attr_ResultImpl( info, nodeTest, m_ctx );
                }
                break;
            }
            case XQStep::CHILD:
            {
                LG_XQILLA_D << "XQStep::CHILD option..." << endl;
                if( !isAttrNode() )
                {
                    if(!nodeTest->getNameWildcard())
                    {
                        string n = tostr(nodeTest->getNodeName());
                        LG_XQILLA_D << "specific child selected n:" << n << endl;

                        if( m_ctx->isSubContextBound( n ) )
                        {
                            fh_context cc = m_ctx->getSubContext( n );
                            FQilla_Node* theNode = new FQilla_Node(
                                (DynamicContext*)dynamicContext, cc, "" );
                            return new FQilla_SingleNode_ResultImpl( info, nodeTest, cc, theNode );
                        }
                    }
                    
                    
                    return new FQilla_Iterator_ResultImpl<Context::iterator>( info,
                                                                              nodeTest,
                                                                              m_ctx, m_ctx->begin(),
                                                                              m_ctx->end() );
                }
                break;
            }
            case XQStep::DESCENDANT: 
            case XQStep::DESCENDANT_OR_SELF:
            {
                LG_XQILLA_D << "XQStep::DESCENDANT/XQStep::DESCENDANT_OR_SELF" << endl;
            
                ctxlist_t* cl = new ctxlist_t;
                fh_context c = m_ctx;
                if( axis == XQStep::DESCENDANT_OR_SELF )
                {
                    cl->push_back( c );
                }
                getAllChildren( m_ctx, cl );
                return new FQilla_Iterator_ResultImpl<ctxlist_t::iterator>( info, nodeTest, m_ctx, cl->begin(), cl->end(), cl );
                break;
            }
            case XQStep::FOLLOWING:
            {
                if( m_forceNoParent )
                    return 0;
            
                if( m_ctx->isParentBound() )
                {
                    ctxlist_t* cl = new ctxlist_t;
                
                    fh_context p = m_ctx->getParent();
                    Context::iterator    e = p->end();
                    Context::iterator iter = p->find( m_ctx->getDirName() );
                    ++iter;
                    for( ; iter != e ; ++iter )
                    {
                        cl->push_back( *iter );
                        getAllChildren( *iter, cl );
                    }
                    return new FQilla_Iterator_ResultImpl<ctxlist_t::iterator>( info, nodeTest, p, cl->begin(), cl->end(), cl );
                }
            }
            case XQStep::FOLLOWING_SIBLING:
            {
                if( m_forceNoParent )
                    return 0;

                if( m_ctx->isParentBound() )
                {
                    fh_context p = m_ctx->getParent();
                    Context::iterator    e = p->end();
                    Context::iterator iter = p->find( m_ctx->getDirName() );
                    ++iter;
                    return new FQilla_Iterator_ResultImpl<Context::iterator>( info, nodeTest, p, iter, e );
                }
//            return new FollowingSiblingAxis(info, nodeTest, fNode, this, nodeTest);
            }
            case XQStep::NAMESPACE: {
//            if(fNode->getNodeType() == DOMNode::ELEMENT_NODE) {
//                return new NamespaceAxis(info, fNode, this, nodeTest);
//            }
                break;
            }
            case XQStep::PARENT: {
                Node::Ptr p = dmParent( dynamicContext );
                return new FQilla_SingleNode_ResultImpl( info, nodeTest, m_ctx, p );
            }
            case XQStep::PRECEDING:
            {
                if( m_forceNoParent )
                    return 0;

                if( m_ctx->isParentBound() )
                {
                    ctxlist_t* cl = new ctxlist_t;
                
                    fh_context p = m_ctx->getParent();
                    Context::iterator iter = p->begin();
                    Context::iterator    e = p->find( m_ctx->getDirName() );
                    for( ; iter != e ; ++iter )
                    {
                        cl->push_back( *iter );
                        getAllChildren( *iter, cl );
                    }
                    return new FQilla_Iterator_ResultImpl<ctxlist_t::iterator>( info, nodeTest, p, cl->begin(), cl->end(), cl );
                }
            }
            case XQStep::PRECEDING_SIBLING:
            {
                if( m_forceNoParent )
                    return 0;

                if( m_ctx->isParentBound() )
                {
                    fh_context p = m_ctx->getParent();
                    Context::iterator iter = p->begin();
                    Context::iterator    e = p->find( m_ctx->getDirName() );
                    return new FQilla_Iterator_ResultImpl<Context::iterator>( info, nodeTest, p, iter, e );
                }
            }
            case XQStep::SELF: {
                return new FQilla_SingleNode_ResultImpl( info, nodeTest, m_ctx, this );
            }
            }

            return 0;
        
        }
        ATBooleanOrDerived::Ptr FQilla_Node::dmIsId(const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "dmIsId()" << endl;
            return 0;
        }
        ATBooleanOrDerived::Ptr FQilla_Node::dmIsIdRefs(const DynamicContext* dynamicContext) const
        {
            LG_XQILLA_D << "dmIsIdRefs()" << endl;
            return 0;
        }
        const XMLCh* FQilla_Node::getTypeURI() const
        {
            LG_XQILLA_D << "getTypeURI()" << endl;
            return FunctionConstructor::XMLChXPath2DatatypesURI;
        }
        const XMLCh* FQilla_Node::getTypeName() const
        {
            LG_XQILLA_D << "getTypeName()" << endl;

            if( m_eaname.empty() )
            {
                return DocumentCacheParser::g_szUntyped;
            }
            return ATUntypedAtomic::fgDT_UNTYPEDATOMIC;
        }

        fh_context
        FQilla_Node::getFerrisContext() const
        {
            return m_ctx;
        }
    
        bool
        FQilla_Node::isAttrNode() const
        {
            return !m_eaname.empty();
        }


        /////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////

        ///Macro used to facilitate the creation of functions
        template<class TYPE>
        class FuncFactoryTemplate : public FuncFactory
        {
        public:

            FuncFactoryTemplate()
                {
                }

            virtual ASTNode *createInstance(const VectorOfASTNodes &args, XPath2MemoryManager* memMgr) const
                {
                    return new (memMgr) TYPE(args, memMgr);
                }

            virtual const XMLCh *getName() const
                {
                    return TYPE::name;
                }

            virtual const XMLCh *getURI() const
                {
                    return TYPE::XMLChFunctionURI;
                }

            virtual const XMLCh *getQName() const
                {
                    return getName();
                }
            
            virtual unsigned int getMinArgs() const
                {
                    return TYPE::minArgs;
                }

            virtual unsigned int getMaxArgs() const
                {
                    return TYPE::maxArgs;
                }
        };

        

        const XMLCh FunctionFerrisDoc::name[] = {
            chLatin_f, chLatin_e, chLatin_r, chLatin_r, chLatin_i, chLatin_s,
            chDash,
            chLatin_d, chLatin_o, chLatin_c, 
            chNull 
        };
        const unsigned int FunctionFerrisDoc::minArgs = 1;
        const unsigned int FunctionFerrisDoc::maxArgs = 1;

        /**
         * fn:ferris-doc($uri as xs:string?) as document?
         **/
        FunctionFerrisDoc::FunctionFerrisDoc(const VectorOfASTNodes &args, XPath2MemoryManager* memMgr)
            : XQFunction(name, minArgs, maxArgs, "string?", args, memMgr)
        {
        }

        ASTNode* FunctionFerrisDoc::staticResolution(StaticContext *context)
        {
            return resolveArguments(context);
        }

        ASTNode *FunctionFerrisDoc::staticTyping(StaticContext *context)
        {
            _src.clear();

            _src.setProperties(StaticResolutionContext::DOCORDER | StaticResolutionContext::GROUPED |
                               StaticResolutionContext::PEER | StaticResolutionContext::SUBTREE | StaticResolutionContext::ONENODE);
            _src.getStaticType().flags = StaticType::DOCUMENT_TYPE;
            _src.availableDocumentsUsed(true);

            return calculateSRCForArguments(context);
        }

        Sequence FunctionFerrisDoc::collapseTreeInternal(DynamicContext* context, int flags) const
        {
            Sequence uriArg = getParamNumber(1,context)->toSequence(context);
  
            if (uriArg.isEmpty())
            {
                return Sequence(context->getMemoryManager());
            }
  
            const XMLCh* uri = uriArg.first()->asString(context);
//             if(!XPath2Utils::isValidURI(uri, context->getMemoryManager()))
//                 XQThrow(FunctionException,
//                         X("FunctionFerrisDoc::collapseTreeInternal"),
//                         X("Invalid argument to fn:doc function [err:FODC0005]"));

            try
            {
//                return context->resolveDocument(uri, this);
                Sequence result(context->getMemoryManager());

                fh_context m_ctx = Resolve( tostr(uri) );
                FQilla_Node* qn = new FQilla_Node( context, m_ctx, "", true );
                result.addItem( qn );
                return result;
            } 
            //TODO:  once DocumentCacheImpl can throw different errors, we should be able to throw the correct corresponding error messages.
            catch(XMLParseException &e)
            {
                XQThrow(FunctionException,
                        X("FunctionFerrisDoc::collapseTreeInternal"),
                        e.getError());
            }
            return Sequence(context->getMemoryManager());
        }
        

        void RegisterFerrisXQillaFunctions( XQContextImpl* context ) 
        {
//             if( XQDynamicContextImpl* dc = dynamic_cast<XQDynamicContextImpl*>( context ) )
//             {
//                 StaticContext* sc = (StaticContext*)dc->_staticContext;
//                 context = sc;
//             }

            context->addCustomFunction( new FuncFactoryTemplate<FunctionFerrisDoc>() );
            LG_XQILLA_D << "Added custom ferris xqilla functions..." << endl;
        }
        


        /////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////
        /////////////////////////////////////////////////////////////////////////////////
        

        
        FQilla_ItemFactory::FQilla_ItemFactory( ItemFactory* ifactory )
//                                                 const DocumentCache* dc,
//                                                 MemoryManager* memMgr )
            :
                                              //            ItemFactoryImpl( dc, memMgr ),
        m_ifactory( ifactory )
//        m_outputDocument( 0 )
        {
        }
        
        FQilla_ItemFactory::~FQilla_ItemFactory()
        {
        }



        Node::Ptr FQilla_ItemFactory::cloneNode(const Node::Ptr node, const DynamicContext *context) const
        {
            return m_ifactory->cloneNode( node, context );
        }
        
        Node::Ptr FQilla_ItemFactory::createTextNode(const XMLCh *value, const DynamicContext *context) const
        {
            return m_ifactory->createTextNode( value, context );
        }
        
        Node::Ptr FQilla_ItemFactory::createCommentNode(const XMLCh *value, const DynamicContext *context) const
        {
            return m_ifactory->createCommentNode( value, context );
        }
        
        Node::Ptr FQilla_ItemFactory::createPINode(const XMLCh *name, const XMLCh *value, const DynamicContext *context) const
        {
            return m_ifactory->createPINode( name, value, context );
        }
        
        Node::Ptr FQilla_ItemFactory::createAttributeNode(const XMLCh *uri, const XMLCh *prefix, const XMLCh *name,
                                                          const XMLCh *value, const DynamicContext *context) const
        {
            return m_ifactory->createAttributeNode( uri, prefix, name, value, context );
        }


        /**
         * Copy all children of 'c' to 'target'
         */
        void copyChildren( const DynamicContext* dc,
                           DOMDocument* document, fh_context c, DOMElement* target )
        {
            string txt = getStrAttr( c, "content", "" );
            if( !txt.empty() )
            {
                XML::setChildText( document, (DOMElement*)target, txt );
            }

            fh_shouldIncludeAttribute sia = getShouldIncludeAttribute( dc );
            
            AttributeCollection::AttributeNames_t an;
            c->getAttributeNames( an );
            for( AttributeCollection::AttributeNames_t::iterator iter = an.begin();
                 iter != an.end(); iter++ )
            {
                std::string eaname = *iter;
                if( sia->shouldInclude( c, eaname ) )
                {
                    string v = getStrAttr( c, eaname, "" );
                    LG_XQILLA_D << "copying attribute:" << eaname << " value:" << v << endl;
                    ensureAttribute( document, target, eaname, v );
                }
            }
            


            Context::iterator    e = c->end();
            Context::iterator iter = c->begin();
            for( ; iter != e ; ++iter )
            {
                string rdn = (*iter)->getDirName();

                string t = getStrAttr( *iter, "libferris-element-name", "" );
                if( !t.empty() )
                    rdn = t;
                
                string crdn = canonicalElementName( rdn );
                DOMNode *newChild = document->createElement(X( crdn.c_str() ));
                target->appendChild(newChild);

                copyChildren( dc, document, *iter, (DOMElement*)newChild );
            }
        }
        
        
        Node::Ptr FQilla_ItemFactory::createElementNode(const XMLCh *uri, const XMLCh *prefix, const XMLCh *name,
                                                        const std::vector<Node::Ptr> &attrList,
                                                        const std::vector<ItemFactory::ElementChild> &childList,
                                                        const DynamicContext *context) const
        {
//            return m_ifactory->createElementNode( uri, prefix, name, attrList, childList, context );

            StaticContext::ConstructionMode constrMode=context->getConstructionMode();
            bool nsPreserveMode=context->getPreserveNamespaces();
            bool nsInheritMode=context->getInheritNamespaces();

             DOMDocument *document = getOutputDocument(context);
             DOMElement *element = document->createElementNS(uri, name);
             if(prefix != 0 && !XPath2Utils::equals(XMLUni::fgZeroLenString, prefix))
                 element->setPrefix(prefix);

            for(std::vector<Node::Ptr>::const_iterator a = attrList.begin(); a != attrList.end(); ++a)
            {
                if( const NodeImpl *nodeImpl = (const NodeImpl *)(*a)->getInterface(Item::gXQilla) )
                {
                    LG_XQILLA_D << "createElementNode(attr) xqilla native node" << endl;
                    
                    const DOMNode* attr=nodeImpl->getDOMNode();
                    DOMAttr* exists=element->getAttributeNodeNS(attr->getNamespaceURI(), attr->getLocalName());

                    LG_XQILLA_D << "createElementNode(attr) xqilla native node localName:" << tostr(attr->getLocalName()) << endl;

                    DOMAttr* imported = (DOMAttr*)document->importNode(const_cast<DOMNode*>(attr),true);
                    if(constrMode == StaticContext::CONSTRUCTION_MODE_PRESERVE)
                        XPath2Utils::copyAttributeType(document, imported, (const DOMAttr*)attr);
                    if(context->getDebugCallback()) context->getDebugCallback()->ReportClonedNode(const_cast<DynamicContext*>(context), attr, imported);

                    element->setAttributeNodeNS(imported);
                }
                else
                {
                    LG_XQILLA_D << "createElementNode(attr) ferris-xqilla native node" << endl;
                    
                    const FQilla_Node* nodeImpl = (const FQilla_Node*)(*a)->getInterface(FQilla_Node::fQilla);
                    assert( nodeImpl != 0 );

                    LG_XQILLA_D << "createElementNode(attr) ferris-xqilla native node eaname:" << nodeImpl->m_eaname << endl;
                    ensureAttribute( document, element, nodeImpl->m_eaname,
                                     getStrAttr( nodeImpl->m_ctx, nodeImpl->m_eaname, "" ) );
                }
                
                
            }

//             std::vector<const XMLCh*> inScopePrefixes;
//             if(!nsInheritMode)
//             {
//                 DOMNode* elt=element;
//                 while(elt!=NULL)
//                 {
//                     // check if this node has a namespace, but the prefix is not declared in the attribute map
//                     const XMLCh* uri=elt->getNamespaceURI();
//                     if(uri && *uri)
//                     {
//                         const XMLCh* prefix=elt->getPrefix();
//                         if (!XPath2Utils::containsString(inScopePrefixes, prefix))
//                             inScopePrefixes.push_back(prefix);
//                     }
//                     DOMNamedNodeMap* attrMap=elt->getAttributes();
//                     for(XMLSize_t i=0;i<attrMap->getLength();i++)
//                     {
//                         DOMNode* attr=attrMap->item(i);
//                         if(XPath2Utils::equals(attr->getPrefix(), XMLUni::fgXMLNSString) || XPath2Utils::equals(attr->getNodeName(), XMLUni::fgXMLNSString))
//                         {
//                             const XMLCh* prefix=attr->getPrefix()==NULL?XMLUni::fgZeroLenString:attr->getLocalName();
//                             if (!XPath2Utils::containsString(inScopePrefixes, prefix))
//                                 inScopePrefixes.push_back(prefix);
//                         }
//                     }
//                     elt=elt->getParentNode();
//                 }
//             }

            for(std::vector<ItemFactory::ElementChild>::const_iterator i = childList.begin(); i != childList.end(); ++i)
            {
                if( const NodeImpl *nodeImpl = (const NodeImpl *)(*i)->getInterface(Item::gXQilla) )
                {
                    LG_XQILLA_D << "createElementNode(element) xqilla native node" << endl;
                    LG_XQILLA_D << "createElementNode(element) xqilla native node localName:" << tostr(nodeImpl->getDOMNode()->getLocalName()) << endl;
//                     LG_XQILLA_D << "createElementNode(element) xqilla native node year:"
//                                 << getAttribute( (DOMElement*)(nodeImpl->getDOMNode()), "year" )
//                                 << endl;
                    

                    DOMNode *newChild = NULL;
                    if(nodeImpl->getDOMNode()->getOwnerDocument() == document)
                    {
                        LG_XQILLA_D << "createElementNode(element) xqilla native node (same doc) i->clone:" << i->clone << endl;
                        
                        if(i->clone) {
                            newChild = nodeImpl->getDOMNode()->cloneNode(true);
//                             if(constrMode == StaticContext::CONSTRUCTION_MODE_PRESERVE && nodeImpl->dmNodeKind()==Node::element_string)
//                                 XPath2Utils::copyElementType(newChild->getOwnerDocument(), (DOMElement*)newChild, (DOMElement*)nodeImpl->getDOMNode());
//                             if(context->getDebugCallback()) context->getDebugCallback()->ReportClonedNode(const_cast<DynamicContext*>(context), nodeImpl->getDOMNode(), newChild);
                        }
                        else
                        {
                            LG_XQILLA_D << "createElementNode(element) xqilla native node (diff doc)" << endl;
                            newChild = const_cast<DOMNode*>(nodeImpl->getDOMNode());
                        }
                    }
                    else {
                        newChild = document->importNode(const_cast<DOMNode*>(nodeImpl->getDOMNode()),true);
                        if(constrMode == StaticContext::CONSTRUCTION_MODE_PRESERVE && nodeImpl->dmNodeKind()==Node::element_string)
                            XPath2Utils::copyElementType(newChild->getOwnerDocument(), (DOMElement*)newChild, (DOMElement*)nodeImpl->getDOMNode());
                        if(context->getDebugCallback()) context->getDebugCallback()->ReportClonedNode(const_cast<DynamicContext*>(context), nodeImpl->getDOMNode(), newChild);
                    }

//                     if(!nsPreserveMode && newChild->getNodeType()==DOMNode::ELEMENT_NODE)
//                     {
//                         DOMElement* childElem=(DOMElement*)newChild;
      
//                         DOMNamedNodeMap* attrMap=childElem->getAttributes();
//                         for(XMLSize_t i=0;i<attrMap->getLength();)
//                         {
//                             bool bPreserved=true;
//                             DOMNode* attr=attrMap->item(i);

//                             if(XPath2Utils::equals(attr->getPrefix(), XMLUni::fgXMLNSString) || XPath2Utils::equals(attr->getNodeName(), XMLUni::fgXMLNSString))
//                             {
//                                 const XMLCh* prefix=attr->getPrefix()==NULL?XMLUni::fgZeroLenString:attr->getLocalName();
//                                 // copy this namespace only if needed by the element name...
//                                 if(!XPath2Utils::equals(childElem->getPrefix(), prefix))
//                                 {
//                                     bPreserved=false;
//                                     //... or by its attributes
//                                     for(XMLSize_t j=0;j<attrMap->getLength();j++)
//                                         if(XPath2Utils::equals(attrMap->item(j)->getPrefix(), prefix))
//                                         {
//                                             bPreserved=true;
//                                             break;
//                                         }
//                                     if(!bPreserved)
//                                         attrMap->removeNamedItemNS(attr->getNamespaceURI(), attr->getLocalName());
//                                 }
//                             }
//                             if(bPreserved)
//                                 i++;
//                         }
//                     }

                    element->appendChild(newChild);
                }
                else
                {
                    LG_XQILLA_D << "createElementNode(element) ferris-xqilla native node" << endl;
                    
                    const FQilla_Node* nodeImpl = (const FQilla_Node*)(*i)->getInterface(FQilla_Node::fQilla);
                    assert( nodeImpl != 0 );

                    LG_XQILLA_D << "createElementNode(element) ferris-xqilla native node ctx:" << nodeImpl->m_ctx->getURL() << endl;
                    LG_XQILLA_D << "createElementNode(element) ferris-xqilla native node i->clone:" << i->clone << endl;
                    
                    string rdn = nodeImpl->m_ctx->getDirName();
                    LG_XQILLA_D << "createElementNode(element) ...2" << endl;
                    string crdn = canonicalElementName( rdn );
                    LG_XQILLA_D << "createElementNode(element) ...3 crdn:" << crdn << endl;
                    DOMNode *newChild = document->createElement(X( crdn.c_str() ));
                    LG_XQILLA_D << "createElementNode(element) ...4" << endl;
                    
                    // FIXME: maybe copy nodeImpl -> attributes and children to newChild?
                    // we should do this because the above is all deep copies,
                    // BUT: of all the attributes how do we know which to copy?

                    copyChildren( context, document, nodeImpl->m_ctx, (DOMElement*)newChild );
                    LG_XQILLA_D << "createElementNode(element) ...5" << endl;
                    
//                    ensureAttribute( document, (DOMElement*)newChild, "foo", "bar" );
//                     {
//                         string txt = getStrAttr( nodeImpl->m_ctx, "content", "" );
//                         if( !txt.empty() )
//                         {
//                             XML::setChildText( document, (DOMElement*)newChild, txt );
//                         }
//                     }
                    
                    element->appendChild(newChild);
                    LG_XQILLA_D << "createElementNode(element) ferris-xqilla done..." << endl;
                }
            }  


            return new NodeImpl(element, context);
        }
        
        Node::Ptr
        FQilla_ItemFactory::createDocumentNode(const std::vector<Node::Ptr> &childList, const DynamicContext *context) const
        {
            return m_ifactory->createDocumentNode( childList, context );
        }
        
        const DOMNode*
        FQilla_ItemFactory::createNamespaceNode(const XMLCh* prefix, const XMLCh* uri,
                            const DOMNode* parentNode, const DynamicContext *context) const
        {
            return m_ifactory->createNamespaceNode( prefix, uri, parentNode, context );
        }
        

        AnyAtomicType::AtomicObjectType
        FQilla_ItemFactory::getPrimitiveTypeIndex(const XMLCh* typeURI, const XMLCh* typeName, bool &isPrimitive) const
        {
            return m_ifactory->getPrimitiveTypeIndex( typeURI, typeName, isPrimitive );
        }
        
        AnyAtomicType::Ptr
        FQilla_ItemFactory::createDerivedFromAtomicType(AnyAtomicType::AtomicObjectType typeIndex, const XMLCh* value,
                                                        const DynamicContext* context)
        {
            return m_ifactory->createDerivedFromAtomicType( typeIndex, value, context );
        }
        
        AnyAtomicType::Ptr
        FQilla_ItemFactory::createDerivedFromAtomicType(AnyAtomicType::AtomicObjectType typeIndex,
                                                        const XMLCh* typeURI,
                                                        const XMLCh* typeName,
                                                        const XMLCh* value,
                                                        const DynamicContext* context)
        {
            return m_ifactory->createDerivedFromAtomicType( typeIndex, typeURI, typeName, value, context );
        }
        
        AnyAtomicType::Ptr
        FQilla_ItemFactory::createDerivedFromAtomicType(const XMLCh* typeURI,
                                                        const XMLCh* typeName,
                                                        const XMLCh* value,
                                                        const DynamicContext* context)
        {
            return m_ifactory->createDerivedFromAtomicType( typeURI, typeName, value, context );
        }
        

        ATDoubleOrDerived::Ptr
        FQilla_ItemFactory::createDouble(const MAPM value, const DynamicContext* context)
        {
            return m_ifactory->createDouble( value, context );
        }
        
        ATDoubleOrDerived::Ptr
        FQilla_ItemFactory::createDouble(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createDouble( value, context );
        }
            
        ATDoubleOrDerived::Ptr
        FQilla_ItemFactory::createDoubleOrDerived(const XMLCh* typeURI,
                                                  const XMLCh* typeName,
                                                  const MAPM value,
                                                  const DynamicContext* context)
        {
            return m_ifactory->createDoubleOrDerived( typeURI, typeName, value, context );
        }
            
        ATDoubleOrDerived::Ptr
        FQilla_ItemFactory::createDoubleOrDerived(const XMLCh* typeURI,
                                                  const XMLCh* typeName,
                                                  const XMLCh* value,
                                                  const DynamicContext* context)
        {
            return m_ifactory->createDoubleOrDerived( typeURI, typeName, value, context );
        }
            

        ATFloatOrDerived::Ptr
        FQilla_ItemFactory::createFloat(const MAPM value, const DynamicContext* context)
        {
            return m_ifactory->createFloat( value, context );
        }
        
        ATFloatOrDerived::Ptr
        FQilla_ItemFactory::createFloat(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createFloat( value, context );
        }
            
        ATFloatOrDerived::Ptr
        FQilla_ItemFactory::createFloatOrDerived(const XMLCh* typeURI,
                                                 const XMLCh* typeName,
                                                 const MAPM value,
                                                 const DynamicContext* context)
        {
            return m_ifactory->createFloatOrDerived( typeURI, typeName, value, context );
        }
            
        ATFloatOrDerived::Ptr
        FQilla_ItemFactory::createFloatOrDerived(const XMLCh* typeURI,
                                                 const XMLCh* typeName,
                                                 const XMLCh* value,
                                                 const DynamicContext* context)
        {
            return m_ifactory->createFloatOrDerived( typeURI, typeName, value, context );
        }

        ATDecimalOrDerived::Ptr FQilla_ItemFactory::createInteger(const int value, const DynamicContext* context)
        {
            return m_ifactory->createInteger( value, context );
        }
        ATDecimalOrDerived::Ptr FQilla_ItemFactory::createInteger(const MAPM value, const DynamicContext* context)
        {
            return m_ifactory->createInteger( value, context );
        }
        ATDecimalOrDerived::Ptr FQilla_ItemFactory::createInteger(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createInteger( value, context );
        }
        ATDecimalOrDerived::Ptr FQilla_ItemFactory::createNonNegativeInteger(const MAPM value, const DynamicContext* context)
        {
            return m_ifactory->createNonNegativeInteger( value, context );
        }
            

        ATDecimalOrDerived::Ptr FQilla_ItemFactory::createDecimal(const MAPM value, const DynamicContext* context)
        {
            return m_ifactory->createDecimal( value, context );
        }
        ATDecimalOrDerived::Ptr FQilla_ItemFactory::createDecimal(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createDecimal( value, context );
        }
        ATDecimalOrDerived::Ptr FQilla_ItemFactory::createDecimalOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                           const MAPM value, const DynamicContext* context)
        {
            return m_ifactory->createDecimalOrDerived( typeURI, typeName, value, context );
        }
        ATDecimalOrDerived::Ptr FQilla_ItemFactory::createDecimalOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                           const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createDecimalOrDerived( typeURI, typeName, value, context );
        }
            

        
        ATDateOrDerived::Ptr FQilla_ItemFactory::createDate(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createDate( value, context );
        }
        ATDateOrDerived::Ptr FQilla_ItemFactory::createDateOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                     const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createDateOrDerived( typeURI, typeName, value, context );
        }

        ATDateTimeOrDerived::Ptr FQilla_ItemFactory::createDateTime(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createDateTime( value, context );
        }
        ATDateTimeOrDerived::Ptr FQilla_ItemFactory::createDateTimeOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                             const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createDateTimeOrDerived( typeURI, typeName, value, context );
        }


        ATTimeOrDerived::Ptr FQilla_ItemFactory::createTime(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createTime( value, context );
        }
        ATTimeOrDerived::Ptr FQilla_ItemFactory::createTimeOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                     const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createTimeOrDerived( typeURI, typeName, value, context );
        }
            

        ATGDayOrDerived::Ptr FQilla_ItemFactory::createGDayOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                     const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createGDayOrDerived( typeURI, typeName, value, context );
        }
        ATGMonthDayOrDerived::Ptr FQilla_ItemFactory::createGMonthDayOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                               const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createGMonthDayOrDerived( typeURI, typeName, value, context );
        }
        ATGMonthOrDerived::Ptr FQilla_ItemFactory::createGMonthOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                         const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createGMonthOrDerived( typeURI, typeName, value, context );
        }
        ATGYearMonthOrDerived::Ptr FQilla_ItemFactory::createGYearMonthOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                                 const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createGYearMonthOrDerived( typeURI, typeName, value, context );
        }
        ATGYearOrDerived::Ptr FQilla_ItemFactory::createGYearOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                       const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createGYearOrDerived( typeURI, typeName, value, context );
        }

        
        ATDurationOrDerived::Ptr FQilla_ItemFactory::createDayTimeDuration(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createDayTimeDuration( value, context );
        }
            
        ATDurationOrDerived::Ptr FQilla_ItemFactory::createDayTimeDuration(const MAPM &seconds, const DynamicContext* context)
        {
            return m_ifactory->createDayTimeDuration( seconds, context );
        }
        ATDurationOrDerived::Ptr FQilla_ItemFactory::createYearMonthDuration(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createYearMonthDuration( value, context );
        }
        ATDurationOrDerived::Ptr FQilla_ItemFactory::createYearMonthDuration(const MAPM &months, const DynamicContext* context)
        {
            return m_ifactory->createYearMonthDuration( months, context );
        }
        ATDurationOrDerived::Ptr FQilla_ItemFactory::createDurationOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                             const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createDurationOrDerived( typeURI, typeName, value, context );
        }
            

        ATStringOrDerived::Ptr FQilla_ItemFactory::createString(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createString( value, context );
        }
        ATStringOrDerived::Ptr FQilla_ItemFactory::createStringOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                         const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createStringOrDerived( typeURI, typeName, value, context );
        }
            

        ATUntypedAtomic::Ptr FQilla_ItemFactory::createUntypedAtomic(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createUntypedAtomic( value, context );
        }

        ATBooleanOrDerived::Ptr FQilla_ItemFactory::createBoolean(bool value, const DynamicContext* context)
        {
            return m_ifactory->createBoolean( value, context );
        }
        ATBooleanOrDerived::Ptr FQilla_ItemFactory::createBoolean(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createBoolean( value, context );
        }
        ATBooleanOrDerived::Ptr FQilla_ItemFactory::createBooleanOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                           bool value, const DynamicContext* context)
        {
            return m_ifactory->createBooleanOrDerived( typeURI, typeName, value, context );
        }
        ATBooleanOrDerived::Ptr FQilla_ItemFactory::createBooleanOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                           const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createBooleanOrDerived( typeURI, typeName, value, context );
        }

        ATAnyURIOrDerived::Ptr FQilla_ItemFactory::createAnyURI(const XMLCh* value, const DynamicContext* context)
        {
            return m_ifactory->createAnyURI( value, context );
        }

        ATQNameOrDerived::Ptr FQilla_ItemFactory::createQName(const XMLCh* uri, const XMLCh* prefix, const XMLCh* name,
                                                              const DynamicContext* context)
        {
            return m_ifactory->createQName( uri, prefix, name, context );
        }
            
        ATQNameOrDerived::Ptr FQilla_ItemFactory::createQNameOrDerived(const XMLCh* typeURI, const XMLCh* typeName,
                                                                       const XMLCh* uri, const XMLCh* prefix,
                                                                       const XMLCh* name, const DynamicContext* context)
        {
            return m_ifactory->createQNameOrDerived( typeURI, typeName, uri, prefix, name, context );
        }
            

        DOMDocument*
        FQilla_ItemFactory::getOutputDocument(const DynamicContext *context) const
        {
//            getDOM( context );
//            return GetImpl( m_outputDocument );
            return ((ItemFactoryImpl*)m_ifactory)->getOutputDocument( context );
        }
            
        
//         fh_domdoc
//         FQilla_ItemFactory::getDOM( const DynamicContext *context ) const
//         {
//             return 0;
            
// //             if( !m_outputDocument )
// //             {
// //                 m_outputDocument = Factory::makeDOM( "" );
// //             }
            
// //             return m_outputDocument;
            

// //              if( m_outputDocument )
// //                  return m_outputDocument;

// //              m_outputDocument = getOutputDocument( context );

// // //             static list< fh_domdoc > explicitLeakList;
// // //             explicitLeakList.push_back( m_outputDocument );
// //              return m_outputDocument;
//         }
        


        
            
            
//         Ferris_XQDynamicContextImpl::Ferris_XQDynamicContextImpl(
//             const StaticContext *staticContext,
//             XERCES_CPP_NAMESPACE_QUALIFIER MemoryManager* memMgr)
//             :
//             XQDynamicContextImpl( staticContext, memMgr )
//         {
//         }
        
//             Ferris_XQDynamicContextImpl::~Ferris_XQDynamicContextImpl()
//         {
//         }
        
        

        
    };
};
