/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: xfsutil.cpp,v 1.2 2005/09/04 06:44:06 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#ifdef FERRIS_HAVE_XFSPROGS
#include <xfsutil.hh>
#include <libxfs.h>
#endif
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/vfs.h>
#include <Ferris.hh>
#include <Resolver_private.hh>

namespace Ferris
{

    bool isXFS( int fd )
    {
#ifndef FERRIS_HAVE_XFSPROGS
        return false;
#else
        struct statfs	buf;

        fstatfs(fd, &buf);
        if (buf.f_type == XFS_SUPER_MAGIC)
        {
            return true;
        }
        return false;
#endif
    }
    
    bool isXFS( fh_context c )
    {
#ifndef FERRIS_HAVE_XFSPROGS
        return false;
#else
        const std::string s = c->getDirPath();
        return isXFS( s );
#endif
    }
    
    bool isXFS( const std::string& s )
    {
#ifndef FERRIS_HAVE_XFSPROGS
        return false;
#else
        FerrisURL fu = FerrisURL::fromString( s );

        LG_NATIVE_D << " isXFS s:" << s
                    << " scheme:" << fu.getScheme()
                    << endl;

        if( fu.getScheme() != "file" )
        {
            LG_NATIVE_D << "isXFS s:" << s
                        << " not using a file scheme: " << fu.getScheme()
                        << endl;
            return false;
        }

        int fd = open( s.c_str(), O_RDONLY);
        if (fd < 0)
        {
            return false;
        }

        bool ret = isXFS( fd );
        LG_NATIVE_D << " isXFS() s:" << s << " ret:" << ret << endl;
        close(fd);
        return ret;
#endif
    }
    
};
