/******************************************************************************
*******************************************************************************
*******************************************************************************

    ferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: ferris-import-desktop-file.cpp,v 1.1 2005/07/04 09:18:13 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <Ferris.hh>
#include <popt.h>

using namespace std;
using namespace Ferris;

const string PROGRAM_NAME = "ferris-import-desktop-file";

void usage(poptContext optCon, int exitcode, char *error, char *addl)
{
    poptPrintUsage(optCon, stderr, 0);
    if (error) fprintf(stderr, "%s: %s0", error, addl);
    exit(exitcode);
}

int main( int argc, char** argv )
{
    try
    {
        const char*   DstNameCSTR     = 0;
        unsigned long Verbose         = 0;
        unsigned long MakeTargetDirectory = 0;

        struct poptOption optionsTable[] =
            {
                { "verbose", 'v', POPT_ARG_NONE, &Verbose, 0,
                  "show what is happening", "" },

                { "target-directory", 'd', POPT_ARG_STRING, &DstNameCSTR, 0,
                  "Explicitly specify where in the apps tree to import to", "Applications" },

                { "create-target-directory", 'p', POPT_ARG_NONE, &MakeTargetDirectory, 0,
                  "create target-directory if it doesn't exist already", "" },
                
                FERRIS_POPT_OPTIONS
                POPT_AUTOHELP
                POPT_TABLEEND
            };
        poptContext optCon;

        optCon = poptGetContext(PROGRAM_NAME.c_str(), argc, (const char**)argv, optionsTable, 0);
        poptSetOtherOptionHelp(optCon, "[OPTIONS]* src1 src2 ...");

        /* Now do options processing */
        char c=-1;
        while ((c = poptGetNextOpt(optCon)) >= 0)
        {
//         switch (c) {
//         }
        }

        if (argc < 2)
        {
            poptPrintHelp(optCon, stderr, 0);
            exit(1);
        }

        while( const char* pathCSTR = poptGetArg(optCon) )
        {
            string path = pathCSTR;
        
            try
            {
                fh_context c = Resolve( path );

                if( DstNameCSTR )
                {
                    if( MakeTargetDirectory )
                    {
                        fh_context c = Shell::acquireContext( DstNameCSTR );
                    }
                    
                    fh_context destc = Resolve( DstNameCSTR );
                    importDesktopFileTo( destc, c );
                }
                else
                {
                    importDesktopFile( c );
                }
            }
            catch( exception& e )
            {
                cerr << "error:" << e.what() << endl;
            }
        }
    }
    catch( exception& e )
    {
        cerr << "error:" << e.what() << endl;
        exit(1);
    }
    return 0;
}


