/******************************************************************************
*******************************************************************************
*******************************************************************************

    ferris internal metadata worker
    Copyright (C) 2007 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: ferris-internal-metadata-worker.cpp,v 1.3 2007/08/31 21:30:13 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include "config.h"
#include <Ferris/Ferris.hh>
#include <Ferris/Ferris_private.hh>
#include <Ferris/MatchedEAGenerators.hh>
#include <Ferris/MetadataServer_private.hh>

#include <popt.h>
#include <signal.h>
#include <time.h>

#include <dbus-c++/dbus.h>
#include <dbus-c++/glib-integration.h>
#include "DBusGlue/ferris-internal-metadata-worker-server-glue.h"

//#include <gtkmm.h>

using namespace std;
using namespace Ferris;

const string PROGRAM_NAME = "ferris-internal-metadata-worker";

void usage(poptContext optCon, int exitcode, char *error, char *addl)
{
    poptPrintUsage(optCon, stderr, 0);
    if (error) fprintf(stderr, "%s: %s0", error, addl);
    exit(exitcode);
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////



int RETURN_CODE = 0;
bool running = true;

namespace Ferris
{
    guint32 glib_idle_timeout_Interval = 30 * 60;
    void resetup_glib_idle_to_close_callback_function()
    {
        resetup_glib_idle_to_close_callback_function( glib_idle_timeout_Interval, &running );
    }
};



namespace Ferris
{
    class MetadataWorker
        :
        public net::sf::witme::Libferris::Metadata::Worker,
        public DBus::IntrospectableAdaptor,
        public DBus::ObjectAdaptor
    {
        typedef stringset_t m_pluginNames_t;
        m_pluginNames_t m_pluginNames;
    public:
    
        MetadataWorker( DBus::Connection& connection,
                        const stringset_t& pluginNames );
        ~MetadataWorker();

        std::vector< ::DBus::Byte > get( const DBus::String& earl, const DBus::String& name );
        void set( const DBus::String& earl, const DBus::String& name, const std::vector< ::DBus::Byte >& value );
        ::DBus::Int32 ping();
        
    };

    MetadataWorker::MetadataWorker( DBus::Connection& connection,
                                    const stringset_t& pluginNames )
        :
        DBus::ObjectAdaptor( connection, "/net/sf/witme/libferris/Metadata" ),
        m_pluginNames( pluginNames )
    {
        LG_MDSERV_D << "MetadataWorker()" << endl;
    }

    MetadataWorker::~MetadataWorker()
    {
        LG_MDSERV_D << "~MetadataWorker()" << endl;
    }

    std::vector< ::DBus::Byte >
    MetadataWorker::get( const DBus::String& earl, const DBus::String& eaname )
    {
        resetup_glib_idle_to_close_callback_function();

        LG_MDSERV_I << "MetadataWorker::get() url:" << earl << " eaname:" << eaname << endl;
        
        stringstream ss;
        try
        {
            fh_context c = Resolve( earl );

            LG_MDSERV_D << "MetadataWorker::get() calling getStrAttr()."
                        << " url:" << earl << " eaname:" << eaname << endl;
            
            std::string v = c->getStrAttr_UsingRestrictedPlugins( eaname, m_pluginNames );
//            ss << v;

            LG_MDSERV_D << "MetadataWorker::get(ok) url:" << earl
                        << " eaname:" << eaname
                        << " ret.sz:" << v.size()
                        << endl;
            if( eaname != "rgba-32bpp" )
            {
                LG_MDSERV_D << "MetadataWorker::get(ok) url:" << earl
                            << " eaname:" << eaname
                            << " ret:" << v
                            << endl;
            }

            std::vector< ::DBus::Byte > ret( v.begin(), v.end() );
            return ret;
        }
        catch( exception& e )
        {
            ss << e.what();
            cerr << ss.str() << endl;
            LG_MDSERV_W << "MetadataWorker::get(fail) url:" << earl
                        << " eaname:" << eaname
                        << " e:" << ss.str()
                        << endl;
            
            throw DBus::Error("net.sf.witme.Libferris.MetadataExtractionError", ss.str().c_str() );
        }
    }

    void
    MetadataWorker::set( const DBus::String& earl,
                         const DBus::String& eaname,
                         const std::vector< ::DBus::Byte >& value )
    {
        resetup_glib_idle_to_close_callback_function();
        LG_MDSERV_W << "MetadataWorker::set() not implemented!" << endl;
        throw DBus::Error("net.sf.witme.Libferris.MetadataExtractionError",
                          "MetadataWorker::set() not implemented!" );
    }
    
    ::DBus::Int32
    MetadataWorker::ping()
    {
        LG_MDSERV_D << "MetadataWorker::ping()" << endl;

        resetup_glib_idle_to_close_callback_function();
        ::DBus::Int32 t = time( 0 );
        return t;
    }
};




////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

using namespace Ferris;
DBus::Glib::BusDispatcher dispatcher;

void niam( int sig )
{
//	dispatcher.leave();
    running = false;
}



int main( int argc, const char** argv )
{
    signal(SIGTERM, niam);
	signal(SIGINT, niam);

	DBus::default_dispatcher = &dispatcher;
    dispatcher.attach(NULL);

    setForceOutOfProcessMetadataOff( true );

    unsigned long ShowVersion        = 0;
    const char*   PluginNames_CSTR   = 0;
    const char*   FormatDisplayCSTR  = 0;
    const char*   DBusServerName_CSTR = 0;
    
    struct poptOption optionsTable[] = {

        { "version", 0, POPT_ARG_NONE, &ShowVersion, 0,
          "show version information and quit", 0 },

        { "plugin-names", 0, POPT_ARG_STRING, &PluginNames_CSTR, 0,
          "comma seperated list of plugins to use for metadata handling", 0 },

        { "dbus-server-name", 0, POPT_ARG_STRING, &DBusServerName_CSTR, 0,
          "name to use for this dbus server object", 0 },
        
        /*
         * Standard Ferris options
         */
//        FERRIS_POPT_OPTIONS
        POPT_AUTOHELP
        POPT_TABLEEND
    };
    poptContext optCon;


    optCon = poptGetContext(PROGRAM_NAME.c_str(), argc, argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* ...");

    /* Now do options processing, get portname */
    char c=-1;
    while ((c = poptGetNextOpt(optCon)) >= 0)
    {}

    if (argc < 1 || !PluginNames_CSTR )
    {
        cerr << "Arguments incorrect!" << endl;
//        poptPrintHelp(optCon, stderr, 0);
        poptPrintUsage(optCon, stderr, 0);
        exit(1);
    }
    
    if( ShowVersion )
    {
        cout << "ferris-internal-metadata-worker" << nl
             << "version: $Id: ferris-internal-metadata-worker.cpp,v 1.3 2007/08/31 21:30:13 ben Exp $\n"
             << "ferris   version: " << VERSION << nl
             << "Written by Ben Martin, aka monkeyiq" << nl
             << nl
             << "Copyright (C) 2007 Ben Martin" << nl
             << "This is free software; see the source for copying conditions.  There is NO\n"
             << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE."
             << endl;
        exit(0);
    }

    string DBusServerName = "net.sf.witme.Libferris.Metadata.Worker";
    if( DBusServerName_CSTR )
        DBusServerName = DBusServerName_CSTR;
    
        
    stringset_t pluginNames;
    Util::parseCommaSeperatedList( PluginNames_CSTR, pluginNames );

    resetup_glib_idle_to_close_callback_function();
    
    LG_MDSERV_D << "about to start new worker process" << endl;
    LG_MDSERV_D << "Registering as:" << DBusServerName << endl;
    
    try
    {

        DBus::Connection conn = DBus::Connection::SessionBus();
        conn.request_name( DBusServerName.c_str() );

        LG_MDSERV_D << "about to start worker object" << endl;
        MetadataWorker server( conn, pluginNames );

        LG_MDSERV_D << "about to enter dispatcher loop" << endl;
//        dispatcher.enter();
        
//        Gtk::Main::run( server );

        while( running )
            g_main_iteration( true );

    }
    catch( NoSuchContextClass& e )
    {
        cerr << "e:" << e.what() << endl;
        LG_MDSERV_W << "Worker process exiting with error:" << e.what() << endl;
        exit(1);
    }
    catch( exception& e )
    {
        cerr << "cought:" << e.what() << endl;
        LG_MDSERV_W << "Worker process exiting with error:" << e.what() << endl;
        exit(1);
    }
        
    poptFreeContext(optCon);
    cout << flush;
    LG_MDSERV_D << "Worker process exiting" << endl;
    return RETURN_CODE;
}
