/******************************************************************************
*******************************************************************************
*******************************************************************************

    ferris cp GTK2 client
    Copyright (C) 2002 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: gfcp.cpp,v 1.2 2006/08/13 11:47:59 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <algorithm>
#include <iomanip>

#include <FerrisCopy.hh>
#include <General.hh>

#include <FerrisUI/gtkferriscellrenderertext.hh>
#include <FerrisUI/gtkferristreestore.hh>
#include <FerrisUI/TreeStoreDriver.hh>
#include <FerrisUI/GtkFerris.hh>
#include <FerrisUI/FerrisCopyUI.hh>

#include <popt.h>
#include <sys/types.h>
#include <sys/stat.h>
//#include <sys/time.h>
//#include <signal.h>
#include <unistd.h>

#include <glib.h>
#include <gtk/gtk.h>

using namespace std;
using namespace Ferris;
using namespace FerrisUI;


const string PROGRAM_NAME = "gfcp";


void usage(poptContext optCon, int exitcode, char *error, char *addl)
{
    poptPrintUsage(optCon, stderr, 0);
    if (error) fprintf(stderr, "%s: %s0", error, addl);
    exit(exitcode);
}

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

class FerrisCopy_GTK;

FERRIS_SMARTPTR( FerrisCopy_GTK, fh_cp_gtk );



class FerrisCopy_GTK
    :
    public FerrisCopy,
    public GTK_TreeWalkClient
{
    typedef FerrisCopy       _Base;
    typedef FerrisCopy_GTK   _Self;

    FerrisCopy_SignalHandler      sighandler;

    friend class FerrisCopy_SignalHandler;

public:

    FerrisCopy_GTK()
        :
        sighandler( this, this )
        {
            makeMainWindow( PROGRAM_NAME );
        }

    virtual ~FerrisCopy_GTK()
        {
        }


    void addBodyElements( GtkWidget* table, int& r )
        {
            addSourceRow(      table, r );
            addDestinationRow( table, r );
            addStartTimeRow(   table, r );
            addElapsedTimeRow( table, r );
            addSpeedRow(       table, r );
            addProgressRow(    table, r );
            addVCRControls(    table );
        }

    int main( int argc, char** argv );
};

int
FerrisCopy_GTK::main( int argc, char** argv )
{
    const char* DstNameCSTR              = 0;
    struct poptOption optionsTable[] =
        {
            { "target-directory", 0, POPT_ARG_STRING, &DstNameCSTR, 0,
              "Specify destination explicity, all remaining URLs are assumed to be source files",
              "DIR" },
                    
            FERRIS_COPY_OPTIONS( this )
            FERRIS_POPT_OPTIONS
            POPT_AUTOHELP
            POPT_TABLEEND
        };
    poptContext optCon;

    optCon = poptGetContext(PROGRAM_NAME.c_str(), argc, (const char**)argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* src1 src2 ... dst");

    /* Now do options processing, get portname */
    char c=-1;
    while ((c = poptGetNextOpt(optCon)) >= 0)
    {
//                 switch (c) {
//                 }
    }

    getPoptCollector()->ArgProcessingDone( optCon );
    setVerbose( true );
    
    if (argc < 3)
    {
        poptPrintHelp(optCon, stderr, 0);
        exit(1);
    }

    showMainWindow( Sloth );
    
    try
    {
        typedef vector<string> srcs_t;
        srcs_t srcs;
                
        while( const char* RootNameCSTR = poptGetArg(optCon) )
        {
            string RootName = RootNameCSTR;
            srcs.push_back( RootName );
        }
                
        /*
         * Setup the destination from either explicit command line or last arg
         */
        string DstName;
        if( DstNameCSTR )
        {
            setDstIsDirectory( true );
            DstName = DstNameCSTR;
        }
        else
        {
            DstName = srcs.back();
            srcs.pop_back();
        }
                
//        cerr << "dst:" << DstName << endl;
        setDstURL( DstName );
        
        {
            fh_stringstream ss;
            ss << "dst: " << DstName;
            gtk_label_set_text( GTK_LABEL(m_dstlab), g_strdup( tostr(ss).c_str() ) );
        }
        
        /************************************************************/
        /************************************************************/
        /************************************************************/

        populateModelWithSelection( m_treemodel, srcs.begin(), srcs.end() );
        makeDefaultColumnViews();
        processAllPendingEvents();

        sighandler.restartTimer();
        
        for( srcs_t::iterator iter = srcs.begin(); iter != srcs.end(); ++iter )
        {
            string SrcName = *iter;
//            cerr << "src:" << SrcName << endl;
                    
            setSrcURL( SrcName );
            copy();
        }

        if( getShouldRunMainLoop( AutoClose, hadUserInteraction ))
        {
            runMainWindow( Sloth );
        }
    }
    catch( Quit_Requested& e )
    {
        cerr << "cought quit request, exiting." << endl;
        exit(0);
    }
    catch( NoSuchContextClass& e )
    {
        fh_stringstream ss;
        ss << "Invalid context class given e:" << e.what() << endl;
        cerr << tostr(ss);
        GtkWidget* dialog = gtk_message_dialog_new (
            0,
            GtkDialogFlags(GTK_DIALOG_MODAL),
            GTK_MESSAGE_ERROR,
            GTK_BUTTONS_CLOSE,
            tostr(ss).c_str()
            );
        gtk_dialog_run (GTK_DIALOG (dialog));
        gtk_widget_destroy (dialog);
        exit(1);
    }
    catch( exception& e )
    {
        fh_stringstream ss;
        ss << "Error: " << e.what() << endl;
        cerr << tostr(ss);
        GtkWidget* dialog = gtk_message_dialog_new (
            0,
            GtkDialogFlags(GTK_DIALOG_MODAL),
            GTK_MESSAGE_ERROR,
            GTK_BUTTONS_CLOSE,
            tostr(ss).c_str()
            );
        gtk_dialog_run (GTK_DIALOG (dialog));
        gtk_widget_destroy (dialog);
        exit(1);
    }
    poptFreeContext(optCon);
    return 0;
}

/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/

int main( int argc, char** argv )
{
//    cerr << "gfcp main() " << endl;
    gtk_init( &argc, &argv );
    fh_cp_gtk obj = new FerrisCopy_GTK();
    return obj->main( argc, argv );
}

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/



