/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: ferris-capplet-auth.cpp,v 1.7 2007/04/08 21:30:13 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <typeinfo>

#include "config.h"

#ifdef HAVE_DTL
#include <DTL.h>
#endif

#include <glib.h>
#include <gtk/gtk.h>

#include <Ferris.hh>

#ifdef HAVE_WEBPHOTOS_API
#include "Ferris/WebPhotos_private.hh"
#endif

#include <FerrisUI/GtkFerris.hh>
#include <Ferris_private.hh>
#include <Enamel_priv.hh>

#include <string>
#include <iostream>
#include <iomanip>

#include <libferrisldapshared.hh>

#ifdef HAVE_SQLPLUS
#include <mysql++.h>
#include <libferrissqlplusshared.hh>
using namespace mysqlpp;
#endif

#ifdef HAVE_LIBPQXX
#include <pqxx/connection>
#include <pqxx/tablewriter>
#include <pqxx/transaction>
#include <pqxx/nontransaction>
#include <pqxx/tablereader>
#include <pqxx/tablewriter>
#include <pqxx/result>
#include <pqxx/cursor>
#include <libferrispostgresqlshared.hh>
using namespace PGSTD;
using namespace pqxx;
#endif


#ifdef HAVE_DTL
#include <libferrisdtlshared.hh>
using namespace dtl;
#endif

#ifdef HAVE_OBBY
#include <libferrisobbyshared.hh>
#endif

using namespace std;
using namespace FerrisUI;

namespace Ferris
{
    const string PROGRAM_NAME = "ferris-capplet-auth";
    const string DBNAME = FDB_SECURE;

    GtkWidget* gtk_window;
    GtkWidget* gtk_scrolledwindow;
    GtkWidget* gtk_table;
    GtkNotebook*   gtk_notebook;
    GtkTreeStore*  w_treemodel;
    GtkWidget*     w_treeview;
    enum {
        C_SERV=0,
//         C_USER,
//         C_PASS,
        C_COUNT
    };

    GtkTreeViewColumn* w_cols[C_COUNT];

    GtkWidget* sql_username;
    GtkWidget* sql_password;

    GtkWidget* psql_username;
    GtkWidget* psql_password;
    
    GtkWidget* odbc_username;
    GtkWidget* odbc_password;
    
    GtkWidget* ldap_username;
    GtkWidget* ldap_password;
    GtkWidget* ldap_basedn;
    GtkWidget* ldap_lookup_basedn;

    GtkWidget* obby_username;
    GtkWidget* obby_password;
    
    string treestr( GtkTreeIter *iter, int col )
    {
        GtkTreeView*  tv = GTK_TREE_VIEW(w_treeview);
        GtkTreeModel* tm = GTK_TREE_MODEL(w_treemodel);
        GtkTreeStore* ts = GTK_TREE_STORE(w_treemodel);

        GValue value;
        memset( &value, 0, sizeof(value));
        gtk_tree_model_get_value( tm, iter, col, &value );
        string ret = g_value_get_string(&value);
        g_value_unset ( &value );

        return ret;
    }

/************************************************************/
/************************************************************/
/************************************************************/

    GtkWidget* w_flickr_status_label = 0;
    GtkWidget* w_flickr_default_resize_combo = 0;
    GtkWidget* w_flickr_include_ea_is_present = 0;
    GtkWidget* w_flickr_include_ea_and_value = 0;
    GtkWidget* w_flickr_default_public;
    GtkWidget* w_flickr_default_friend;
    GtkWidget* w_flickr_default_family;


    GtkWidget* w_twothreehq_status_label = 0;
    GtkWidget* w_twothreehq_default_resize_combo = 0;
    GtkWidget* w_twothreehq_include_ea_is_present = 0;
    GtkWidget* w_twothreehq_include_ea_and_value = 0;
    GtkWidget* w_twothreehq_default_public;
    GtkWidget* w_twothreehq_default_friend;
    GtkWidget* w_twothreehq_default_family;
    
    
#ifdef HAVE_WEBPHOTOS_API

    fh_webPhotos flickrWebPhotos = 0;
    fh_webPhotos twothreehqWebPhotos = 0;

    void update_webphotos_status( fh_webPhotos wf, GtkWidget* lab )
    {
        LG_WEBPHOTO_D << "haveToken:" << wf->haveToken()
                      << " token:" << wf->getToken() << endl;
        
        if( wf->haveToken() )
        {
            stringstream ss;
            ss << "authentication OK";

            
            string fn = wf->getUserName();
            if( !fn.empty() )
            {
                ss << endl << "username:" << fn;
            }

            gtk_label_set_text( GTK_LABEL(lab), ss.str().c_str() );
        }
        else
        {
            gtk_label_set_text( GTK_LABEL(lab), "authentication needed" );
        }
    }


    
    
    string
    flickr_request_frob( fh_webPhotos wf )
    {
        stringmap_t args;
        args.insert( make_pair( "method", "flickr.auth.getFrob" ) );
        args.insert( make_pair( "api_key", wf->getAPIKey() ) );

        fh_context res = wf->signedCall( "flickr.auth.getFrob", args, true );

        LG_WEBPHOTO_W << "rsp:" << res->isSubContextBound( "rsp" ) << endl;
        LG_WEBPHOTO_W << "have-frob:" << res->isSubContextBound( "frob" ) << endl;

        
        if( res->isSubContextBound( "frob" ) )
        {
            fh_context e = res->getSubContext( "frob" );
            string frob = getStrAttr( e, "content", "" );
            if( !frob.empty() )
                return frob;

            stringstream ss;
            ss << "Error. no frob string!" << endl;
            LG_WEBPHOTO_W << tostr(ss);
            Throw_WebPhotoException( tostr(ss), 0 );
        }
        
        stringstream ss;
        ss << "Error. no frob element!" << endl;
        LG_WEBPHOTO_W << tostr(ss);
        Throw_WebPhotoException( tostr(ss), 0 );
    }

    void authenticate_with_flickr_run_ffox_cb( GtkButton *button, gpointer user_data )
    {
        const char* earl = (const char*)user_data;
        stringstream ss;
        
        fh_runner r = new Runner();
        r->pushCommandLineArg( "firefox" );
        r->pushCommandLineArg( earl );
        r->setSpawnFlags(
            GSpawnFlags(
                G_SPAWN_SEARCH_PATH |
                G_SPAWN_STDERR_TO_DEV_NULL |
                r->getSpawnFlags()));
        r->Run();
    }
    
    void
    authenticate_with_flickr_cb( GtkButton *button, gpointer user_data )
    {
        fh_webPhotos wf = (WebPhotos*)user_data;
        
        try
        {
            
            LG_WEBPHOTO_D << "authenticate_with_flickr_cb() called" << endl;
            string frob = flickr_request_frob( wf );
//            string frob = "dummy";
            
            LG_WEBPHOTO_D << "authenticate_with_flickr_cb() have_frob:" << frob << endl;

            // 4. Create a login link
            {
                stringmap_t args;
                args["perms"] = "write";
                args["frob"] = frob;
                args["api_key"] = wf->getAPIKey();
                
                string md5string = wf->makeSignedString( args );
                args["api_sig"] = md5string;
                string url = wf->getAuthURL();
                stringstream callss;
                callss << url << "?";
                for( stringmap_t::iterator iter = args.begin(); iter != args.end(); ++iter )
                {
                    callss << iter->first << "=" << iter->second << "&";
                }
                string earl = callss.str();

                LG_WEBPHOTO_D << "link for user to validate frob:" << callss.str() << endl;

                GtkWidget *dialog, *label;
   
                dialog = gtk_dialog_new_with_buttons ("Authenticate using browser",
                                                      0,
                                                      GTK_DIALOG_DESTROY_WITH_PARENT,
                                                      GTK_STOCK_OK,
                                                      GTK_RESPONSE_NONE,
                                                      NULL);
                label = gtk_label_new ("You must now verify the frob using your web browser\n"
                                       "once you are done return and confirm this dialog\n"
                                       "Either click the link to run firefox on the URL\n"
                                       "or copy and paste the URL shown below to your browser\n" );
                gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label);
                GtkWidget* linkb = gtk_link_button_new_with_label( earl.c_str(), "click here to run firefox on the URL" );
                gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), linkb);
                gtk_signal_connect(GTK_OBJECT (linkb), "clicked",
                                   GTK_SIGNAL_FUNC(authenticate_with_flickr_run_ffox_cb), (void*)earl.c_str() );

                GtkWidget* e = gtk_entry_new();
                gtk_entry_set_text( GTK_ENTRY( e ), earl.c_str() );
                gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), e);
                
                gtk_widget_show_all (dialog);
                gtk_dialog_run (GTK_DIALOG (dialog));
                gtk_widget_destroy( dialog );
            }

            LG_WEBPHOTO_D << "Dialog closed... attempt to call getToken() starting" << endl;
            // 5. Convert frob to a token
            {
                stringmap_t args;
                args["frob"] = frob;
                fh_context res = wf->signedCall( "flickr.auth.getToken", args, true );

                if( res->isSubContextBound( "auth" ) )
                {
                    // <token>45-76598454353455</token>
                    // <perms>read</perms>
                    // <user nsid="12037949754@N01" username="Bees" fullname="Cal H" />
                    fh_context e = res->getSubContext( "auth" );

                    string token = getStrSubCtx( e, "token", "" );
                    string perms = getStrSubCtx( e, "perms", "" );
                    string nsid = "";
                    string username = "";
                    string fullname = "";
                    
                    if( e->isSubContextBound( "user" ) )
                    {
                        fh_context u = e->getSubContext( "user" );
                        nsid     = getStrAttr( u, "nsid", "" );
                        username = getStrAttr( u, "username", "" );
                        fullname = getStrAttr( u, "fullname", "" );
                    }

                    if( !token.empty() )
                    {
                        wf->setDefaultUsername( username );
                        wf->setUserName( username );
                        wf->setFullName( fullname );
                        wf->setToken( token );
                        LG_WEBPHOTO_D << "Success!" << endl;
                    }
                    else
                    {
                        stringstream ss;
                        ss << "No token could be retrieved from photo service";
                        RunErrorDialog(tostr(ss));
                    }
                }
            }
        }
        catch( exception& e )
        {
            fh_stringstream ss;
            ss << "Error authenticaing with flickr!\n"
               << " e:" << e.what()
               << endl;
            RunErrorDialog(tostr(ss));
        }
        
        if( flickrWebPhotos )
            update_webphotos_status( flickrWebPhotos, w_flickr_status_label );
        if( twothreehqWebPhotos )
            update_webphotos_status( twothreehqWebPhotos, w_twothreehq_status_label );
    }
    
#endif
    
/************************************************************/
/************************************************************/
/************************************************************/


    
static gboolean
my_search_equal_func(GtkTreeModel *model,
                     gint column,
                     const gchar *key,
                     GtkTreeIter *iter,
                     gpointer search_data)
{
    cerr << "key:" << key  << endl;
    string serv = treestr( iter, C_SERV );
    return serv == key;
}

/************************************************************/
/************************************************************/
/************************************************************/
    

    gboolean SaveData_fe( GtkTreeModel *model, GtkTreePath *path,
                          GtkTreeIter *iter, gpointer udata )
    {
        fh_stringstream ss = *((fh_stringstream*)udata);
        GtkTreeView*  tv = GTK_TREE_VIEW( w_treeview);
        GtkTreeModel* tm = GTK_TREE_MODEL(w_treemodel);
        GtkTreeStore* ts = GTK_TREE_STORE(w_treemodel);

        string serv = treestr( iter, C_SERV );
//         string user = treestr( iter, C_USER );
//         string pass = treestr( iter, C_PASS );

        ss << serv << " ";
        {
            fh_stringstream ss;
            ss << "mysql-server-" << serv;
            setEDBString( DBNAME, tostr(ss), "1" );
        }

//         setUserPass( serv, user, pass );
        return 0;
    }

    void SaveData(void)
    {
        GtkTreeView*  tv = GTK_TREE_VIEW( w_treeview);
        GtkTreeModel* tm = GTK_TREE_MODEL(w_treemodel);
        GtkTreeStore* ts = GTK_TREE_STORE(w_treemodel);

        fh_stringstream ss;
        gtk_tree_model_foreach( tm, GtkTreeModelForeachFunc(SaveData_fe), &ss );
        setEDBString( DBNAME, "mysql-server-keys", tostr(ss) );

        GtkTreeSelection* selection = gtk_tree_view_get_selection ( tv );
        gtk_tree_selection_unselect_all( selection );
        
#ifdef HAVE_WEBPHOTOS_API
        {
            if( w_flickr_default_resize_combo )
            {
                string s = gtk_combo_box_get_active_text(
                    GTK_COMBO_BOX( w_flickr_default_resize_combo ));
                if( starts_with( s, "no" ))
                {
                    flickrWebPhotos->setDefaultLargestDimension( 0 );
                }
                else
                {
                    flickrWebPhotos->setDefaultLargestDimension( toint( s ) );
                }
            }

            if( w_flickr_include_ea_is_present )
            {
                string s = gtk_entry_get_text(GTK_ENTRY(w_flickr_include_ea_is_present));
                flickrWebPhotos->setDefaultIncludeEAIsPresentRegex( s );
            }
            if( w_flickr_include_ea_and_value )
            {
                string s = gtk_entry_get_text(GTK_ENTRY(w_flickr_include_ea_and_value));
                flickrWebPhotos->setDefaultIncludeEAandValueRegex( s );
            }

            
            flickrWebPhotos->setDefaultImageProtectionPublic(
                gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(w_flickr_default_public)));
            flickrWebPhotos->setDefaultImageProtectionFriend(
                gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(w_flickr_default_friend)));
            flickrWebPhotos->setDefaultImageProtectionFamily(
                gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(w_flickr_default_family)));
        }


        {
            if( w_twothreehq_default_resize_combo )
            {
                string s = gtk_combo_box_get_active_text(
                    GTK_COMBO_BOX( w_twothreehq_default_resize_combo ));
                if( starts_with( s, "no" ))
                {
                    twothreehqWebPhotos->setDefaultLargestDimension( 0 );
                }
                else
                {
                    twothreehqWebPhotos->setDefaultLargestDimension( toint( s ) );
                }
            }

            if( w_twothreehq_include_ea_is_present )
            {
                string s = gtk_entry_get_text(GTK_ENTRY(w_twothreehq_include_ea_is_present));
                twothreehqWebPhotos->setDefaultIncludeEAIsPresentRegex( s );
            }
            if( w_twothreehq_include_ea_and_value )
            {
                string s = gtk_entry_get_text(GTK_ENTRY(w_twothreehq_include_ea_and_value));
                twothreehqWebPhotos->setDefaultIncludeEAandValueRegex( s );
            }

            
            twothreehqWebPhotos->setDefaultImageProtectionPublic(
                gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(w_twothreehq_default_public)));
            twothreehqWebPhotos->setDefaultImageProtectionFriend(
                gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(w_twothreehq_default_friend)));
            twothreehqWebPhotos->setDefaultImageProtectionFamily(
                gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(w_twothreehq_default_family)));
        }
        
        
#endif
    }

    GtkTreeIter appendNewBlankItem()
    {
        GtkTreeIter iter;
        const char* d = "x";
        gtk_tree_store_append( w_treemodel, &iter, NULL );
        gtk_tree_store_set( w_treemodel, &iter,
                            C_SERV, d,
//                             C_USER, d,
//                             C_PASS, d,
                            -1 );
        return iter;
    }


    void LoadData()
    {
        bool addedany = false;
        GtkTreeIter treeiter;
        int i=0;
        char** p;
        string servdb;
    
        fh_stringstream kss;
        kss << getEDBString( DBNAME, "mysql-server-keys", "" );

        while( getline( kss, servdb, ' ' ))
        { 
            cerr << "loading server-db:" << servdb << endl;
        
            addedany = true;
            gtk_tree_store_append( w_treemodel, &treeiter, NULL );

            string db;
            string serv;
//             userpass_t up = getUserPass( servdb );
//             string user = up.first;
//             string pass = up.second;

            {
                fh_stringstream ss;
                ss << servdb;
                getline( ss, serv, '-' );
                getline( ss, db );
            }
        
        
            gtk_tree_store_set( w_treemodel, &treeiter,
                                C_SERV, serv.c_str(),
//                                 C_USER, user.c_str(),
//                                 C_PASS, pass.c_str(),
                                -1 );
        
        }

        /* bug in gtk2 */
        if( !addedany )
        {
            appendNewBlankItem();
        }
    }

/**********************/

    void add_cb( GtkButton *button, gpointer user_data )
    {
        GtkTreeView*  tv = GTK_TREE_VIEW(w_treeview);
        GtkTreeModel* tm = GTK_TREE_MODEL(w_treemodel);
    
        GtkTreeIter iter = appendNewBlankItem();

        GtkTreePath* path      = gtk_tree_model_get_path( tm, &iter );
        GtkTreeViewColumn* col = gtk_tree_view_get_column( tv, C_SERV );
        // FIXME: gtk2 doesn't like this. it doesn't close off the old cell edit
        gtk_tree_view_set_cursor( tv, path, col, true );
        gtk_tree_path_free( path );
    }


    typedef list< GtkTreeIter* > deliters_t;

    void del_cb_fe( GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer udata)
    {
        deliters_t* x = (deliters_t*)udata;
        x->push_back( gtk_tree_iter_copy( iter ));
    }


    void del_cb( GtkButton *button, gpointer user_data )
    {
        deliters_t deliters;
        GtkTreeView*  tv = GTK_TREE_VIEW(w_treeview);
        GtkTreeModel* tm = GTK_TREE_MODEL(w_treemodel);
        GtkTreeStore* ts = GTK_TREE_STORE(w_treemodel);
        GtkTreeSelection *selection;

        selection = gtk_tree_view_get_selection ( tv );
        gtk_tree_selection_selected_foreach(
            selection, GtkTreeSelectionForeachFunc(del_cb_fe), &deliters );
        gtk_tree_selection_unselect_all( selection );

        for( deliters_t::iterator iter = deliters.begin(); iter != deliters.end(); ++iter )
        {
            gtk_tree_store_remove( ts, *iter );
        }
    }


#ifdef HAVE_SQLPLUS
    void sql_test_fe( GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer udata)
    {
        GtkTreeView*  tv = GTK_TREE_VIEW(w_treeview);
        GtkTreeModel* tm = GTK_TREE_MODEL(w_treemodel);
        GtkTreeStore* ts = GTK_TREE_STORE(w_treemodel);

        string db   = "";
        string serv = treestr( iter, C_SERV );
        string user = gtk_entry_get_text(GTK_ENTRY(sql_username));
        string pass = gtk_entry_get_text(GTK_ENTRY(sql_password));

        try
        {
            Connection con(use_exceptions);
            con.connect( db.c_str(), serv.c_str(), user.c_str(), pass.c_str() );

            fh_stringstream ss;
            ss << "Success connecting to db:" << db
               << " serv:" << serv
               << " user:" << user
               << " pass:" << pass
               << endl;
            RunInfoDialog(tostr(ss));
        }
        catch( exception& e )
        {
            fh_stringstream ss;
            ss << "Error connecting to db:" << db
               << " serv:" << serv
               << " user:" << user
               << " pass:" << pass
               << endl;
            RunErrorDialog(tostr(ss));
        }
    }
#endif

#ifdef HAVE_LIBPQXX
    void psql_test_fe( GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer udata)
    {
        GtkTreeView*  tv = GTK_TREE_VIEW(w_treeview);
        GtkTreeModel* tm = GTK_TREE_MODEL(w_treemodel);
        GtkTreeStore* ts = GTK_TREE_STORE(w_treemodel);

        string db   = "";
        string serv = treestr( iter, C_SERV );
        string user = gtk_entry_get_text(GTK_ENTRY(psql_username));
        string pass = gtk_entry_get_text(GTK_ENTRY(psql_password));

        try
        {
            fh_stringstream ss;
            ss << "host=" << serv << " ";
            if( !user.empty() )
                ss << "user=" << user << " ";
            if( !pass.empty() )
                ss << "password=" << pass << " ";
            if( !db.empty() )
                ss << " dbname=" << db;
            ss << endl;

            connection con( ss.str() );
            fh_stringstream infoss;
            infoss << "Success connecting to db:" << db
                   << " serv:" << serv
                   << " user:" << user
                   << " pass:" << pass
                   << endl;
            RunInfoDialog(tostr(infoss));
        }
        catch( exception& e )
        {
            fh_stringstream ss;
            ss << "Error connecting to db:" << db
               << " serv:" << serv
               << " user:" << user
               << " pass:" << pass
               << " error:" << e.what()
               << endl;
            RunErrorDialog(tostr(ss));
        }
    }
#endif

#ifdef HAVE_DTL
    void odbc_test_fe( GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer udata)
    {
        GtkTreeView*  tv = GTK_TREE_VIEW(w_treeview);
        GtkTreeModel* tm = GTK_TREE_MODEL(w_treemodel);
        GtkTreeStore* ts = GTK_TREE_STORE(w_treemodel);

        string serv = treestr( iter, C_SERV );
        string user = gtk_entry_get_text(GTK_ENTRY(odbc_username));
        string pass = gtk_entry_get_text(GTK_ENTRY(odbc_password));

        try
        {
            fh_stringstream args;
            args << "UID=" << user  << ";"
                 << "PWD=" << pass  << ";"
                 << "DSN=" << serv  << ";"
                 << endl;
            DBConnection con;
            con.Connect( tostr( args ) );

            fh_stringstream ss;
            ss << "Success connecting to"
               << " serv:" << serv
               << " user:" << user
               << " pass:" << pass
               << endl;
            RunInfoDialog(tostr(ss));
        }
        catch( exception& e )
        {
            fh_stringstream ss;
            ss << "Error connecting to"
               << " serv:" << serv
               << " user:" << user
               << " pass:" << pass
               << endl;
            RunErrorDialog(tostr(ss));
        }
    }
#endif
    
    void test_cb( GtkButton *button, gpointer user_data )
    {
        GtkTreeView*  tv = GTK_TREE_VIEW(w_treeview);
        GtkTreeModel* tm = GTK_TREE_MODEL(w_treemodel);
        GtkTreeStore* ts = GTK_TREE_STORE(w_treemodel);
        GtkTreeSelection *selection;

        selection = gtk_tree_view_get_selection ( tv );
        gtk_tree_selection_selected_foreach(
            selection, GtkTreeSelectionForeachFunc(user_data), 0 );
    }

    static void
    cell_edited_cb (GtkCellRendererText *cell, gchar *path_string,
                    gchar *new_text, gpointer udata)
    {
        gint cidx = GPOINTER_TO_INT(udata);
        GtkTreeModel *model = GTK_TREE_MODEL (w_treemodel);
        GtkTreeIter iter;
        GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
        gtk_tree_model_get_iter (model, &iter, path);
        gtk_tree_store_set (GTK_TREE_STORE (model), &iter, cidx, new_text, -1);
        gtk_tree_path_free (path);
    }

    void tv_changed_fe( GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer udata)
    {
        string* sp = (string*)udata;
        *sp = treestr( iter, C_SERV );
    }
    
    void tv_changed_cb( GtkTreeSelection *sel, gpointer )
    {
        static string old;
        string current;
        
        gtk_tree_selection_selected_foreach(
            sel, GtkTreeSelectionForeachFunc(tv_changed_fe), &current  );

        cerr << "tv_changed_cb() old:" << old << " new:" << current << endl;
        
        if( old != current )
        {
            if( old.length() )
            {
                // save the old data
                
#ifdef HAVE_SQLPLUS
                {
                    string username = gtk_entry_get_text(GTK_ENTRY(sql_username));
                    string password = gtk_entry_get_text(GTK_ENTRY(sql_password));
                    setUserPass( old, username, password );
                }
#endif

#ifdef HAVE_LIBPQXX
                {
                    string username = gtk_entry_get_text(GTK_ENTRY(psql_username));
                    string password = gtk_entry_get_text(GTK_ENTRY(psql_password));
                    setPostgreSQLUserPass( old, username, password );
                }
#endif

#ifdef HAVE_DTL
                {
                    string username = gtk_entry_get_text(GTK_ENTRY(odbc_username));
                    string password = gtk_entry_get_text(GTK_ENTRY(odbc_password));
                    setODBCUserPass( old, username, password );
                }
#endif

#ifdef HAVE_LDAP
                {
                    LDAPAuth::LDAPAuthInfo d;
                    d.username      = gtk_entry_get_text(GTK_ENTRY(ldap_username));
                    d.password      = gtk_entry_get_text(GTK_ENTRY(ldap_password));
                    d.basedn        = gtk_entry_get_text(GTK_ENTRY(ldap_basedn));
                    d.lookup_basedn = gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(ldap_lookup_basedn));
                    LDAPAuth::setUserPass( old, d );
                }
#endif
                
#ifdef HAVE_OBBY
                {
                    string username = gtk_entry_get_text(GTK_ENTRY(obby_username));
                    string password = gtk_entry_get_text(GTK_ENTRY(obby_password));
                    setOBBYUserPass( old, username, password );
                }
#endif
            }

            // load the new selection's data

#ifdef HAVE_SQLPLUS
            {
                userpass_t up = getUserPass( current );
                string username = up.first;
                string password = up.second;

                gtk_entry_set_text( GTK_ENTRY(sql_username), username.c_str() );
                gtk_entry_set_text( GTK_ENTRY(sql_password), password.c_str() );
            }
#endif

#ifdef HAVE_LIBPQXX
            {
                userpass_t up = getPostgreSQLUserPass( current );
                string username = up.first;
                string password = up.second;

                gtk_entry_set_text( GTK_ENTRY(psql_username), username.c_str() );
                gtk_entry_set_text( GTK_ENTRY(psql_password), password.c_str() );
            }
#endif

            
#ifdef HAVE_DTL
            {
                userpass_t up = getODBCUserPass( current );
                string username = up.first;
                string password = up.second;

                gtk_entry_set_text( GTK_ENTRY(odbc_username), username.c_str() );
                gtk_entry_set_text( GTK_ENTRY(odbc_password), password.c_str() );
            }
#endif
            
#ifdef HAVE_LDAP
            {
                LDAPAuth::LDAPAuthInfo d = LDAPAuth::getUserPass( current );

                gtk_entry_set_text( GTK_ENTRY(ldap_username), d.username.c_str() );
                gtk_entry_set_text( GTK_ENTRY(ldap_password), d.password.c_str() );
                gtk_entry_set_text( GTK_ENTRY(ldap_basedn), d.basedn.c_str() );
                gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(ldap_lookup_basedn),
                                              d.lookup_basedn );
            }
#endif
            
#ifdef HAVE_OBBY
            {
                userpass_t up = getOBBYUserPass( current );
                string username = up.first;
                string password = up.second;

                gtk_entry_set_text( GTK_ENTRY(obby_username), username.c_str() );
                gtk_entry_set_text( GTK_ENTRY(obby_password), password.c_str() );
            }
#endif
            
            old = current;
        }
    }
    
/**********************/

    void save_and_quit_cb( GtkButton *button, gpointer user_data )
    {
        SaveData();
        gtk_main_quit();
    }

    void quit_cb( GtkButton *button, gpointer user_data )
    {
        gtk_main_quit();
    }


    void make_window()
    {
        GtkWidget* list_scrollwin;
        GtkWidget* sw;
        GtkWidget* b;
        GtkWidget* listarea;

        gtk_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
        gtk_window_set_title ( GTK_WINDOW( gtk_window ), "Ferris authentication" );
        gtk_window_set_default_size (GTK_WINDOW (gtk_window), 600, 450);
        gtk_signal_connect(GTK_OBJECT (gtk_window), "destroy", gtk_main_quit, NULL);
        GtkNotebook* main_widget = GTK_NOTEBOOK(gtk_notebook_new());

        GtkPaned* paned = GTK_PANED(gtk_hpaned_new());
        GtkWidget* tree_frame    = gtk_frame_new (NULL);
        GtkWidget* details_frame = gtk_frame_new (NULL);
        gtk_frame_set_shadow_type (GTK_FRAME (tree_frame),    GTK_SHADOW_IN);
        gtk_frame_set_shadow_type (GTK_FRAME (details_frame), GTK_SHADOW_IN);
        gtk_paned_pack1 (GTK_PANED (paned), tree_frame,    true, false );
        gtk_paned_pack2 (GTK_PANED (paned), details_frame, true, false );
        gtk_frame_set_label( GTK_FRAME (details_frame), "Authentication Information" );
        gtk_notebook_append_page( main_widget, GTK_WIDGET(paned), gtk_label_new( "Servers" ));

        
        int r=0;
    
        {
            GtkToolbar* tb = GTK_TOOLBAR(gtk_toolbar_new());
            
            GtkWidget* vbx = GTK_WIDGET(gtk_vbox_new(0,0));
            GtkWidget* hbx = GTK_WIDGET(gtk_hbox_new(0,0));

            b = gtk_toolbar_append_item( tb, "New",
                                         "create a new server/database connection", "",
                                         gtk_image_new_from_stock(
                                             GTK_STOCK_NEW,
                                             GTK_ICON_SIZE_LARGE_TOOLBAR ),
                                         GTK_SIGNAL_FUNC(add_cb),
                                         0 );

            b = gtk_toolbar_append_item( tb, "Delete",
                                         "delete selected server/database connection(s)", "",
                                         gtk_image_new_from_stock(
                                             GTK_STOCK_DELETE,
                                             GTK_ICON_SIZE_LARGE_TOOLBAR ),
                                         GTK_SIGNAL_FUNC(del_cb),
                                         0 );

//            gtk_container_add(GTK_CONTAINER(hbx), GTK_WIDGET( tb ));
            gtk_container_add(GTK_CONTAINER(hbx), GTK_WIDGET( tb ));
            gtk_box_pack_start(GTK_BOX(vbx), hbx, 0, 0, 0 );
            
            
//             b = gtk_button_new_from_stock( GTK_STOCK_NEW );
//             gtk_container_add(GTK_CONTAINER(hbx), GTK_WIDGET( b ));
//             gtk_signal_connect(GTK_OBJECT (b), "clicked",
//                                GTK_SIGNAL_FUNC(add_cb), NULL);
        
//             b = gtk_button_new_from_stock( GTK_STOCK_DELETE );
//             gtk_container_add(GTK_CONTAINER(hbx), GTK_WIDGET( b ));
//             gtk_signal_connect(GTK_OBJECT (b), "clicked",
//                                GTK_SIGNAL_FUNC(del_cb), NULL);

//             b = gtk_button_new_from_stock( GTK_STOCK_EXECUTE );
//             gtk_container_add(GTK_CONTAINER(hbx), GTK_WIDGET( b ));
//             gtk_signal_connect(GTK_OBJECT (b), "clicked",
//                                GTK_SIGNAL_FUNC(test_cb), NULL);


        
//            gtk_box_pack_start(GTK_BOX(vbx), hbx, 0, 0, 0 );
//            gtk_table_attach_defaults(GTK_TABLE(page), vbx, 0, 1, r, r+1 );

    

            w_treemodel = gtk_tree_store_new( C_COUNT,
                                              G_TYPE_STRING, G_TYPE_STRING,
                                              G_TYPE_STRING, G_TYPE_STRING );
            w_treeview  = gtk_tree_view_new_with_model(GTK_TREE_MODEL( w_treemodel )); 
//    gtk_table_attach_defaults(GTK_TABLE(page), w_treeview, 0, 1, r, r+1 );
//            gtk_box_pack_start(GTK_BOX(vbx), GTK_WIDGET(w_treeview), 1, 1, 0 );
   
            gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (w_treeview), TRUE);
            GObject *selection;
            selection = G_OBJECT (gtk_tree_view_get_selection (GTK_TREE_VIEW (w_treeview)));
//            gtk_tree_selection_set_mode (GTK_TREE_SELECTION (selection), GTK_SELECTION_MULTIPLE);
            gtk_tree_selection_set_mode (GTK_TREE_SELECTION (selection), GTK_SELECTION_SINGLE);
            gtk_tree_view_set_headers_clickable(GTK_TREE_VIEW (w_treeview), true );
            gtk_tree_view_set_enable_search(GTK_TREE_VIEW (w_treeview), true );
            gtk_tree_view_set_search_column(GTK_TREE_VIEW (w_treeview), C_SERV );
            gtk_tree_view_set_search_equal_func(GTK_TREE_VIEW (w_treeview), 
                                                my_search_equal_func, 0, 0 );
            g_signal_connect_data( G_OBJECT( selection ), "changed",
                                   G_CALLBACK (tv_changed_cb), 0, 0, GConnectFlags(0));

            
            /* Setup view */
            GtkCellRenderer* ren;

            ren = gtk_cell_renderer_text_new ();
            g_object_set(ren, "editable", 1, 0 );
            g_signal_connect_data( G_OBJECT( ren ), "edited",
                                   G_CALLBACK (cell_edited_cb),
                                   GINT_TO_POINTER(C_SERV), 0, GConnectFlags(0));
    
            w_cols[C_SERV] = gtk_tree_view_column_new_with_attributes( "Server", ren,
                                                                       "text", C_SERV,
                                                                       NULL);
            gtk_tree_view_append_column( GTK_TREE_VIEW( w_treeview ), w_cols[C_SERV] );
//            gtk_tree_view_column_set_sizing ( w_cols[C_SERV], GTK_TREE_VIEW_COLUMN_RESIZABLE );
            gtk_tree_view_column_set_sort_column_id( w_cols[C_SERV], C_SERV );

//             ren = gtk_cell_renderer_text_new ();
//             g_object_set(ren, "editable", 1, 0 );
//             g_signal_connect_data( G_OBJECT( ren ), "edited",
//                                    G_CALLBACK (cell_edited_cb),
//                                    (gpointer)C_USER, 0, GConnectFlags(0));

//             w_cols[C_USER] = gtk_tree_view_column_new_with_attributes( "Username", ren,
//                                                                        "text", C_USER,
//                                                                        NULL);
//             gtk_tree_view_append_column( GTK_TREE_VIEW( w_treeview ), w_cols[C_USER] );
// //            gtk_tree_view_column_set_sizing ( w_cols[C_USER], GTK_TREE_VIEW_COLUMN_RESIZABLE );
//             gtk_tree_view_column_set_sort_column_id( w_cols[C_USER], C_USER );

//             ren = gtk_cell_renderer_text_new ();
//             g_object_set(ren, "editable", 1, 0 );
//             g_signal_connect_data( G_OBJECT( ren ), "edited",
//                                    G_CALLBACK (cell_edited_cb),
//                                    (gpointer)C_PASS, 0, GConnectFlags(0));
    
//             w_cols[C_PASS] = gtk_tree_view_column_new_with_attributes( "Password", ren,
//                                                                        "text", C_PASS,
//                                                                        NULL);
//             gtk_tree_view_append_column( GTK_TREE_VIEW( w_treeview ), w_cols[C_PASS] );
// //            gtk_tree_view_column_set_sizing ( w_cols[C_PASS], GTK_TREE_VIEW_COLUMN_RESIZABLE );
//             gtk_tree_view_column_set_sort_column_id( w_cols[C_PASS], C_PASS );
    
    
    
    
    

//            list_scrollwin = sw = gtk_scrolled_window_new(NULL, NULL);
//            gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sw), GTK_WIDGET( page ));

            sw = gtk_scrolled_window_new(NULL, NULL);
            gtk_container_add( GTK_CONTAINER( sw ), GTK_WIDGET( w_treeview ));
            
            gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
                                            GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
            
            gtk_box_pack_start(GTK_BOX(vbx), GTK_WIDGET(sw), 1, 1, 0 );
            gtk_container_add( GTK_CONTAINER( tree_frame ), vbx );
            
        }


#ifdef HAVE_WEBPHOTOS_API
        {
            GtkPaned* paned = GTK_PANED(gtk_hpaned_new());
            gtk_notebook_append_page( main_widget, GTK_WIDGET(paned), gtk_label_new( "Photo servers" ));

            GtkNotebook* photoservers = GTK_NOTEBOOK(gtk_notebook_new());
            gtk_paned_pack1 (GTK_PANED (paned), GTK_WIDGET(photoservers), true, false );

            int r=0;
            int c=0;
            GtkWidget* l;
            GtkWidget* e;
            GtkWidget* page = 0;

            r=0; c=0;
//            page = gtk_table_new ( 2, 2, false );
            GtkNotebook* flickrpage = GTK_NOTEBOOK(gtk_notebook_new());
            gtk_notebook_append_page( photoservers,
                                      GTK_WIDGET(flickrpage),
                                      gtk_label_new( "flickr" ));

            fh_webPhotos wf = flickrWebPhotos = Factory::getDefaultFlickrWebPhotos();
            if( wf->haveAPIKey() )
            {
                {
                    page = gtk_table_new ( 2, 2, false );
                    gtk_notebook_append_page( GTK_NOTEBOOK(flickrpage), page,
                                              gtk_label_new( "auth" ));
                    
                    b = gtk_button_new_with_label("Authenticate with flickr");
                    gtk_table_attach_defaults(GTK_TABLE(page), b, c, c+2, r, r+1 );
                    ++c;
                    gtk_signal_connect(GTK_OBJECT (b), "clicked",
                                       GTK_SIGNAL_FUNC(authenticate_with_flickr_cb), GetImpl(wf) );
                    ++r; c=0;
                    
                    l = gtk_label_new ("Status" );
                    gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
                    ++c;
                    l = gtk_label_new ("" );
                    w_flickr_status_label = l;
                    gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
                }
                
                {
                    string s;
                    
                    page = gtk_table_new ( 2, 2, false );
                    gtk_notebook_append_page( GTK_NOTEBOOK(flickrpage), page,
                                              gtk_label_new( "settings" ));
                    r=0; c=0;

                    l = gtk_label_new("Default include EA is Present Regex");
                    gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
                    ++c;
                    w_flickr_include_ea_is_present = e = gtk_entry_new();
                    gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
                    s = flickrWebPhotos->getDefaultIncludeEAIsPresentRegexString();
                    gtk_entry_set_text( GTK_ENTRY(e), s.c_str() );
                    c=0; ++r; 

                    l = gtk_label_new("Default include EA and Value Regex");
                    gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
                    ++c;
                    w_flickr_include_ea_and_value = e = gtk_entry_new();
                    gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
                    s = flickrWebPhotos->getDefaultIncludeEAandValueRegexString();
                    gtk_entry_set_text( GTK_ENTRY(e), s.c_str() );
                    c=0; ++r; 


                    e = gtk_check_button_new_with_label("default permissions public");
                    w_flickr_default_public = e;
                    gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+2, r, r+1 );
                    gtk_toggle_button_set_active(
                        GTK_TOGGLE_BUTTON(e), flickrWebPhotos->isDefaultImageProtectionPublic() );
                    c=0; ++r; 

                    e = gtk_check_button_new_with_label("default permissions friend");
                    w_flickr_default_friend = e;
                    gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+2, r, r+1 );
                    gtk_toggle_button_set_active(
                        GTK_TOGGLE_BUTTON(e), flickrWebPhotos->isDefaultImageProtectionFriend() );
                    c=0; ++r; 

                    e = gtk_check_button_new_with_label("default permissions family");
                    w_flickr_default_family = e;
                    gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+2, r, r+1 );
                    gtk_toggle_button_set_active(
                        GTK_TOGGLE_BUTTON(e), flickrWebPhotos->isDefaultImageProtectionFamily() );
                    c=0; ++r; 
                    
                    

                    l = gtk_label_new ("Default upload resize" );
                    gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
                    ++c;
                    GtkWidget* w = gtk_combo_box_new_text();
                    w_flickr_default_resize_combo = w;
                    gtk_combo_box_append_text( GTK_COMBO_BOX(w), "no resize" );
                    gtk_combo_box_append_text( GTK_COMBO_BOX(w), "320" );
                    gtk_combo_box_append_text( GTK_COMBO_BOX(w), "640" );
                    gtk_combo_box_append_text( GTK_COMBO_BOX(w), "800" );
                    gtk_combo_box_append_text( GTK_COMBO_BOX(w), "1024" );
                    gtk_combo_box_append_text( GTK_COMBO_BOX(w), "1280" );
                    gtk_combo_box_append_text( GTK_COMBO_BOX(w), "1600" );
                    gtk_combo_box_append_text( GTK_COMBO_BOX(w), "1920" );
                    gtk_table_attach_defaults(GTK_TABLE(page), w, c, c+1, r, r+1 );
                
                    {
                        gint index = -1;
                        long v = wf->getDefaultLargestDimension();
                        switch( wf->getDefaultLargestDimension() )
                        {
                        case 0:    index = 0; break;
                        case 320:  index = 1; break;
                        case 640:  index = 2; break;
                        case 800:  index = 3; break;
                        case 1024: index = 4; break;
                        case 1280: index = 5; break;
                        case 1600: index = 6; break;
                        case 1920: index = 7; break;
                        }
                        gtk_combo_box_set_active( GTK_COMBO_BOX(w), index );
                    }
                    ++r; c=0;
                }
                
                update_webphotos_status( wf, w_flickr_status_label );
            }
            else
            {
                l = gtk_label_new("No flickr API and secret defined.\n"
                                  "Please obtain an API key and shared secret from:\n"
                                  "http://www.flickr.com/services/api/auth.howto.desktop.html\n"
                                  "and place them in ~/.ferris/flickr-api-key.txt and flickr-api-secret.txt"
                    );
                gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+2, r, r+1 );
                ++c;
            }

            /********************************************************************************/
            /********************************************************************************/
            /********************************************************************************/
            
            {
                r=0; c=0;
                GtkNotebook* twothreehqpage = GTK_NOTEBOOK(gtk_notebook_new());
                gtk_notebook_append_page( photoservers,
                                          GTK_WIDGET(twothreehqpage),
                                          gtk_label_new( "23hq" ));

                fh_webPhotos wf = twothreehqWebPhotos = Factory::getDefault23hqWebPhotos();
                if( wf->haveAPIKey() )
                {
                    {
                        page = gtk_table_new ( 2, 2, false );
                        gtk_notebook_append_page( GTK_NOTEBOOK(twothreehqpage), page,
                                                  gtk_label_new( "auth" ));
                    
                        b = gtk_button_new_with_label("Authenticate with 23hq");
                        gtk_table_attach_defaults(GTK_TABLE(page), b, c, c+2, r, r+1 );
                        ++c;
                        gtk_signal_connect(GTK_OBJECT (b), "clicked",
                                           GTK_SIGNAL_FUNC(authenticate_with_flickr_cb), GetImpl(wf) );
                        ++r; c=0;
                    
                        l = gtk_label_new ("Status" );
                        gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
                        ++c;
                        l = gtk_label_new ("" );
                        w_twothreehq_status_label = l;
                        gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
                    }
                
                    {
                        string s;
                    
                        page = gtk_table_new ( 2, 2, false );
                        gtk_notebook_append_page( GTK_NOTEBOOK(twothreehqpage), page,
                                                  gtk_label_new( "settings" ));
                        r=0; c=0;

                        l = gtk_label_new("Default include EA is Present Regex");
                        gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
                        ++c;
                        w_twothreehq_include_ea_is_present = e = gtk_entry_new();
                        gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
                        s = twothreehqWebPhotos->getDefaultIncludeEAIsPresentRegexString();
                        gtk_entry_set_text( GTK_ENTRY(e), s.c_str() );
                        c=0; ++r; 

                        l = gtk_label_new("Default include EA and Value Regex");
                        gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
                        ++c;
                        w_twothreehq_include_ea_and_value = e = gtk_entry_new();
                        gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
                        s = twothreehqWebPhotos->getDefaultIncludeEAandValueRegexString();
                        gtk_entry_set_text( GTK_ENTRY(e), s.c_str() );
                        c=0; ++r; 


                        e = gtk_check_button_new_with_label("default permissions public");
                        w_twothreehq_default_public = e;
                        gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+2, r, r+1 );
                        gtk_toggle_button_set_active(
                            GTK_TOGGLE_BUTTON(e), twothreehqWebPhotos->isDefaultImageProtectionPublic() );
                        c=0; ++r; 

                        e = gtk_check_button_new_with_label("default permissions friend");
                        w_twothreehq_default_friend = e;
                        gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+2, r, r+1 );
                        gtk_toggle_button_set_active(
                            GTK_TOGGLE_BUTTON(e), twothreehqWebPhotos->isDefaultImageProtectionFriend() );
                        c=0; ++r; 

                        e = gtk_check_button_new_with_label("default permissions family");
                        w_twothreehq_default_family = e;
                        gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+2, r, r+1 );
                        gtk_toggle_button_set_active(
                            GTK_TOGGLE_BUTTON(e), twothreehqWebPhotos->isDefaultImageProtectionFamily() );
                        c=0; ++r; 
                    
                    

                        l = gtk_label_new ("Default upload resize" );
                        gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
                        ++c;
                        GtkWidget* w = gtk_combo_box_new_text();
                        w_twothreehq_default_resize_combo = w;
                        gtk_combo_box_append_text( GTK_COMBO_BOX(w), "no resize" );
                        gtk_combo_box_append_text( GTK_COMBO_BOX(w), "320" );
                        gtk_combo_box_append_text( GTK_COMBO_BOX(w), "640" );
                        gtk_combo_box_append_text( GTK_COMBO_BOX(w), "800" );
                        gtk_combo_box_append_text( GTK_COMBO_BOX(w), "1024" );
                        gtk_combo_box_append_text( GTK_COMBO_BOX(w), "1280" );
                        gtk_combo_box_append_text( GTK_COMBO_BOX(w), "1600" );
                        gtk_combo_box_append_text( GTK_COMBO_BOX(w), "1920" );
                        gtk_table_attach_defaults(GTK_TABLE(page), w, c, c+1, r, r+1 );
                
                        {
                            gint index = -1;
                            long v = wf->getDefaultLargestDimension();
                            switch( wf->getDefaultLargestDimension() )
                            {
                            case 0:    index = 0; break;
                            case 320:  index = 1; break;
                            case 640:  index = 2; break;
                            case 800:  index = 3; break;
                            case 1024: index = 4; break;
                            case 1280: index = 5; break;
                            case 1600: index = 6; break;
                            case 1920: index = 7; break;
                            }
                            gtk_combo_box_set_active( GTK_COMBO_BOX(w), index );
                        }
                        ++r; c=0;
                    }
                
                    update_webphotos_status( wf, w_twothreehq_status_label );
                }
                else
                {
                    l = gtk_label_new("No twothreehq API and secret defined.\n"
                                      "Please obtain an API key and shared secret from:\n"
                                      "http://www.23hq.com/\n"
                                      "and place them in ~/.ferris/23hq-api-key.txt and 23hq-api-secret.txt"
                        );
                    gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+2, r, r+1 );
                    ++c;
                }
            }
        }
#endif
        
    
        /* ok / cancel part */

        GtkWidget* vbx = GTK_WIDGET(gtk_vbox_new(0,0));
        GtkWidget* hbx = GTK_WIDGET(gtk_hbox_new(0,0));

        b = gtk_button_new_from_stock( GTK_STOCK_OK );
        gtk_container_add(GTK_CONTAINER(hbx), GTK_WIDGET( b ));
        gtk_signal_connect(GTK_OBJECT (b), "clicked",
                           GTK_SIGNAL_FUNC(save_and_quit_cb), NULL);

        b = gtk_button_new_from_stock( GTK_STOCK_CANCEL );
        gtk_container_add(GTK_CONTAINER(hbx), GTK_WIDGET( b ));
        gtk_signal_connect(GTK_OBJECT (b), "clicked",
                           GTK_SIGNAL_FUNC(quit_cb), NULL);

        
        // Many pages to handle the different auth info for each server type
        gtk_notebook = GTK_NOTEBOOK(gtk_notebook_new());

#ifdef HAVE_SQLPLUS
        // sqlplus page
        {
            int r=0;
            int c=0;
            GtkWidget* l;
            GtkWidget* e;
            GtkWidget* page = gtk_table_new ( 2, 2, false );

            l = gtk_label_new("username");
            gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
            ++c;
            sql_username = e = gtk_entry_new();
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
            c=0; ++r; 

            l = gtk_label_new("password");
            gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
            ++c;
            sql_password = e = gtk_entry_new();
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
            c=0; ++r; 

            GtkWidget* v = GTK_WIDGET(gtk_vbox_new(0,0));

            GtkToolbar* tb = GTK_TOOLBAR(gtk_toolbar_new());
            b = gtk_toolbar_append_item( tb, "Test",
                                         "try to connect to selected server/databases", "",
                                         gtk_image_new_from_stock(
                                             GTK_STOCK_EXECUTE,
                                             GTK_ICON_SIZE_LARGE_TOOLBAR ),
                                         GTK_SIGNAL_FUNC(test_cb),
                                         (gpointer)sql_test_fe );
            
            gtk_box_pack_start(GTK_BOX(v), GTK_WIDGET(tb), 0, 0, 0 );
            gtk_box_pack_start(GTK_BOX(v), page, 0, 0, 0 );
            gtk_box_pack_start(GTK_BOX(v), gtk_label_new(""), 1, 1, 0 );
            gtk_notebook_append_page( gtk_notebook, v, gtk_label_new( "sqlplus" ));
        }
#endif

#ifdef HAVE_LIBPQXX
        {
            int r=0;
            int c=0;
            GtkWidget* l;
            GtkWidget* e;
            GtkWidget* page = gtk_table_new ( 2, 2, false );

            l = gtk_label_new("username");
            gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
            ++c;
            psql_username = e = gtk_entry_new();
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
            c=0; ++r; 

            l = gtk_label_new("password");
            gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
            ++c;
            psql_password = e = gtk_entry_new();
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
            c=0; ++r; 

            GtkWidget* v = GTK_WIDGET(gtk_vbox_new(0,0));

            GtkToolbar* tb = GTK_TOOLBAR(gtk_toolbar_new());
            b = gtk_toolbar_append_item( tb, "Test",
                                         "try to connect to selected server/databases", "",
                                         gtk_image_new_from_stock(
                                             GTK_STOCK_EXECUTE,
                                             GTK_ICON_SIZE_LARGE_TOOLBAR ),
                                         GTK_SIGNAL_FUNC(test_cb),
                                         (gpointer)psql_test_fe );
            
            gtk_box_pack_start(GTK_BOX(v), GTK_WIDGET(tb), 0, 0, 0 );
            gtk_box_pack_start(GTK_BOX(v), page, 0, 0, 0 );
            gtk_box_pack_start(GTK_BOX(v), gtk_label_new(""), 1, 1, 0 );
            gtk_notebook_append_page( gtk_notebook, v, gtk_label_new( "postgresql" ));
        }
#endif
        
#ifdef HAVE_DTL
        // odbc page
        {
            int r=0;
            int c=0;
            GtkWidget* l;
            GtkWidget* e;
            GtkWidget* page = gtk_table_new ( 2, 2, false );

            l = gtk_label_new("username");
            gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
            ++c;
            odbc_username = e = gtk_entry_new();
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
            c=0; ++r; 

            l = gtk_label_new("password");
            gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
            ++c;
            odbc_password = e = gtk_entry_new();
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
            c=0; ++r; 

            GtkWidget* v = GTK_WIDGET(gtk_vbox_new(0,0));

            GtkToolbar* tb = GTK_TOOLBAR(gtk_toolbar_new());
            b = gtk_toolbar_append_item( tb, "Test",
                                         "try to connect to selected server/databases", "",
                                         gtk_image_new_from_stock(
                                             GTK_STOCK_EXECUTE,
                                             GTK_ICON_SIZE_LARGE_TOOLBAR ),
                                         GTK_SIGNAL_FUNC(test_cb),
                                         (gpointer)odbc_test_fe );
            
            gtk_box_pack_start(GTK_BOX(v), GTK_WIDGET(tb), 0, 0, 0 );
            gtk_box_pack_start(GTK_BOX(v), page, 0, 0, 0 );
            gtk_box_pack_start(GTK_BOX(v), gtk_label_new(""), 1, 1, 0 );
            gtk_notebook_append_page( gtk_notebook, v, gtk_label_new( "odbc" ));
        }
#endif
        
#ifdef HAVE_LDAP
        // ldap page
        {
            int r=0;
            int c=0;
            GtkWidget* l;
            GtkWidget* e;

            GtkWidget* page = gtk_table_new ( 2, 2, false );

            l = gtk_label_new("username");
            gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
            ++c;
            ldap_username = e = gtk_entry_new();
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
            c=0; ++r; 

            l = gtk_label_new("password");
            gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
            ++c;
            ldap_password = e = gtk_entry_new();
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
            c=0; ++r; 
            
            l = gtk_label_new("basedn");
            gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
            ++c;
            ldap_basedn = e = gtk_entry_new();
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
            c=0; ++r; 

            e = ldap_lookup_basedn = gtk_check_button_new_with_label("lookup basedn from server");
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+2, r, r+1 );
            c=0; ++r; 

            GtkWidget* v = GTK_WIDGET(gtk_vbox_new(0,0));
            gtk_box_pack_start(GTK_BOX(v), page, 0, 0, 0 );
            gtk_box_pack_start(GTK_BOX(v), gtk_label_new(""), 1, 1, 0 );
            gtk_notebook_append_page( gtk_notebook, v, gtk_label_new( "ldap" ));
        }
#endif

#ifdef HAVE_OBBY
        // obby page
        {
            int r=0;
            int c=0;
            GtkWidget* l;
            GtkWidget* e;

            GtkWidget* page = gtk_table_new ( 2, 2, false );

            l = gtk_label_new("username");
            gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
            ++c;
            obby_username = e = gtk_entry_new();
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
            c=0; ++r; 

            l = gtk_label_new("password");
            gtk_table_attach_defaults(GTK_TABLE(page), l, c, c+1, r, r+1 );
            ++c;
            obby_password = e = gtk_entry_new();
            gtk_table_attach_defaults(GTK_TABLE(page), e, c, c+1, r, r+1 );
            c=0; ++r; 
            
            GtkWidget* v = GTK_WIDGET(gtk_vbox_new(0,0));
            gtk_box_pack_start(GTK_BOX(v), page, 0, 0, 0 );
            gtk_box_pack_start(GTK_BOX(v), gtk_label_new(""), 1, 1, 0 );
            gtk_notebook_append_page( gtk_notebook, v, gtk_label_new( "obby" ));
        }
#endif        
        gtk_container_add( GTK_CONTAINER( details_frame ), GTK_WIDGET(gtk_notebook) );
        
//         GtkWidget* topbox = GTK_WIDGET(gtk_hbox_new(0,0));
//         gtk_box_pack_start(GTK_BOX(topbox), list_scrollwin, 1, 1, 0 );
//         gtk_box_pack_start(GTK_BOX(topbox), GTK_WIDGET(gtk_notebook), 0, 0, 0 );
        
//         gtk_box_pack_start(GTK_BOX(vbx), topbox, 1, 1, 0 );
//         gtk_box_pack_start(GTK_BOX(vbx), hbx, 0, 0, 0 );

//         gtk_container_add(GTK_CONTAINER(gtk_window), vbx);


        gtk_box_pack_start(GTK_BOX(vbx), GTK_WIDGET(main_widget), 1, 1, 0 );
        gtk_box_pack_start(GTK_BOX(vbx), GTK_WIDGET(hbx), 0, 0, 0 );
        gtk_container_add(GTK_CONTAINER(gtk_window), vbx);
        
    }

    int kickit( int argc, char** argv )
    {
        poptContext optCon;

        const char* copt   = "copt";
        unsigned long iopt = 0;

        try
        {
            struct poptOption optionsTable[] = {
//             { "root-context-class", 0, POPT_ARG_STRING, &RootContextClass, 0,
//               "Name of the class that handles reading the root context", "Native" },
            
                FERRIS_POPT_OPTIONS
                POPT_AUTOHELP
                POPT_TABLEEND
            };

            gtk_init( &argc, &argv );

            optCon = poptGetContext(PROGRAM_NAME.c_str(), argc, (const char**)argv, optionsTable, 0);
            poptSetOtherOptionHelp(optCon, "[OPTIONS]* RootName1 ...");

            if (argc < 1) {
//          poptPrintHelp(optCon, stderr, 0);
                poptPrintUsage(optCon, stderr, 0);
                exit(1);
            }


            char c=-1;
            while ((c = poptGetNextOpt(optCon)) >= 0)
            {
            }

//        gtk_idle_add( fire_ls_cb, (void*)&ls );
            make_window();
            LoadData();
            gtk_widget_show_all( gtk_window );
            gtk_main();
        }
        catch( exception& e )
        {
            cerr << PROGRAM_NAME << ": cought e:" << e.what() << endl;
            exit(1);
        }
        
    
        poptFreeContext(optCon);
        return 0;
    }


};

int main( int argc, char** argv )
{
    return Ferris::kickit( argc, argv );
}
 

