/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2005 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: libferrisxwin_klipper.cpp,v 1.3 2006/08/13 11:41:37 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include "libferrisxwin_klipper.hh"
#include <qstringlist.h>

namespace Ferris
{
    /******************************************************************************/
    /******************************************************************************/
    /******************************************************************************/

    /**
     */
    class FERRISEXP_CTXPLUGIN klipContext
        :
        public StateLessEAHolder< klipContext, leafContext >
    {
        typedef klipContext                                   _Self;
        typedef StateLessEAHolder< klipContext, leafContext > _Base;

        DCOPClient* getDCOP()
            {
                klipperTopDirectoryContext* p = (klipperTopDirectoryContext*)getParent();
                return p->getDCOP();
            }

        DCOPRef getKlipper()
            {
                DCOPClient* dc = getDCOP();
                DCOPRef klips("klipper", "klipper");
                klips.setDCOPClient( dc );
                return klips;
            }
        
    protected:
        virtual std::string priv_getRecommendedEA()
            {
                return adjustRecommendedEAForDotFiles(this, "name,content");
            }

        virtual void createStateLessAttributes( bool force = false )
            {
                static Util::SingleShot virgin;
                if( virgin() )
                {
                    _Base::createStateLessAttributes( true );
                }
            }

        ferris_ios::openmode getSupportedOpenModes()
            {
                return
                    ios_base::in        |
                    ios_base::out       |
                    ios_base::trunc     |
                    ios_base::binary    ;
            }

        

        virtual fh_istream  priv_getIStream( ferris_ios::openmode m )
            throw (FerrisParentNotSetError,
                   CanNotGetStream,
                   exception)
            {
                fh_stringstream ret;

                int i = toint( getDirName() );
                DCOPRef klips = getKlipper();
                QString qs = klips.call("getClipboardHistoryItem(int)", i);
                ret << qs;
                return ret;
            }

        virtual fh_iostream priv_getIOStream( ferris_ios::openmode m )
            throw (FerrisParentNotSetError,
                   AttributeNotWritable,
                   CanNotGetStream,
                   std::exception)
            {
                fh_stringstream ret;
                int i = toint( getDirName() );
                if( i )
                {
                    stringstream ss;
                    ss << "Can only write to most recent clipboard entry (having name == 0)" << endl;
                    Throw_CanNotGetStream( ss.str(), this );
                }

                if( !(m & ios_base::trunc) )
                {
                    DCOPRef klips = getKlipper();
                    QString qs = klips.call("getClipboardHistoryItem(int)", i);
                    ret << qs;
                }
                ret->getCloseSig().connect( bind( SigC::slot( *this, &_Self::OnStreamClosed ), m ));
                return ret;
            }
        
        virtual void OnStreamClosed( fh_istream& ss, std::streamsize tellp, ferris_ios::openmode m )
            {
                AdjustForOpenMode_Closing( ss, m, tellp );
                const std::string s = StreamToString(ss);

                cerr << "OnStreamClosed() s:" << s << endl;
                QString qs = s.c_str();
                DCOPRef klips = getKlipper();
                klips.call("setClipboardContents(QString)", qs );
            }
        
        
    public:
        
        klipContext( Context* parent, const std::string& rdn )
            :
            _Base( parent, rdn )
            {
                createStateLessAttributes( true );
            }
        
        virtual ~klipContext()
            {
            }
    };
    FERRIS_SMARTPTR( klipContext, fh_klipContext );

    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/

    DCOPClient*
    klipperTopDirectoryContext::getDCOP()
    {
        if( !m_dcop )
        {
            m_dcop = new DCOPClient();
            m_dcop->attach();
            m_dcop->registerAs( "libferris" );
        }
        
        return m_dcop;
    }
    
    
    void
    klipperTopDirectoryContext::priv_read()
    {
        emitExistsEventForEachItemRAII _raii1( this );
        clearContext();
        
        DCOPClient* dc = getDCOP();
        DCOPRef klips("klipper", "klipper");
        klips.setDCOPClient( dc );

        int i = 0;
        QStringList qsl =klips.call("getClipboardHistoryMenu()");
        for( QStringList::const_iterator qsi = qsl.begin(); qsi!=qsl.end(); ++qsi, ++i )
        {
//            string content = *qsi;
            string rdn = tostr(i);

            klipContext* cc = 0;
            cc = priv_ensureSubContext( rdn, cc );
        }
    }
        
    klipperTopDirectoryContext::klipperTopDirectoryContext( Context* parent, const std::string& rdn )
        :
        _Base( parent, rdn ),
        m_dcop( 0 )
    {
        createStateLessAttributes();
    }
    klipperTopDirectoryContext::klipperTopDirectoryContext()
        :
        m_dcop( 0 )
    {
        createStateLessAttributes();
        createAttributes();
    }
    
    klipperTopDirectoryContext::~klipperTopDirectoryContext()
    {
    }
    std::string
    klipperTopDirectoryContext::priv_getRecommendedEA()
    {
        return adjustRecommendedEAForDotFiles(this, "name");
    }
    std::string
    klipperTopDirectoryContext::getRecommendedEA()
    {
        return adjustRecommendedEAForDotFiles(this, "name");
    }
    
    klipperTopDirectoryContext*
    klipperTopDirectoryContext::priv_CreateContext( Context* parent, string rdn )
    {
        klipperTopDirectoryContext* ret = new klipperTopDirectoryContext();
        ret->setContext( parent, rdn );
        return ret;
    }
        
    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/
};
