/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2002 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: libcreationtdb.cpp,v 1.1 2005/07/12 04:21:16 ben Exp $

    *******************************************************************************
    *******************************************************************************
    ******************************************************************************/

#include <FerrisCreationPlugin.hh>

#ifdef FERRIS_HAVE_TDB
#include <tdb.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

using namespace std;

namespace Ferris
{

    class CreationStatelessFunctorTDB
        :
        public CreationStatelessFunctor
    {
    public:
        virtual fh_context create( fh_context c, fh_context md );
    };

    extern "C"
    {
        fh_CreationStatelessFunctor
        Create()
        {
            return new CreationStatelessFunctorTDB();
        }
    };

    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/
    
    fh_context
    CreationStatelessFunctorTDB::create( fh_context c, fh_context md )
    {
#ifdef FERRIS_HAVE_TDB
    
        fh_context newc = SubCreate_file( c, md );
        LG_FERRISCREATE_D << "Create tdb at newc->getDirPath:" << newc->getURL() << endl;

        mode_t mode = getModeFromMetaData( md );
        mode_t oldumask = 0;
        bool ignoreUMask = toint(getStrSubCtx( md, "ignore-umask", "0" ));

        if( ignoreUMask ) oldumask = umask( 0 );
        TDB_CONTEXT* db = tdb_open( (char*)newc->getDirPath().c_str(), 0,
                                    TDB_NOLOCK | TDB_CLEAR_IF_FIRST,
                                    O_RDWR | O_CREAT | O_TRUNC | O_LARGEFILE, mode );
        if( ignoreUMask ) umask( oldumask );

        if( !db )
        {
            fh_stringstream ss;
            ss << "SL_SubCreate_dir() error creating tdb file"
               << " URL:" << newc->getURL()
               << endl;
            Throw_FerrisCreateSubContextFailed( tostr(ss), GetImpl(c) );
        }

        tdb_close( db );
        return newc;
    
#endif        

        fh_stringstream ss;
        ss << "SL_SubCreate() should not have gotten here, you have no "
           << "tdb at compile time"
           << " url:" << c->getURL()
           << endl;
        Throw_FerrisCreateSubContextNotSupported( tostr(ss), GetImpl(c) );
    }
};
