/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2004 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: libferrisredlandeashared.hh,v 1.2 2005/12/30 14:51:56 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#ifndef _ALREADY_INCLUDED_FERRIS_REDLAND_EA_SHARED_H_
#define _ALREADY_INCLUDED_FERRIS_REDLAND_EA_SHARED_H_

#include <FerrisEAPlugin.hh>
#include <Ferris/Redland_private.hh>

using namespace std;

namespace Ferris
{
    using namespace RDF;

    

    class RedlandByteArrayAttribute
        :
        public EA_Atom_ReadWrite
    {

    protected:

        /************************************************************/
        /*** methods that subclasses will override ******************/
        /************************************************************/

        /**
         * Get the RDF model that we are working with
         * default is to use getDefaultFerrisModel().
         */
        virtual RDF::fh_model getModel( Context* c );
        /**
         * This method allows attributes to have a URI prefix appended
         * to them automatically. You could override this and return rdn
         * to make no extra prefixing occur.
         * Warning the following use redlandea::getPredicateURI() directly:
         * Create()
         */
        virtual std::string getPredicateURI( const std::string& rdn );
        

        /************************************************************/
        /************************************************************/
        /************************************************************/
        
        ferris_ios::openmode getSupportedOpenModes();
        void createOnDiskAttribute( const fh_context& parent, const std::string& uri );
    
    public:

        fh_iostream priv_getStream( Context* c, const std::string& rdn,
                               fh_model m, fh_node subj, fh_node pred, EA_Atom* atom );
        virtual fh_iostream getStream( Context* c, const std::string& rdn, EA_Atom* atom );
        virtual void setStream( Context* c, const std::string& rdn, EA_Atom* atom, fh_istream ss );

        RedlandByteArrayAttribute(
            const fh_context& parent,
            const string& rdn,
            const string& uri,
            bool forceCreate = false
            );
        virtual ~RedlandByteArrayAttribute();

        static RedlandByteArrayAttribute* Create( const fh_context& parent,
                                                  const string& rdn,
                                                  bool forceCreate = false );
    };
 
};

#endif
