/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: libftxidxclucene.cpp,v 1.4 2006/12/07 06:49:43 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/


#include <Ferris/FullTextIndexerMetaInterface.hh>
#include <Ferris/Ferris.hh>
#include <Ferris/FullTextIndexer_private.hh>
#include <Ferris/FerrisCLucene.hh>

#include <CLucene.h>
#include <CLucene/util/stringreader.h>
#include <CLucene/util/Reader.h>
#include <CLucene/util/StringBuffer.h>

//typedef wchar_t TCHAR;
typedef char char_t;

using namespace std;

using namespace lucene;
using namespace lucene::index;
using namespace lucene::document;
using namespace lucene::analysis;
using namespace lucene::analysis::standard;
using namespace lucene::search;
using namespace lucene::util;
using namespace lucene::queryParser;


namespace Ferris
{
//     wstring widen( const string& s )
//     {
//         wstring ret;
//         Util::utf8_to_wstring( ret, s );
//         return ret;
//     }
//     string narrow( const wstring& s )
//     {
//         string ret;
//         Util::wstring_to_utf8( ret, s );
//         return ret;
//     }

//     typedef Loki::SmartPtr< ::lucene::document::Field, 
//                             Loki::RefLinked, 
//                             Loki::DisallowConversion, 
//                             FerrisLoki::FerrisExSmartPointerChecker, 
//                             Loki::DefaultSPStorage >  l_field;

//    typedef Hits* l_Hits;
//     typedef Loki::SmartPtr< Hits, 
//                             Loki::RefLinked, 
//                             Loki::DisallowConversion, 
//                             FerrisLoki::FerrisExSmartPointerChecker, 
//                             Loki::DefaultSPStorage >  l_Hits;

//    typedef Query* l_Query;
//     typedef Loki::SmartPtr< Query, 
//                             Loki::RefLinked, 
//                             Loki::DisallowConversion, 
//                             FerrisLoki::FerrisExSmartPointerChecker, 
//                             Loki::DefaultSPStorage >  l_Query;
//     typedef TermQuery* l_TermQuery;
// //     typedef Loki::SmartPtr< TermQuery, 
// //                             Loki::RefLinked, 
// //                             Loki::DisallowConversion, 
// //                             FerrisLoki::FerrisExSmartPointerChecker, 
// //                             Loki::DefaultSPStorage >  l_TermQuery;
//     typedef BooleanQuery* l_BooleanQuery;
// //     typedef Loki::SmartPtr< BooleanQuery, 
// //                             Loki::RefLinked, 
// //                             Loki::DisallowConversion, 
// //                             FerrisLoki::FerrisExSmartPointerChecker, 
//                             Loki::DefaultSPStorage >  l_BooleanQuery;

//     typedef Loki::SmartPtr< Term, 
//                             Loki::RefLinked, 
//                             Loki::AllowConversion, 
//                             FerrisLoki::FerrisExSmartPointerChecker, 
//                             Loki::DefaultSPStorage >  l_Term;

    
namespace FullTextIndex 
    {
        wstring W( const string& s )
        {
            wstring ret;
            return Util::utf8_to_wstring( ret, s );
        }

        string N( const wstring& ws )
        {
            string ret;
            return Util::wstring_to_utf8( ret, ws );
        }
        

        /****************************************/
        /****************************************/
        /****************************************/

        class FullTextIndexerCLucene
            :
            public MetaFullTextIndexerInterface
        {
            map< int, string > m_docIDHash;
            

            Analyzer* newAnalyzer();
            
        protected:
            
            virtual void Setup();
            virtual void CreateIndexBeforeConfig( fh_context c,
                                                  bool caseSensitive,
                                                  bool dropStopWords,
                                                  StemMode stemMode,
                                                  const std::string& lex_class,
                                                  fh_context md );
            virtual void CreateIndex( fh_context c,
                                      bool caseSensitive,
                                      bool dropStopWords,
                                      StemMode stemMode,
                                      const std::string& lex_class,
                                      fh_context md );
            virtual void addToIndex( fh_context c,
                                     fh_docindexer di );

            virtual docNumSet_t& addAllDocumentsMatchingTerm(
                const std::string& term,
                docNumSet_t& output );
            virtual void cleanDocumentIDCache();
            virtual std::string resolveDocumentID( docid_t );

        public:

            FullTextIndexerCLucene();

            docNumSet_t&
            ExecuteLuceneFullTextQuery( const std::string& queryString,
                                        docNumSet_t& docnums );
            virtual void* BuildCLuceneQuery( const std::string& qstr, void* current_query );
            
            
            static MetaFullTextIndexerInterface* Create();

            virtual bool getIndexMethodSupportsIsFileNewerThanIndexedVersion()
                { return true; }
            virtual bool isFileNewerThanIndexedVersion( const fh_context& c );
        };

        /************************************************************/
        /************************************************************/
        /************************************************************/



        class MyAnalyzer
            :
            public Analyzer
        {
            bool caseSensitive;
            bool dropStopWords;
            long stemMode;
            list< string > stopTable;
            char** stopWords;

            list< wstring > wstopTable;
            typedef wchar_t wstopWordsCHAR;
            wstopWordsCHAR** wstopWords;


            char**
            stringvec_to_CSTRvec( const list<string>& v )
                {
                    int l = v.size();
                    int i = 0;

                    list<string>::const_iterator ci = v.begin();
                    char** ret = g_new0(char*, l+2 );
                    for( i=0; i < l; ++i )
                    {
                        ret[i] = g_strdup( ci->c_str());
                        ++ci;
                    }
        
                    ret[i] = 0;
                    return ret;
                }
            
        public:
            MyAnalyzer( bool caseSensitive,
                        bool dropStopWords,
                        long stemMode,
                        stringlist_t droplist )
                :
                caseSensitive( caseSensitive ),
                dropStopWords( dropStopWords ),
                stemMode( stemMode ),
                stopWords( 0 ),
                stopTable( droplist )
                {
                    stopWords = stringvec_to_CSTRvec( stopTable );
                        
                    for( stringlist_t::iterator si = droplist.begin(); si != droplist.end(); ++si )
                        wstopTable.push_back( W( *si ) );
                    
                    wstopWords = new wstopWordsCHAR*[ wstopTable.size()+2 ];
                    int i = 0;
                    for( list<wstring>::iterator si = wstopTable.begin(); si != wstopTable.end(); ++si )
                    {
                        wstopWords[ i ] = (wstopWordsCHAR*)si->c_str();
                        ++i;
                    }
                }

            // 0.9.x
            TokenStream* tokenStream(const TCHAR* fieldName, util::Reader* reader)
//            TokenStream& tokenStream(const char_t* fieldName, lucene::util::Reader* reader)
                {
                    TokenStream* ret = new StandardTokenizer( reader );
                    ret = new StandardFilter( ret, true );
                    if( !caseSensitive )
                        ret = new LowerCaseFilter( ret, true );

                    if( dropStopWords )
                    {
                        ret = new StopFilter( ret, true, stopWords ); //, stopTable.size() );
                    }
//                     if( stemMode == (jint)STEM_PORTER )
//                         ret = new PorterStemFilter( ret );

                    return ret;
                    
//                    return new LowerCaseTokenizer( reader );
//                    return new StopFilter(new LowerCaseTokenizer(reader), stopTable);
                }
        };

        /**************************************************/
        /**************************************************/

        Analyzer*
        FullTextIndexerCLucene::newAnalyzer()
        {
//            return new StandardAnalyzer();

            stringlist_t sl = Util::parseCommaSeperatedList( 
                getConfig( IDXMGR_STOPWORDSLIST_K, IDXMGR_STOPWORDSLIST_DEFAULT ) );
            return new MyAnalyzer( isCaseSensitive(),
                                   getDropStopWords(),
                                   getStemMode(), sl );
        }
        
        
        FullTextIndexerCLucene::FullTextIndexerCLucene()
        {
        }
        
        
        void
        FullTextIndexerCLucene::Setup()
        {
        }

        void
        FullTextIndexerCLucene::CreateIndexBeforeConfig( fh_context c,
                                                        bool caseSensitive,
                                                        bool dropStopWords,
                                                        StemMode stemMode,
                                                        const std::string& lex_class,
                                                        fh_context md )
        {
            try
            {
                Analyzer* a = new StandardAnalyzer();
                l_IndexWriter writer = new IndexWriter( getPath().c_str(), a, true );
                writer->close();
            }
            catch ( exception& e )
            {
                throw;
            }
        }
        
        
        void
        FullTextIndexerCLucene::CreateIndex( fh_context c,
                                            bool caseSensitive,
                                            bool dropStopWords,
                                            StemMode stemMode,
                                            const std::string& lex_class,
                                            fh_context md )
        {
        }

        
        void
        FullTextIndexerCLucene::addToIndex( fh_context c,
                                           fh_docindexer di )
        {
            LG_IDX_D << "addToIndex() c:" << c->getURL() << endl;

            string s;
            fh_istream iss;
            try
            {
                fh_attribute a = c->getAttribute( "as-text" );
                iss = a->getIStream();
            }
            catch( exception& e )
            {
                LG_IDX_W << "WARNING, Failed to obtain plaintext for url:" << c->getURL()
                         << " error:" << e.what() << endl;
                return;
            }

            streamsize bytesDone    = 0;
            streamsize signalWindow = 0;
            streamsize totalBytes   = toType<streamsize>(getStrAttr( c, "size", "0" ));
            SignalThrottle throttle = getSignalThrottle();
            bool isCaseSensitive = this->isCaseSensitive();

            lucene::document::Document* doc = new lucene::document::Document();

//            IndexModifier writer = new IndexModifier( getPath().c_str(), *newAnalyzer(), false );
//            int32_t delCount = writer->deleteDocuments( new Term( "url", c->getURL().c_str() ) );
            {
//                IndexReader* writer = new IndexReader( getPath().c_str() );
                l_IndexReader writer = IndexReader::open( getPath().c_str() );
                
                {
                    typedef stringset_t skipnames_t;
                    skipnames_t skipnames;
                    skipnames.insert("content");
                    skipnames.insert("url");
                    skipnames.insert("ferris-url");
                    skipnames.insert("ferris-path");
                    skipnames.insert("ferris-ftx-modified");

                    l_Term t = new Term( "ferris-url", c->getURL().c_str() );
                    l_TermQuery pq = new TermQuery( t );
                    auto_ptr<Searcher> searcher( new IndexSearcher( getPath().c_str() ) );
                    l_Hits hits = searcher->search( pq );
                    if( hits->length() )
                    {
                        lucene::document::Document& olddoc = hits->doc(0);

                        DocumentFieldEnumeration* iter = olddoc.fields();
                        while( iter->hasMoreElements() )
                        {
                            Field* f = iter->nextElement();
                            if( !skipnames.count( f->name() ) )
                            {
                                doc->add( *f );
                            }
                        }
                    }
                }
                
                int32_t delCount = writer->deleteDocuments( new Term( "ferris-url", c->getURL().c_str() ) );
                writer->close();
            }

            l_IndexWriter writer = new IndexWriter( getPath().c_str(), newAnalyzer(), false );
            writer->setMaxFieldLength( 512*1024*1024 );
//            writer->setMaxFieldLength(IndexWriter::DEFAULT_MAX_FIELD_LENGTH);
            cerr << "IndexWriter::DEFAULT_MAX_FIELD_LENGTH:" << IndexWriter::DEFAULT_MAX_FIELD_LENGTH << endl;


            string mtime = getStrAttr( c, "mtime", "0" );
            doc->add( *Field::Keyword( "url",  c->getURL().c_str() ));
            doc->add( *Field::Keyword( "ferris-url",  c->getURL().c_str() ));
            doc->add( *Field::Keyword( "ferris-path", c->getDirPath().c_str() ));
            doc->add( *Field::Keyword( "ferris-ftx-modified", mtime.c_str() ));

            
            {
                LG_IDX_D << "addToIndex() c:" << c->getURL() <<  endl;
                
                StringBuffer str;
                string s;
                while( getline( iss, s ) )
                {
                    str.append( s.c_str() );
                    str.append( "\n" );
                }
//                 string contentstr = StreamToString( iss );
//                 cerr << "contentstr.sz:" << contentstr.size() << endl;
//                 str.append(contentstr.c_str());
                
                doc->add( *Field::Text( "contents", str.getBuffer() ));
            }
            
            writer->addDocument( doc );
            
            writer->optimize();
            writer->close();
        }

        std::string
        FullTextIndexerCLucene::resolveDocumentID( docid_t id )
        {
            return m_docIDHash[ id ];
        }
        
        void
        FullTextIndexerCLucene::cleanDocumentIDCache()
        {
            m_docIDHash.clear();
        }

        docNumSet_t&
        FullTextIndexerCLucene::ExecuteLuceneFullTextQuery( const std::string& queryString,
                                                            docNumSet_t& docnums )
        {
            LG_IDX_D << "ExecuteLuceneFullTextQuery() query:" << queryString << endl;
            addAllDocumentsMatchingTerm( queryString, docnums );
            return docnums;
        }

        void*
        FullTextIndexerCLucene::BuildCLuceneQuery( const std::string& qstr, void* current_query_vp )
        {
            ::lucene::search::BooleanQuery* query = (BooleanQuery*)(current_query_vp);

            string path = getPath();
            Searcher* searcher = new IndexSearcher( path.c_str() );
            Analyzer* analyzer = newAnalyzer();

            Query* subq = QueryParser::parse( qstr.c_str(), "contents", analyzer);
            query->add( subq, false, true, false );
            
            return query;
        }
        
        
        docNumSet_t&
        FullTextIndexerCLucene::addAllDocumentsMatchingTerm(
            const std::string& term_const,
            docNumSet_t& output )
        {
            try
            {
                
                string term = term_const;
                string path = getPath();
                Searcher* searcher = new IndexSearcher( path.c_str() );
                Analyzer* analyzer = newAnalyzer();

// FIXME: FIXME!                
//                 if( getStemMode() == STEM_PORTER )
//                 {
//                     PorterStemmer* s = new PorterStemmer();
//                     term = tostr( s->stem( tojstr( term ) ) );
//                 }
                if( getDropStopWords() )
                {
                    if( getStopWords().count( term_const ) )
                    {
                        cerr << "dropword! term:" << term_const << endl;
                        return output;
                    }
                }

                l_Query query = QueryParser::parse( term.c_str(), "contents", analyzer);
                LG_IDX_D << "Searching for:" << term << endl;

                l_Hits hits = searcher->search( GetImpl(query) );
                LG_IDX_D << "total matching documents:" << hits->length() << endl;

                for (int i = 0; i < hits->length(); i++)
                {
                    lucene::document::Document doc = hits->doc(i);
                    int docID     = hits->id( i );
                    output.insert( docID );
                    m_docIDHash[ docID ] = doc.get("ferris-url");
                }
            }
            catch ( exception& e )
            {
                LG_IDX_D << "error:" << e.what() << endl;
                throw;
            }
        }
        
        /**************************************************/
        /**************************************************/

        bool
        FullTextIndexerCLucene::isFileNewerThanIndexedVersion( const fh_context& c )
        {
            bool ret = true;

            time_t ct = getTimeAttr( c, "mtime", 0 );
            if( !ct )
                return ret;

            LG_IDX_D << "isFileNewerThanIndexedVersion() c:" << c->getURL() << endl;


            
            l_Term t = new Term( "ferris-url", c->getURL().c_str() );
            l_TermQuery pq = new TermQuery( t );

            auto_ptr<Searcher> searcher( new IndexSearcher( getPath().c_str() ) );
            l_Hits hits = searcher->search( pq );

            LG_IDX_D << "hits:" << hits->length() << " c:" << c->getURL() << endl;
            if( hits->length() )
            {
                time_t cachetime = 0;

                lucene::document::Document doc = hits->doc(0);

                if( !doc.get("ferris-ftx-modified") )
                    return ret;
                
                string s = doc.get("ferris-ftx-modified");
                if( s.empty() )
                    return ret;
                
                cachetime = toType<time_t>( s );
                LG_IDX_D << " cachetime:" << cachetime << " currenttime:" << ct
                         << " ret:" << (cachetime < ct) << endl;
                return( cachetime < ct );
            }

            return ret;
        }
    };
};


extern "C"
{
    FERRISEXP_API Ferris::FullTextIndex::MetaFullTextIndexerInterface* Create()
    {
        return new Ferris::FullTextIndex::FullTextIndexerCLucene();
    }
};
